#include <C/Filters/TRN_Filter.h>
#include <C/Common/TRN_API_Config.hpp>

// All C++ standard library and internal header files must be inside these guards.
// C standard header files (e.g. stdio.h) are recommended to be outside of these guards.
#if TRN_ENABLE_HIDE_SYMBOL_DIRECTIVES
#pragma GCC visibility push(hidden)
#endif // TRN_ENABLE_HIDE_SYMBOL_DIRECTIVES

#include <C/Common/TRN_Exception.hpp>
#include <Common/UString.h>
#include <Filters/CWrapFilterBase.h>
#include <Filters/Filter.h>
#include <Filters/StdFile.h>
#ifdef TRN_HAS_FILESYSTEM
#include <Filters/MappedFile.hpp>
#endif
#include <Filters/ASCII85Encode.h>
#include <Filters/FlateEncode.h>
#include <Filters/MemoryFilter.h>
#include <PDF/Image/Image2RGB.h>
#include <PDF/Image/Image2RGBA.h>

#ifdef ATLSTREAM
#include <Objidl.h>
#endif // ATLSTREAM

using namespace Common;
using namespace Filters;
using namespace PDF;
using namespace SDF;

#ifdef ATLSTREAM
class ATLStream : public BufFilter
{
public:
	ATLStream(IStream* stream) : Filters::BufFilter(AutoPtr<Filters::Filter>(NULL), 1024, true), m_max_stm_length((size_t)-1), m_pos(0), m_stream(stream), m_file_handle_owner(true), m_COMUnInitReq(false)
	{
		assert(m_stream);
		if (m_stream)
			m_stream->AddRef();
		m_thread_id = GetCurrentThreadId();
	}
	ATLStream(ATLStream& file) : 
		Filters::BufFilter(file, true),
		m_max_stm_length(file.m_max_stm_length),
		m_pos(file.m_pos),
		m_stream(file.m_stream),
		m_file_handle_owner(false),
		m_COMUnInitReq(false)
	{
		m_thread_id = GetCurrentThreadId();
	}
	virtual ~ATLStream()
	{
		if (m_file_handle_owner && m_stream) 
		{
			m_stream->Release();
		}
		if(m_COMUnInitReq)
			CoUninitialize();
	}
	void VerifyCom()
	{
		// keep getting CO_E_NOTINITIALIZED errors, this is to handles those
		DWORD id = GetCurrentThreadId();
		if(m_thread_id != id)
		{
			m_thread_id = id;
			HRESULT hr = CoInitialize(NULL);
			if (hr == S_OK)
			{
				m_COMUnInitReq = true;
			}
			else if (hr == S_FALSE)
			{
				CoUninitialize();
			}
			assert(hr == S_OK || hr == S_FALSE);
		}
	}
	const char* GetName() const { return "ATLStream"; }
	void Flush() {}
	size_t Process() {
		// return if the end of file segment was reached
		if (m_max_stm_length == 0) return 0;

		// Make sure that some other filter which is using the same stream 
		// didn't reposition internal file pointer or have corrupted the stream.
		//assert(ferror(m_stream) == 0);
		assert(m_stream != 0);

		//ptrdiff_t cur_pos = FFTELL(m_stream);
		LARGE_INTEGER dlibMove = {};
		ULARGE_INTEGER plibNewPosition = {};
		VerifyCom();
		HRESULT hr = m_stream->Seek(dlibMove, STREAM_SEEK_CUR, &plibNewPosition);
		assert(hr == S_OK);
		ptrdiff_t cur_pos = (ptrdiff_t)(plibNewPosition.QuadPart);
		if (cur_pos != m_pos) 
		{
			//BASE_ASSERT(FFSEEK(m_stream, m_pos, e_begin) == 0, "Operation file seek failed");
			dlibMove.QuadPart = m_pos;
			plibNewPosition.QuadPart = 0;
			VerifyCom();
			hr = m_stream->Seek(dlibMove, STREAM_SEEK_SET, &plibNewPosition);
			assert(hr == S_OK);
			//assert(m_pos == FFTELL(m_stream));
			assert(m_pos == (ptrdiff_t)plibNewPosition.QuadPart);
			Reset();
		}

		// Fill the non available area with data
		//size_t  count_bytes = fread(mp_buf_begin, 1, Min(m_max_stm_length, m_buf_size), m_stream);
		ULONG bytes_read = 0;
		VerifyCom();
		hr = m_stream->Read(mp_buf_begin, Min(m_max_stm_length, m_buf_size), &bytes_read);
		size_t count_bytes = (size_t)bytes_read;
		m_max_stm_length -= count_bytes;

		return count_bytes;
	}
	void SetStreamLength(size_t bytes) { m_max_stm_length = bytes; }
	bool CanSeek() { return true; }
	ptrdiff_t Tell() { return m_pos; }
	AutoPtr<Filter> CreateInputIterator () {
		BASE_ASSERT(CanSeek() && IsInputFilter(), "CreateInputIterator is not supported by this filter");
		AutoPtr<Filter> ret(new ATLStream(*this));
		return ret;
	}
	void Seek(ptrdiff_t offset, ReferencePos origin) {
		assert(CanSeek()); 
		DWORD dwOrigin;
		LARGE_INTEGER dlibMove = {};
		ULARGE_INTEGER plibNewPosition = {};
		switch(origin)
		{
			case Filter::e_begin : dwOrigin = STREAM_SEEK_SET; break;
			case Filter::e_end : dwOrigin = STREAM_SEEK_END; break;
			case Filter::e_cur : dwOrigin = STREAM_SEEK_CUR; break;
			default: assert(false); dwOrigin = STREAM_SEEK_CUR;					
		}
		dlibMove.QuadPart = offset;
		VerifyCom();
		HRESULT hr = m_stream->Seek(dlibMove, dwOrigin, &plibNewPosition);
		assert(hr == S_OK);
		m_pos = (ptrdiff_t)plibNewPosition.QuadPart;
		Reset();	// Among other things Reset will zero m_consumed_count
		//m_write_pos = m_pos;  // Set the current write position (used only in DataWrite method)
	}
	void Consume(size_t bytes) { m_pos += bytes; BufFilter::Consume(bytes); }
	ptrdiff_t FileSize() {
		STATSTG m_statstg;
		m_stream->Stat(&m_statstg, STATFLAG_NONAME);		
		return (ptrdiff_t)m_statstg.cbSize.QuadPart;
	}
protected:
	size_t		m_max_stm_length;
	ptrdiff_t	m_pos;
	IStream*	m_stream;
	bool		m_file_handle_owner;
	bool		m_COMUnInitReq;
	DWORD		m_thread_id;
};
#endif // ATLSTREAM

#if TRN_ENABLE_HIDE_SYMBOL_DIRECTIVES
#pragma GCC visibility pop
#endif // TRN_ENABLE_HIDE_SYMBOL_DIRECTIVES

#ifdef ATLSTREAM
TRN_API	TRN_FilterCreateFromStream (void* stm, TRN_Filter* result)
{
	BEX;
	*result=(TRN_Filter)new ATLStream((IStream*)stm);
	EEX;
}
#endif // ATLSTREAM

TRN_API TRN_FilterCreateASCII85Encode (TRN_Filter input_filter, TRN_UInt32 line_width, TRN_Size buf_sz, TRN_Filter* result)
{
	BEX;
	AutoPtr<Filter> in_filter((Filter*)input_filter);
	*result=(TRN_Filter)new ASCII85Encode(in_filter,line_width,buf_sz);
	EEX;
}

TRN_API TRN_FilterCreateFlateEncode (TRN_Filter input_filter, int compression_level, TRN_Size buf_sz, TRN_Filter* result)
{
	BEX;
	AutoPtr<Filter> in_filter((Filter*)input_filter);
	*result=(TRN_Filter)new FlateEncode(in_filter,compression_level,buf_sz);
	EEX;
}

#ifdef TRN_HAS_FILESYSTEM
TRN_API TRN_FilterCreateMappedFileFromUString (const TRN_UString filename, TRN_Filter* result)
{
	BEX;

	UString temp(UStrCast(filename));
	*result=(TRN_Filter)new MappedFile(temp);

	EEX;
}

TRN_API TRN_FilterMappedFileCompare(TRN_Filter mf1,TRN_Filter mf2, TRN_Bool* result)
{
	BEX;
	MappedFile* temp1=dynamic_cast<MappedFile*>((Filter*)mf1);
	BASE_ASSERT(temp1!=0,("The first filter is not a MappedFile"));

	MappedFile* temp2=dynamic_cast<MappedFile*>((Filter*)mf2);
	BASE_ASSERT(temp2!=0,("The second filter is not a MappedFile"));

	*result=BToTB(((*((MappedFile*)mf1))==(*((MappedFile*)mf2))));
	EEX;
}
#endif

TRN_API TRN_FilterCreateMemoryFilter(TRN_Size buf_sz, TRN_Bool is_input, TRN_Filter* result)
{
	BEX;
	*result=(TRN_Filter)new MemoryFilter(buf_sz,TBToB(is_input));
	EEX;
}

TRN_API TRN_FilterCreateImage2RGBFromElement(TRN_Element elem, TRN_Filter* result)
{
	BEX;
	*result=(TRN_Filter)new Image2RGB(EleCast(elem));
	EEX;
}

TRN_API TRN_FilterCreateImage2RGBFromObj(TRN_Obj obj, TRN_Filter* result)
{
	BEX;
	*result=(TRN_Filter)new Image2RGB(ObjCast(obj));
	EEX;
}

TRN_API TRN_FilterCreateImage2RGB(TRN_Image img, TRN_Filter* result)
{
	BEX;
	Image temp(ImgCast(img));
	*result=(TRN_Filter)new Image2RGB(&temp);
	EEX;
}

TRN_API TRN_FilterCreateImage2RGBAFromElement(TRN_Element elem, TRN_Bool premultiply, TRN_Filter* result)
{
	BEX;
	*result=(TRN_Filter)new Image2RGBA(EleCast(elem),TBToB(premultiply));
	EEX;
}

TRN_API TRN_FilterCreateImage2RGBAFromObj(TRN_Obj obj, TRN_Bool premultiply, TRN_Filter* result)
{
	BEX;
	*result=(TRN_Filter)new Image2RGBA(ObjCast(obj),TBToB(premultiply));
	EEX;
}

TRN_API TRN_FilterCreateImage2RGBA(TRN_Image img, TRN_Bool premultiply, TRN_Filter* result)
{
	BEX;
	Image temp(ImgCast(img));
	*result=(TRN_Filter)new Image2RGBA(temp,TBToB(premultiply));
	EEX;
}

TRN_API TRN_FilterDestroy (TRN_Filter filter)
{
	BEX;
	delete (Filter*)filter;
	EEX;
}

#ifdef TRN_HAS_FILESYSTEM
TRN_API TRN_FilterMappedFileFileSize(TRN_Filter filter, TRN_Size* result)
{
	BEX;
	MappedFile* temp=dynamic_cast<MappedFile*>((Filter*)filter);
	BASE_ASSERT(temp!=0,"This filter is not a MappedFile");
	UInt64 result_inner = temp->FileSize();
	*result = (TRN_Size)Filter::VerifyMemSized(result_inner);
	EEX;
}
#endif

TRN_API TRN_FilterAttachFilter(TRN_Filter filter, TRN_Filter attach_filter)
{
	BEX;
	AutoPtr<Filter> a_filter((Filter*)attach_filter);
	((Filter*)filter)->AttachFilter(a_filter);
	EEX;
}

TRN_API TRN_FilterReleaseAttachedFilter (TRN_Filter filter, TRN_Filter* result)
{
	BEX;
	*result=(TRN_Filter)((Filter*)filter)->ReleaseAttachedFilter().release();
	EEX;
}

TRN_API TRN_FilterGetAttachedFilter (TRN_Filter filter,	TRN_Filter* result)
{
	BEX;
	*result=(TRN_Filter)((Filter*)filter)->GetAttachedFilter();
	EEX;
}

TRN_API TRN_FilterGetSourceFilter (TRN_Filter filter, TRN_Filter* result)
{
	BEX;
	*result=(TRN_Filter)((Filter*)filter)->GetSourceFilter();
	EEX;
}

TRN_API TRN_FilterGetName (TRN_Filter const filter, const char** result)
{
	BEX;
	*result=((Filter*)filter)->GetName();
	EEX;
}

TRN_API TRN_FilterGetDecodeName (TRN_Filter const filter, const char** result)
{
	BEX;
	*result=((Filter*)filter)->GetDecodeName();
	EEX;
}
TRN_API TRN_FilterBegin (TRN_Filter filter, TRN_UChar** result)
{
	BEX;
	*result=((Filter*)filter)->Begin();
	EEX;
}

TRN_API TRN_FilterSize (TRN_Filter filter, TRN_Size* result)
{
	BEX;
	*result=((Filter*)filter)->Size();
	EEX;
}

TRN_API TRN_FilterConsume (TRN_Filter filter, TRN_Size num_bytes)
{
	BEX;
	((Filter*)filter)->Consume(num_bytes);
	EEX;
}

TRN_API TRN_FilterCount (TRN_Filter filter, TRN_Size* result)
{
	BEX;
	UInt64 result_inner = ((Filter*)filter)->Count();
	*result = (TRN_Size)Filter::VerifyMemSized(result_inner);
	EEX;
}

TRN_API TRN_FilterSetCount (TRN_Filter filter, TRN_Size new_count, TRN_Size* result)
{
	BEX;
	UInt64 result_inner = ((Filter*)filter)->SetCount(new_count);
	*result = (TRN_Size)Filter::VerifyMemSized(result_inner);
	EEX;
}

TRN_API TRN_FilterSetStreamLength (TRN_Filter filter, TRN_Size bytes)
{
	BEX;
	((Filter*)filter)->SetStreamLength(bytes);
	EEX;
}

TRN_API TRN_FilterFlush (TRN_Filter filter)
{
	BEX;
	((Filter*)filter)->Flush();
	EEX;
}

TRN_API TRN_FilterFlushAll (TRN_Filter filter)
{
	BEX;
	((Filter*)filter)->FlushAll();
	EEX;
}

TRN_API TRN_FilterIsInputFilter (TRN_Filter filter, TRN_Bool* result)
{
	BEX;
	*result=BToTB(((Filter*)filter)->IsInputFilter());
	EEX;
}

TRN_API TRN_FilterCanSeek (TRN_Filter filter, TRN_Bool* result)
{
	BEX;
	*result=BToTB(((Filter*)filter)->CanSeek());
	EEX;
}

TRN_API TRN_FilterSeek (TRN_Filter filter, TRN_Ptrdiff offset, enum TRN_FilterReferencePos origin)
{
	BEX;
	((Filter*)filter)->Seek(offset,(Filter::ReferencePos)origin);
	EEX;
}

TRN_API TRN_FilterTell (TRN_Filter filter, TRN_Ptrdiff* result)
{
	BEX;
	UInt64 result_inner = ((Filter*)filter)->Tell();	
	*result = Filter::VerifyMemSized(result_inner);
	EEX;
}

TRN_API TRN_FilterCreateInputIterator (TRN_Filter filter, TRN_Filter* result)
{
	BEX;
	*result=(TRN_Filter)((Filter*)filter)->CreateInputIterator().release();
	EEX;
}

TRN_API TRN_FilterGetFilePath(TRN_Filter const filter, TRN_UString* result)
{
	BEX;
	UString temp=((Filter*)filter)->GetFilePath();
	*result=(TRN_UString)(temp.mp_impl);
	temp.mp_impl=0;
	EEX;
}

TRN_API TRN_FilterWriteToFile(TRN_Filter filter, TRN_UString p, TRN_Bool append)
{
	BEX;
	UString path(UStrCast(p));
	((Filter*)filter)->WriteToFile(path, TBToB(append));
	EEX;
}

TRN_API TRN_FilterMemoryFilterGetBuffer(TRN_Filter filter, TRN_UChar** result)
{
	BEX;
	MemoryFilter* temp=dynamic_cast<MemoryFilter*>((Filter*)filter);
	BASE_ASSERT(temp!=0,("This filter is not a MemoryFilter"));
	*result=temp->GetBuffer();
	EEX;
}
	
TRN_API TRN_FilterMemoryFilterSetAsInputFilter(TRN_Filter filter)
{
	BEX;
	MemoryFilter* temp=dynamic_cast<MemoryFilter*>((Filter*)filter);
	BASE_ASSERT(temp!=0,("This filter is not a MemoryFilter"));
	temp->SetAsInputFilter();
	EEX;
}

TRN_API TRN_FilterMemoryFilterReset(TRN_Filter filter)
{
    BEX;
    MemoryFilter* temp = dynamic_cast<MemoryFilter*>((Filter*) filter);
    BASE_ASSERT(temp != 0, ("This filter is not a MemoryFilter"));
    temp->Reset();
    EEX;
}

TRN_API TRN_PathCompare(TRN_UString p1, TRN_UString p2, TRN_Bool* result)
{
	BEX;
	UString path1(UStrCast(p1)),path2(UStrCast(p2));
	*result=StdFile::Equivalent(path1,path2);
	EEX;
}
namespace trn {
	UInt64 NullTruncFunction(void*, UInt64)
	{
		return 0;
	}
}

TRN_API TRN_FilterCreateCustom(TRN_FilterStdFileOpenMode mode, void* user_data, TRN_SeekProc _seek, TRN_TellProc _tell, TRN_FlushProc _flush, TRN_ReadProc _read, TRN_WriteProc _write, TRN_CreateInputIteratorProc _iterator, TRN_DestroyProc _destroy, TRN_Filter* result)
{
	BEX;
	*result=(TRN_Filter)new CWrapFilterBase((Filters::StdFile::OpenMode)mode, user_data, _seek, _tell, trn::NullTruncFunction, _flush, _read, _write, _iterator, _destroy);
	EEX;
}

TRN_API TRN_FilterCreateCustomWithStruct(TRN_FilterStdFileOpenMode mode, void* user_data, TRN_CustomFilterCallbacks callbacks, TRN_Filter* result)
{
    BEX;
    *result=(TRN_Filter)new CWrapFilterBase((Filters::StdFile::OpenMode)mode, user_data, callbacks._seek, callbacks._tell, trn::NullTruncFunction, callbacks._flush, callbacks._read, callbacks._write, callbacks._createCustomIterator, callbacks._destroy);
    EEX;
}
