#include <C/Common/TRN_UString.h>
#include <C/SDF/TRN_SignatureHandler.h>
#include <C/Common/TRN_API_Config.hpp>

// All C++ standard library and internal header files must be inside these guards.
// C standard header files (e.g. stdio.h) are recommended to be outside of these guards.
#if TRN_ENABLE_HIDE_SYMBOL_DIRECTIVES
#pragma GCC visibility push(hidden)
#endif // TRN_ENABLE_HIDE_SYMBOL_DIRECTIVES

#include <cstring>
#include <C/Common/TRN_Exception.hpp>
#include <Common/Common.h>
#include <Common/Exception.h>
#include <SDF/SignatureHandler.h>

namespace trn {
namespace SDF {

// Internal class that extends the Signature base for CWrap users.
class CWrapSignatureHandler : public trn::SDF::SignatureHandler
{
public:
    CWrapSignatureHandler() : m_get_name(NULL),
                        m_append_data(NULL),
                        m_reset(NULL),
                        m_create_signature(NULL),
                        /*m_validate_signature(NULL),*/
                        m_destructor(NULL),
                        mp_client_impl(NULL),
                        mp_name_string(NULL)
    { }

    virtual UString GetName() const
    {
        BASE_ASSERT(m_get_name != NULL, "Missing implementation for SignatureHandler GetName.");
        TRN_UStringCreate((TRN_UString*) &mp_name_string); //TODO: Handle TRN_Exception
        TRN_Exception ex = m_get_name((TRN_UString*) &mp_name_string, mp_client_impl);
        if (ex != NULL) {
            // the throw below will be handled by PDFDoc::Save method.
            throw (Common::Exception(*((Common::Exception*) ex)));
        }
        return (UString(mp_name_string));
    }

    virtual void AppendData(const UInt8* in_data, const size_t in_length)
    {
        BASE_ASSERT(m_append_data != NULL, "Missing implementation for SignatureHandler AppendData.");
        TRN_SignatureData data;
        data.data = (TRN_UInt8*) in_data;
        data.length = in_length;
        TRN_Exception ex = m_append_data(data, mp_client_impl);
        if (ex != NULL) {
            // the throw below will be handled by PDFDoc::Save method.
            throw (Common::Exception(*((Common::Exception*) ex)));
        }
        return;
    }

    virtual bool Reset()
    {
        BASE_ASSERT(m_reset != NULL, "Missing implementation for SignatureHandler Reset.");
        TRN_Bool result = BToTB(false);
        TRN_Exception ex = m_reset(&result, mp_client_impl);
        if (ex != NULL) {
            // the throw below will be handled by PDFDoc::Save method.
            throw (Common::Exception(*((Common::Exception*) ex)));
        }
        return (TBToB(result));
    }

    virtual std::vector<UInt8> CreateSignature()
    {
        BASE_ASSERT(m_create_signature != NULL, "Missing implementation for SignatureHandler CreateSignature.");
        TRN_SignatureData data;
        TRN_Exception ex = m_create_signature(&data, mp_client_impl);
        if (ex != NULL) {
            // the throw below will be handled by PDFDoc::Save method.
            throw (Common::Exception(*((Common::Exception*) ex)));
        }
        std::vector<UInt8> result;
        if (data.data != NULL) {
            result.resize(data.length);
            memcpy(&(result[0]), data.data, data.length);
        }
        return (result);
    }

	/*
    virtual VerificationResult VerifySignature(const PDF::DigitalSignatureField& in_digsig, const SDF::VerificationOptions& in_opts)
    {
        
        BASE_ASSERT(m_validate_signature != NULL, "Missing implementation for SignatureHandler VerifySignature.");
        TRN_ValidateSignatureResult vresult = m_validate_signature((TRN_Obj) in_sigDict, mp_client_impl);
        SignatureVerificationResult result;
        result.m_isvalid = TBToB(vresult.valid);
        result.m_errorcode = vresult.error_code;
        result.m_message = UString(vresult.message);
        
		SDF::VerificationResult result(in_digsig);
        return (result);
    }
	*/


    virtual ~CWrapSignatureHandler()
    {
        if (m_destructor != NULL)
            m_destructor(mp_client_impl);

        if (mp_name_string != NULL)
            TRN_UStringDestroy(mp_name_string);
    }

    TRN_SignatureHandlerGetNameFunction m_get_name;
    TRN_SignatureHandlerAppendDataFunction m_append_data;
    TRN_SignatureHandlerResetFunction m_reset;
    TRN_SignatureHandlerCreateSignatureFunction m_create_signature;
    //TRN_SignatureHandlerValidateSignatureFunction m_validate_signature;
    TRN_SignatureHandlerDestructorFunction m_destructor;
    void* mp_client_impl;
private:
    TRN_UString mp_name_string;
    CWrapSignatureHandler(CWrapSignatureHandler& src) { }
    CWrapSignatureHandler& operator=(CWrapSignatureHandler& src) { }
}; // class CWrapSignatureHandler

}; // namespace SDF
}; // namespace trn

#if TRN_ENABLE_HIDE_SYMBOL_DIRECTIVES
#pragma GCC visibility pop
#endif // TRN_ENABLE_HIDE_SYMBOL_DIRECTIVES

// C Wrapper for creating custom signatures
TRN_API TRN_SignatureHandlerCreate(TRN_SignatureHandlerGetNameFunction get_name,
                            TRN_SignatureHandlerAppendDataFunction append_data,
                            TRN_SignatureHandlerResetFunction reset,
                            TRN_SignatureHandlerCreateSignatureFunction create_signature,
                            /*TRN_SignatureHandlerValidateSignatureFunction validate_signature,*/
                            TRN_SignatureHandlerDestructorFunction destructor,
                            void* unused,
                            TRN_SignatureHandler* signature_handler)
{
    BEX;
    trn::SDF::CWrapSignatureHandler* temp = new trn::SDF::CWrapSignatureHandler();
    *signature_handler = (TRN_SignatureHandler) temp;
    temp->m_get_name = get_name;
    temp->m_append_data = append_data;
    temp->m_reset = reset;
    temp->m_create_signature = create_signature;
    //temp->m_validate_signature = validate_signature;
    temp->m_destructor = destructor;
    temp->mp_client_impl = unused;
    
    EEX;
}

TRN_API TRN_SignatureHandlerDestroy(TRN_SignatureHandler signature_handler, void** unused)
{
    BEX;
    trn::SDF::CWrapSignatureHandler* temp = (trn::SDF::CWrapSignatureHandler*) signature_handler;
    if (unused != NULL) {
        *unused = temp->mp_client_impl;
        // set the destructor for userimpl to NULL
        temp->m_destructor = NULL;
        temp->mp_client_impl = NULL;
    }
    delete(temp);
    EEX;
}

TRN_API TRN_SignatureHandlerGetName(TRN_SignatureHandler signature_handler, TRN_UString* result)
{
    BEX;
    trn::SDF::CWrapSignatureHandler* temp = (trn::SDF::CWrapSignatureHandler*) signature_handler;
    BASE_ASSERT(temp->m_get_name != NULL, "TRN_SignatureHandlerGetName undefined.");
    TRN_Exception ex = temp->m_get_name(result, temp->mp_client_impl);
    if (ex != NULL) {
        // the EEX macro will deal with this
        throw (Common::Exception(*((Common::Exception*) ex)));
    }
    EEX;
}

TRN_API TRN_SignatureHandlerAppendData(TRN_SignatureHandler signature_handler, const TRN_SignatureData data)
{
    BEX;
    trn::SDF::CWrapSignatureHandler* temp = (trn::SDF::CWrapSignatureHandler*) signature_handler;
    BASE_ASSERT(temp->m_append_data != NULL, "TRN_SignatureHandlerAppendData undefined.");
    TRN_Exception ex = temp->m_append_data(data, temp->mp_client_impl);
    if (ex != NULL) {
        // the EEX macro will deal with this
        throw (Common::Exception(*((Common::Exception*) ex)));
    }
    EEX;
}

TRN_API TRN_SignatureHandlerReset(TRN_SignatureHandler signature_handler, TRN_Bool* result)
{
    BEX;
    trn::SDF::CWrapSignatureHandler* temp = (trn::SDF::CWrapSignatureHandler*) signature_handler;
    BASE_ASSERT(temp->m_reset != NULL, "TRN_SignatureHandlerReset undefined.");
    TRN_Exception ex = temp->m_reset(result, temp->mp_client_impl);
    if (ex != NULL) {
        // the EEX macro will deal with this
        throw (Common::Exception(*((Common::Exception*) ex)));
    }
    EEX;
}

TRN_API TRN_SignatureHandlerCreateSignature(TRN_SignatureHandler signature_handler, TRN_SignatureData* result)
{
    BEX;
    trn::SDF::CWrapSignatureHandler* temp = (trn::SDF::CWrapSignatureHandler*) signature_handler;
    BASE_ASSERT(temp->m_create_signature != NULL, "TRN_SignatureHandlerCreateSignature undefined.");
    TRN_Exception ex = temp->m_create_signature(result, temp->mp_client_impl);
    if (ex != NULL) {
        // the EEX macro will deal with this
        throw (Common::Exception(*((Common::Exception*) ex)));
    }
    EEX;
}
/*
TRN_API TRN_SignatureHandlerValidateSignature(TRN_SignatureHandler signature_handler, TRN_Obj sig_dict, TRN_ValidateSignatureResult* result)
{
    BEX;
    trn::SDF::CWrapSignatureHandler* temp = (trn::SDF::CWrapSignatureHandler*) signature_handler;
    if (temp->m_validate_signature == NULL)
        throw (std::exception("TRN_SignatureHandlerValidateSignature undefined."));
    *result = temp->m_validate_signature(sig_dict, temp->mp_client_impl);
    EEX;
}
*/
TRN_API TRN_SignatureHandlerDestructor(TRN_SignatureHandler signature_handler)
{
    BEX;
    trn::SDF::CWrapSignatureHandler* temp = (trn::SDF::CWrapSignatureHandler*) signature_handler;
    BASE_ASSERT(temp->m_destructor != NULL, "TRN_SignatureHandlerDestructor undefined.");
    temp->m_destructor(temp->mp_client_impl);
    EEX;
}

TRN_API TRN_SignatureHandlerGetUserImpl(TRN_SignatureHandler signature_handler, void** result)
{
    BEX;
    trn::SDF::CWrapSignatureHandler* temp = (trn::SDF::CWrapSignatureHandler*) signature_handler;
    if (result != NULL)
        *result = temp->mp_client_impl;
    EEX;
}
