#include <C/Common/TRN_Matrix2D.h>
#include <C/Common/TRN_API_Config.hpp>

// All C++ standard library and internal header files must be inside these guards.
// C standard header files (e.g. stdio.h) are recommended to be outside of these guards.
#if TRN_ENABLE_HIDE_SYMBOL_DIRECTIVES
#pragma GCC visibility push(hidden)
#endif // TRN_ENABLE_HIDE_SYMBOL_DIRECTIVES

#include <cmath>
#include <C/Common/TRN_Exception.hpp>

#if TRN_ENABLE_HIDE_SYMBOL_DIRECTIVES
#pragma GCC visibility pop
#endif // TRN_ENABLE_HIDE_SYMBOL_DIRECTIVES

TRN_API TRN_Matrix2DCopy(const TRN_Matrix2D* m, TRN_Matrix2D* result)
{
	result->m_a=m->m_a;
	result->m_b=m->m_b;
	result->m_c=m->m_c;
	result->m_d=m->m_d;
	result->m_h=m->m_h;
	result->m_v=m->m_v;
	return 0;
}

TRN_API TRN_Matrix2DSet (TRN_Matrix2D* matrix, double a, double b, double c, double d, double h, double v)
{
	matrix->m_a=a;
	matrix->m_b=b;
	matrix->m_c=c;
	matrix->m_d=d;
	matrix->m_h=h;
	matrix->m_v=v;
	return 0;
}

TRN_API TRN_Matrix2DConcat(TRN_Matrix2D* matrix, double a, double b, double c, double d, double h, double v)
{
	double a1 = matrix->m_a;
	double b1 = matrix->m_b;
	double c1 = matrix->m_c;
	double d1 = matrix->m_d;

	matrix->m_a = a * a1 + b * c1;
	matrix->m_b = a * b1 + b * d1;
	matrix->m_c = c * a1 + d * c1;
	matrix->m_d = c * b1 + d * d1;
	matrix->m_h = h * a1 + v * c1 + matrix->m_h;
	matrix->m_v = h * b1 + v * d1 + matrix->m_v;
	return 0;
}




TRN_API TRN_Matrix2DEquals (const TRN_Matrix2D* m1, const TRN_Matrix2D* m2, TRN_Bool* result)
{
	if (m1->m_a == m2->m_a &&
		m1->m_b == m2->m_b &&
		m1->m_c == m2->m_c &&
		m1->m_d == m2->m_d &&
		m1->m_h == m2->m_h &&
		m1->m_v == m2->m_v) 
	{
		*result=(TRN_Bool)true;
	}
	else
		*result=0;
	return 0;
}


TRN_API TRN_Matrix2DMult(const TRN_Matrix2D* matrix, double* in_out_x, double* in_out_y)
{
	double in_x = (*in_out_x);
	(*in_out_x) = (*in_out_x)*(matrix->m_a) + (*in_out_y)*(matrix->m_c) + (matrix->m_h);
	(*in_out_y) = in_x*(matrix->m_b) + (*in_out_y)*(matrix->m_d) + matrix->m_v;
	return 0;
}


TRN_API TRN_Matrix2DInverse (const TRN_Matrix2D* matrix, TRN_Matrix2D* result)
{
	double det = matrix->m_a * matrix->m_d - matrix->m_b * matrix->m_c;
	if (det == 0) 
	{
		TRN_Matrix2DCreateIdentityMatrix(result);
		std::string t("The given matrix is not invertible");
		return GetException(t);
	}

	result->m_a = matrix->m_d / det;
	result->m_c = -matrix->m_c / det;
	result->m_h = (matrix->m_c * matrix->m_v - matrix->m_h * matrix->m_d) / det;

	result->m_b = -matrix->m_b / det;
	result->m_d = matrix->m_a / det;
	result->m_v = -(matrix->m_a * matrix->m_v - matrix->m_h * matrix->m_b) / det;
	return 0;
}


TRN_API TRN_Matrix2DTranslate (TRN_Matrix2D* matrix, double h, double v)
{
	matrix->m_h += h;
	matrix->m_v += v;
	return 0;
}


TRN_API TRN_Matrix2DScale (TRN_Matrix2D* matrix, double h, double v)
{
	matrix->m_a *= h;
	matrix->m_d *= v;
	return 0;
}


TRN_API TRN_Matrix2DCreateZeroMatrix (TRN_Matrix2D* result)
{
	result->m_a=0;
	result->m_b=0;
	result->m_c=0;
	result->m_d=0;
	result->m_h=0;
	result->m_v=0;
	return 0;
}


TRN_API TRN_Matrix2DCreateIdentityMatrix (TRN_Matrix2D* result)	
{
	result->m_a=1;
	result->m_b=0;
	result->m_c=0;
	result->m_d=1;
	result->m_h=0;
	result->m_v=0;
	return 0;
}


TRN_API TRN_Matrix2DCreateRotationMatrix (const double angle, TRN_Matrix2D* result)
{
	double c = cos(angle);
	double s = sin(angle);
	result->m_a=c;
	result->m_b=-s;
	result->m_c=s;
	result->m_d=c;
	result->m_h=0;
	result->m_v=0;
	return 0;
}
