#include <C/PDF/TRN_PDFRasterizer.h>
#include <PDF/Render/PDFAGG/RenderPage.hpp>
#include <C/Common/TRN_API_Config.hpp>
#include <PDF/Annots/AnnotationSource.hpp>

// All C++ standard library and internal header files must be inside these guards.
// C standard header files (e.g. stdio.h) are recommended to be outside of these guards.
#if TRN_ENABLE_HIDE_SYMBOL_DIRECTIVES
#pragma GCC visibility push(hidden)
#endif // TRN_ENABLE_HIDE_SYMBOL_DIRECTIVES

#include <C/Common/TRN_Exception.hpp>
#include <C/Common/TRN_Vector.h>
#include <C/Common/TRN_UString.h>
#include <PDF/PDFRasterizer.h>
#include <SDF/License.hpp>
#include <Common/ArrayRef.hpp>
#include <Filters/MemoryFilter.h>
#include <Common/UStringImpl.hpp>
#include <C/Common/TRN_PolymorphicVector.hpp>
#if TRN_ENABLE_HIDE_SYMBOL_DIRECTIVES
#pragma GCC visibility pop
#endif // TRN_ENABLE_HIDE_SYMBOL_DIRECTIVES

using namespace SDF;
using namespace PDF;

TRN_API TRN_PDFRasterizerCreate(enum TRN_PDFRasterizerType type, TRN_PDFRasterizer* result)
{
	BEX;
	License::Instance().VerifyPermission(License::e_perm_rast);

	*result=(TRN_PDFRasterizer)new PDFRasterizer(false, (PDFRasterizer::Type)type);
	EEX;
}

TRN_API TRN_PDFRasterizerDestroy(TRN_PDFRasterizer r)
{
	BEX;
	delete (PDFRasterizer*)r;
	EEX;
}

TRN_API TRN_PDFRasterizerRasterizeToMemory(TRN_PDFRasterizer r,TRN_Page page, TRN_UChar* in_out_image_buffer, 
				int width, int height, int stride, 
				int num_comps,
				TRN_Bool demult,
				const TRN_Matrix2D* device_mtx,
				const TRN_Rect* clip,
				const TRN_Rect* scrl_clp_regions,
				volatile TRN_Bool* cancel)
{
	BEX;
	Page temp=Page(PageCast(page));
	RastCast(r).Rasterize(temp,in_out_image_buffer,width,height,stride, num_comps,
		TBToB(demult), M2DCast(device_mtx),(const Rect*)clip,(const iAnnotationSource*)0,TBToBP(cancel));
	EEX;
}

TRN_API TRN_PDFRasterizerRasterizeAnnot(TRN_PDFRasterizer r, TRN_Annot annot, TRN_Page page, 
	const TRN_Matrix2D* device_mtx, TRN_Bool demult, volatile TRN_Bool* cancel, TRN_OwnedBitmap* result)
{
	BEX;
	Annot a(reinterpret_cast<SDF::Obj*>(annot));
	Rect clip = a.GetRect();
	SingleAnnotSource source(a);
	Page temp = Page(PageCast(page));
	*result = (TRN_OwnedBitmap)RastCast(r).RasterizeAnnot(a,temp, M2DCast(device_mtx), TBToB(demult), TBToBP(cancel)).release();
	EEX;
}

#if defined(_WIN32) && !defined(__WINRT__)
TRN_API TRN_PDFRasterizerRasterizeToDevice(TRN_PDFRasterizer r,TRN_Page page, void* hdc, 
		const TRN_Matrix2D* device_mtx, 
		const TRN_Rect* clip,int dpi,
		volatile TRN_Bool* cancel)
{
	BEX;
	Page temp=Page(PageCast(page));
	RastCast(r).Rasterize(temp,hdc,M2DCast(device_mtx),(const Rect*)clip,dpi,TBToBP(cancel));
	EEX;
}
#endif

TRN_API TRN_PDFRasterizerSetDrawAnnotations(TRN_PDFRasterizer r,TRN_Bool render_annots)
{
	BEX;
	RastCast(r).SetDrawAnnotations(TBToB(render_annots));
	EEX;
}

TRN_API TRN_PDFRasterizerSetHighlightFields(TRN_PDFRasterizer r,TRN_Bool highlight)
{
	BEX;
	RastCast(r).SetHighlightFields(TBToB(highlight));
	EEX;
}

TRN_API TRN_PDFRasterizerSetGamma(TRN_PDFRasterizer r, double expgamma)
{
	BEX;
	RastCast(r).SetGamma(expgamma);
	EEX;
}

TRN_API TRN_PDFRasterizerSetOCGContext(TRN_PDFRasterizer r, TRN_OCGContext ctx)
{
	BEX;
	RastCast(r).SetOCGContext((OCG::Context*)ctx);
	EEX;
}

TRN_API TRN_PDFRasterizerSetPrintMode(TRN_PDFRasterizer r, TRN_Bool is_printing)
{
	BEX;
	RastCast(r).SetPrintMode(TBToB(is_printing));
	EEX;
}

TRN_API TRN_PDFRasterizerSetAntiAliasing(TRN_PDFRasterizer r, TRN_Bool enable_aa)
{
	BEX;
	RastCast(r).SetAntiAliasing(TBToB(enable_aa));
	EEX;
}

TRN_API TRN_PDFRasterizerSetPathHinting(TRN_PDFRasterizer r, TRN_Bool enable_hinting)
{
	BEX;
	RastCast(r).SetPathHinting(TBToB(enable_hinting));
	EEX;
}

TRN_API TRN_PDFRasterizerSetThinLineAdjustment(TRN_PDFRasterizer r, TRN_Bool grid_fit, TRN_Bool stroke_adjust)
{
	BEX;
	RastCast(r).SetThinLineAdjustment(TBToB(grid_fit), TBToB(stroke_adjust));
	EEX;
}

TRN_API TRN_PDFRasterizerSetImageSmoothing(TRN_PDFRasterizer r, TRN_Bool smoothing_enabled, TRN_Bool hq_image_resampling)
{
	BEX;
	RastCast(r).SetImageSmoothing(TBToB(smoothing_enabled), TBToB(hq_image_resampling));
	EEX;
}

TRN_API TRN_PDFRasterizerSetOverprint(TRN_PDFRasterizer r, enum TRN_PDFRasterizerOverprintPreviewMode op)
{
	BEX;
	RastCast(r).SetOverprint(PDFRasterizer::OverprintPreviewMode(op));
	EEX;
}

TRN_API TRN_PDFRasterizerSetCaching(TRN_PDFRasterizer r, TRN_Bool enabled)
{
	BEX;
	RastCast(r).SetCaching(TBToB(enabled));
	EEX;
}

TRN_API TRN_PDFRasterizerSetAnnotationState(TRN_PDFRasterizer r, TRN_Annot annot, enum TRN_AnnotState new_view_state)
{
	BEX;
	Annot a((Obj*)annot);
	RastCast(r).SetAnnotationState(a.GetSDFObj(), (Annot::AnnotationState)new_view_state);
	EEX;
}

TRN_API TRN_PDFRasterizerSetErrorReportProc(TRN_PDFRasterizer r, TRN_RasterizerErrorReportProc error_proc, void* data)
{
	BEX;
	RastCast(r).SetErrorReportProc((PDFRasterizer::ErrorReportProc)error_proc,data);
	EEX;
}

TRN_API TRN_PDFRasterizerSetRasterizerType (TRN_PDFRasterizer r, enum TRN_PDFRasterizerType type)
{
	BEX;
	RastCast(r).SetRasterizerType((PDFRasterizer::Type)type);
	EEX;
}

TRN_API TRN_PDFRasterizerGetRasterizerType (TRN_PDFRasterizer r, enum TRN_PDFRasterizerType* result)
{
	BEX;
	*result=(enum TRN_PDFRasterizerType)RastCast(r).GetRasterizerType();
	EEX;
}

TRN_API TRN_PDFRasterizerSetColorPostProcessMode(TRN_PDFRasterizer r, enum TRN_PDFRasterizerColorPostProcessMode mode)
{
	BEX;
	RastCast(r).SetColorPostProcessing((PDFRasterizer::ColorPostProcessMode)mode);
	EEX;
}

TRN_API TRN_PDFRasterizerGetColorPostProcessMode(TRN_PDFRasterizer r, enum TRN_PDFRasterizerColorPostProcessMode* result)
{
	BEX;
	*result=(enum TRN_PDFRasterizerColorPostProcessMode)RastCast(r).GetColorPostProcessing();
	EEX;
}

TRN_API TRN_PDFRasterizerGetChunkRenderer(TRN_PDFRasterizer r,TRN_Page page, TRN_UChar* in_out_image_buffer, 
										  int width, int height, int stride, 
										  int num_comps,
										  TRN_Bool demult,
										  const TRN_Matrix2D* device_mtx,
										  const TRN_Rect* clip,
										  const TRN_Rect* scrl_clp_regions,
										  volatile TRN_Bool* cancel,
										  TRN_ChunkRenderer* result)
{
	BEX;
	Page temp=Page(PageCast(page));
	*result=(TRN_ChunkRenderer)(RastCast(r).GetChunkRenderer(temp,in_out_image_buffer,width,height,stride, num_comps,
		TBToB(demult), M2DCast(device_mtx),(const Rect*)clip,(const iAnnotationSource*)0,TBToBP(cancel))
		.release());
	EEX;
}

TRN_API TRN_PDFRasterizerGetChunkRendererPath(TRN_PDFRasterizer r, TRN_Page page, TRN_UString file_path,
	int width, int height,
	TRN_Bool demult,
	const TRN_Matrix2D* device_mtx,
	const TRN_Rect* clip,
	const TRN_Rect* scrl_clp_regions,
	volatile TRN_Bool* cancel,
	TRN_ChunkRenderer* result)
{
	BEX;
	Page temp = Page(PageCast(page));
	UString path(file_path);
	*result = (TRN_ChunkRenderer)(RastCast(r).GetChunkRenderer(temp, path, width, height,
		TBToB(demult), M2DCast(device_mtx), (const Rect*)clip, (const iAnnotationSource*)0, TBToBP(cancel))
		.release());
	EEX;
}


TRN_API TRN_ChunkRendererRenderNext(TRN_ChunkRenderer cr, TRN_Bool* result)
{
	BEX;
	*result=BToTB(((ChunkRenderer*)cr)->RenderNext());
	EEX;
}
TRN_API TRN_ChunkRendererDestroy(TRN_ChunkRenderer cr)
{
	BEX;
	delete ((ChunkRenderer*)cr);
	EEX;
}

TRN_API TRN_PDFRasterizerUpdateBuffer(TRN_PDFRasterizer r)
{
	BEX;
	RastCast(r).UpdateBuffer();
	EEX;
}

TRN_API TRN_PDFRasterizerEnableDisplayListCaching(TRN_PDFRasterizer r, TRN_Bool enabled)
{
	BEX;
	RastCast(r).SetDispCache(TBToB(enabled));
	EEX;
}

TRN_API TRN_PDFRasterizerRasterizeSeparations(TRN_PDFRasterizer r, TRN_Page page, int width, int height, const TRN_Matrix2D* mtx,
	const TRN_Rect* clip, volatile TRN_Bool* cancel, TRN_Vector* result)
{
	BEX;
	ConstArrayRef<PDF::Rect> clip_rects;
	if (clip)
	{
		clip_rects = ConstArrayRef<PDF::Rect>(&RectCast(clip), 1);
	}
	else
	{
		clip_rects = ConstArrayRef<PDF::Rect>();
	}
	FlexVector<tOwnedBuffer> output_buffers;
	Page pdf_page(PageCast(page));
	FlexVector<OutputSeparation> separations = RastCast(r).RasterizeSeparations(pdf_page, Common::Size2D<int>(width, height), M2DCast(mtx), clip_rects, output_buffers, TBToBP(cancel));
	
	PolyVectorSeparation* final_separations = new PolyVectorSeparation();
	*result = (TRN_Vector) final_separations;
	final_separations->m_vector.resize(separations.size());
	try
	{
		
		for (UInt32 i = 0; i < separations.size(); ++i)
		{
			OutputSeparation& current = separations[i];

			TRN_Separation new_separation;
			UInt32 separation_buffer_size = current.m_bitmap.W()*current.m_bitmap.H();
			
			new_separation.data = new UChar[separation_buffer_size];
			AutoPtr<Filters::Filter> channel_filter(current.GetFilterForChannel());
			Filters::FilterReader reader(*channel_filter);
			reader.Read(new_separation.data, separation_buffer_size);

			new_separation.data_size = separation_buffer_size;

			new_separation.c = current.m_separation_color.c();
			new_separation.m = current.m_separation_color.m();
			new_separation.y = current.m_separation_color.y();
			new_separation.k = current.m_separation_color.k();
			new_separation.separation_name = (TRN_UString) IMPL_TRN_UStringCreateFromCharString(current.m_separation_name.c_str(), signed32(current.m_separation_name.size()), UString::e_utf8);	
			final_separations->m_vector[i] = new_separation;
		}
	}
	catch (...)
	{
		TRN_VectorDestroy(*result);
		throw;
	}
	EEX;
}

TRN_API TRN_SeparationDestroy(TRN_Separation* sep)
{
	BEX
	delete[] sep->data;
	EEX;
}

