//---------------------------------------------------------------------------------------
// Copyright (c) 2001-2020 by PDFTron Systems Inc. All Rights Reserved.
// Consult legal.txt regarding legal and license information.
//---------------------------------------------------------------------------------------

#import "PTCollaborationAnnotation.h"
#import "PTToolManager.h"

#import <Foundation/Foundation.h>

@class PTCollaborationManager;
@class PTToolManager;

NS_ASSUME_NONNULL_BEGIN

/**
 * The `PTCollaborationServer` protocol allows the adopting class to be used for communicating
 * annotation changes that occur on a shared document.
 */
@protocol PTCollaborationServerCommunication <NSObject>
@required

/**
 * The `PTCollaborationManager` that registered this service implementation in
 * `-[PTCollaborationManager registerServerCommunicationComponent:]`.
 */
@property (nonatomic, weak, nullable) PTCollaborationManager* collaborationManager;

/**
 * The implementation of this method sends the local annotation change to the server.
 * This method is called by `PTCollaborationManager`.
 *
 * @param collaborationAnnotation an object representing the collaboration annotation.
 */
-(void)localAnnotationAdded:(PTCollaborationAnnotation*)collaborationAnnotation;

/**
 * The implementation of this method sends the local annotation change to the server.
 * This method is called by `PTCollaborationManager`.
 *
 * @param collaborationAnnotation an object representing the collaboration annotation.
 */
-(void)localAnnotationModified:(PTCollaborationAnnotation*)collaborationAnnotation;

/**
 * The implementation of this method sends the local annotation change to the server.
 * This method is called by `PTCollaborationManager`.
 *
 * @param collaborationAnnotation an object representing the collaboration annotation.
 */
-(void)localAnnotationRemoved:(PTCollaborationAnnotation*)collaborationAnnotation;

/**
 *
 * This method is called by the local document viewer when it is ready for remote annotations.
 *
 * The implemention of this method applies all remotely stored annotations to the local document.
 *
 */
-(void)documentLoaded;

/**
 * The unqiue ID that represents the current user.
 */
@property (nonatomic, readonly, nullable, copy) NSString *userID;

/**
 * The unqiue ID that represents the current document.
 */
@property (nonatomic, readonly, nullable, copy) NSString *documentID;

@end

/**
 *
 * The object that is responsible for sending local changes to the remote server, and for receiving remote changes
 * from the server.
 *
 */
@interface PTCollaborationManager : NSObject

/**
 * Intializes a new PTCollaborationManager.
 *
 * @param toolManager The tool manager that this object uses to register for annotation change events that
 * occur locally, and to push back changes that happen remotely (via its annotManager).
 *
 * @param userId The identifier for the current collaboration user.
 */
- (instancetype)initWithToolManager:(PTToolManager *)toolManager userId:(NSString *)userId NS_DESIGNATED_INITIALIZER;

/**
 * The tool manager used by this object to receive local annotation change events.
 */
@property (nonatomic, readonly, strong) PTToolManager *toolManager;

#pragma mark - Initial annotations

/**
 * Call this method to load an initial remote annotation. This should generally be done immediately
 * after the document itself is loaded.
 *
 * @param collaborationAnnotation an object representing the collaboration annotation.
 */
- (void)loadInitialRemoteAnnotation:(PTCollaborationAnnotation*)collaborationAnnotation;

/**
 * Call this method to load a list of initial remote annotations. This should generally be done immediately
 * after the document itself is loaded.
 *
 * @param collaborationAnnotations an array of objects representing the collaboration annotations.
 */
- (void)loadInitialRemoteAnnotations:(NSArray<PTCollaborationAnnotation *> *)collaborationAnnotations;

#pragma mark - Importing XFDF

/**
 * Imports the annotations specified by the XFDF string. The imported annotations are considered to
 * be part of the initial set of annotations.
 *
 * @param xfdfString the XFDF string to import
 */
- (void)importAnnotationsWithXFDFString:(NSString *)xfdfString;

/**
* Imports the annotations specified by the XFDF string.
*
* @param xfdfString the XFDF string to import
*
* @param isInitial whether the imported annotations should be considered to be part of the initial
* set of annotations.
*/
- (void)importAnnotationsWithXFDFString:(NSString *)xfdfString isInitial:(BOOL)isInitial;

/**
 * Imports the annotations specified by the XFDF command string. The command string can contain <add>,
 * <modify>, and <delete> tags. The imported annotations are considered to be part of the initial
 * set of annotations.
 *
 * @param xfdfCommand the XFDF command string to import
 */
- (void)importAnnotationsWithXFDFCommand:(NSString *)xfdfCommand;

/**
* Imports the annotations specified by the XFDF command string. The command string can contain <add>,
* <modify>, and <delete> tags.
*
* @param xfdfCommand the XFDF command string to import
*
* @param isInitial whether the imported annotations should be considered to be part of the initial
* set of annotations.
*/
- (void)importAnnotationsWithXFDFCommand:(NSString *)xfdfCommand isInitial:(BOOL)isInitial;

#pragma mark - Remote annotation events

/**
 * Call this method to communicate a change that occured remotely that needs to be reflected on the device.
 *
 * @param collaborationAnnotation an object representing the collaboration annotation.
 */
-(void)remoteAnnotationAdded:(PTCollaborationAnnotation*)collaborationAnnotation;

/**
 * Call this method to communicate a change that occured remotely that needs to be reflected on the device.
 *
 * @param collaborationAnnotations an array of objects representing the collaboration annotations.
 */
-(void)remoteAnnotationsAdded:(NSArray<PTCollaborationAnnotation *> *)collaborationAnnotations;

/**
 * Call this method to communicate a change that occured remotely that needs to be reflected on the device.
 *
 * @param collaborationAnnotation an object representing the collaboration annotation.
 */
-(void)remoteAnnotationModified:(PTCollaborationAnnotation*)collaborationAnnotation;

/**
 * Call this method to communicate a change that occured remotely that needs to be reflected on the device.
 *
 * @param collaborationAnnotations an array of objects representing the collaboration annotations.
 */
-(void)remoteAnnotationsModified:(NSArray<PTCollaborationAnnotation *> *)collaborationAnnotations;

/**
 * Call this method to communicate a change that occured remotely that needs to be reflected on the device.
 *
 * @param collaborationAnnotation an object representing the collaboration annotation.
 */
-(void)remoteAnnotationRemoved:(PTCollaborationAnnotation*)collaborationAnnotation;

/**
 * Call this method to communicate a change that occured remotely that needs to be reflected on the device.
 *
 * @param collaborationAnnotations an array of objects representing the collaboration annotations.
 */
-(void)remoteAnnotationsRemoved:(NSArray<PTCollaborationAnnotation *> *)collaborationAnnotations;

/**
 * Establishes a "connection" between the `PTCollaborationManager`, that sends local changes to and receives remote
 * changes from the communicationComponent.
 *
 * param communicationComponent The object that the `PTCollaborationManager` uses to inform the server that a local
 * change occured, and likewise is used by the `communicationComponent` to inform the device that a remote change
 * occured.
 *
 */
-(void)registerServerCommunicationComponent:(id<PTCollaborationServerCommunication>)communicationComponent;

/**
 * Call to demark the end of one or more changes to a document for grouping when querying for changes to communicate
 * to the server.
 *
 * @param label A label to associate with the group of changes.
 *
 */
-(void)takeSnapshot:(NSString*)label;



-(instancetype)init NS_UNAVAILABLE;

@end

#pragma mark - Notifications

/**
 * Posted when remote annotations are added.
 *
 * The notification object is the collaboration manager that posted the notification. The `userInfo` dictionary
 * contains entries for the `PTCollaborationManagerAnnotationUserInfoKey` key.
 */
PT_EXPORT const NSNotificationName PTCollaborationManagerRemoteAnnotationsAddedNotification;

/**
 * Posted when remote annotations are modified.
 *
 * The notification object is the collaboration manager that posted the notification. The `userInfo` dictionary
 * contains entries for the `PTCollaborationManagerAnnotationUserInfoKey` key.
 */
PT_EXPORT const NSNotificationName PTCollaborationManagerRemoteAnnotationsModifiedNotification;

/**
 * Posted when remote annotations are removed.
 *
 * The notification object is the collaboration manager that posted the notification. The `userInfo` dictionary
 * contains entries for the `PTCollaborationManagerAnnotationUserInfoKey` key.
 */
PT_EXPORT const NSNotificationName PTCollaborationManagerRemoteAnnotationsRemovedNotification;

#pragma mark - Notification user info keys

/**
 * The key for an `NSArray` of `PTCollaborationAnnotation` objects.
 */
PT_EXPORT NSString * const PTCollaborationManagerAnnotationsUserInfoKey;

/**
 * The key for an `NSNumber` object containing a `BOOL`.
 */
PT_EXPORT NSString * const PTCollaborationManagerIsInitialUserInfoKey;

NS_ASSUME_NONNULL_END
