//---------------------------------------------------------------------------------------
// Copyright (c) 2001-2020 by PDFTron Systems Inc. All Rights Reserved.
// Consult legal.txt regarding legal and license information.
//---------------------------------------------------------------------------------------

#import "ToolsDefines.h"

#import <Foundation/Foundation.h>

NS_ASSUME_NONNULL_BEGIN

PT_EXPORT NSString * const PTToolsSettingsCategoryKey;
PT_EXPORT NSString * const PTToolsSettingsSettingKey;
PT_EXPORT NSString * const PTToolsSettingsCategoryDefaultValueKey;
PT_EXPORT NSString * const PTToolsSettingsSettingNameKey;
PT_EXPORT NSString * const PTToolsSettingsCategoryDescriptionKey;
PT_EXPORT NSString * const PTToolsSettingsSettingKeyKey;
PT_EXPORT NSString * const PTToolsSettingsMinOSKey;
PT_EXPORT NSString * const PTToolsSettingsUnavailableKey;

/**
 *
 * A class that manages user prefrences.
 *
 * The settings in this class are use by the `PTDocumentViewController` to control its behaviour, except as noted in the documentation for each setting.
 *
 */
@interface PTToolsSettingsManager : NSObject

/**
 * The global instance of the shared tools settings manager.
 */
@property (nonatomic, class, readonly, strong) PTToolsSettingsManager *sharedManager;

/**
 *
 * The default settings. These are the values that are contained in the settings plist file, `PTToolsSettings.plist`, not necessarily the current values.
 *
 */
@property (nonatomic, retain, readonly) NSArray<NSDictionary<NSString *, id> *> *toolsDefaultSettings;

/**
 * Determines if freehand annotation tool is shown in the main toolbar.
 *
 * Default is true.
 */
@property (nonatomic, assign) BOOL showInkInMainToolbar;

/**
 * Determines if the text search tool is shown in the main toolbar.
 *
 * Default is true on iPad, false on iPhone.
 */
@property (nonatomic, assign) BOOL showTextSearchInMainToolbar;

/**
 * Determines if shape annotations are automatically selected after the user creates them.
 *
 * Default is true.
 */
@property (nonatomic, assign) BOOL selectAnnotationAfterCreation;

/**
 * Determines if tabs for documents are enabled in the viewer.
 *
 * Default is true.
 *
 * @note this setting is not used by the `PTTabbedDocumentViewController`, and should instead be
 * read and acted upon by the app itself.
 */
@property (nonatomic, assign) BOOL tabsEnabled;

/**
 * Determines if the toolbars should be hidden automatically after 5 seconds.
 *
 * Default is false.
 */
@property (nonatomic, assign) BOOL automaticallyHideToolbars;

/**
 * Determines if PDFTron should execute any JavaScript contained in a PDF.
 *
 * Default is true.
 */
@property (nonatomic, assign) BOOL javascriptEnabled;

/**
 * Determines if an Apple Pencil should immediately draw ink (as opposed to function like a finger touch).
 *
 * Default is true.
 */
@property (nonatomic, assign) BOOL applePencilDrawsInk;

/**
 * Determines if freehand (ink) annotations should be created using PenckilKit (iOS 13.1+), or a standard curved line.
 *
 * Default is true.
 */
@property (nonatomic, assign) BOOL freehandUsesPencilKit;

/**
 * Determines if the document slider should show mini thumbnails (or use a standard `UISlider`).
 *
 * Default is true.
 */
@property (nonatomic, assign) BOOL showThumbnailsInSlider;

/**
 * Determines if the device should be stopped from dimming when a document is displayed.
 * Default is false.
 *
 * @note this setting is not used by the `PTDocumentVIewController`, and should instead be read and acted upon by the app itself.
 */
@property (nonatomic, assign) BOOL stopScreenFromDimming;

/**
 * Returns the boolean value for a given setting key value.
 */
-(BOOL)boolForKey:(NSString*)key;

/**
 * Sets the boolean value for a given setting key value.
 */
-(void)setBool:(BOOL)value forKey:(NSString*)key;

@end

NS_ASSUME_NONNULL_END
