//---------------------------------------------------------------------------------------
// Copyright (c) 2001-2020 by PDFTron Systems Inc. All Rights Reserved.
// Consult legal.txt regarding legal and license information.
//---------------------------------------------------------------------------------------

#import "ToolsConfig.h"
#import "PTCollaborationManager.h"
#import "PTCollaborationManager+Private.h"

#import "PTAnnotationManager.h"
#import "PTErrors.h"
#import "PTToolsUtil.h"

#import "PTAnnot+PTAdditions.h"
#import "PTFDFDoc+PTAdditions.h"

#pragma mark - Notification names

// Notification name values are equal to their constant's name.
const NSNotificationName PTCollaborationManagerRemoteAnnotationsAddedNotification = PT_NS_STRINGIFY(PTCollaborationManagerRemoteAnnotationsAddedNotification);
const NSNotificationName PTCollaborationManagerRemoteAnnotationsModifiedNotification = PT_NS_STRINGIFY(PTCollaborationManagerRemoteAnnotationsModifiedNotification);
const NSNotificationName PTCollaborationManagerRemoteAnnotationsRemovedNotification = PT_NS_STRINGIFY(PTCollaborationManagerRemoteAnnotationsRemovedNotification);

#pragma mark - Notification user info keys

// User info key values are equal to their constant's name.
NSString * const PTCollaborationManagerAnnotationsUserInfoKey = PT_NS_STRINGIFY(PTCollaborationManagerAnnotationsUserInfoKey);

NSString * const PTCollaborationManagerIsInitialUserInfoKey = PT_NS_STRINGIFY(PTCollaborationManagerIsInitialUserInfoKey);

NS_ASSUME_NONNULL_BEGIN

@interface PTCollaborationManager ()

@property (strong, nonatomic) PTExternalAnnotManager *externalAnnotManager;

@property (nonatomic, strong, nullable) id<PTCollaborationServerCommunication> server;

@end

NS_ASSUME_NONNULL_END

@implementation PTCollaborationManager

- (instancetype)initWithToolManager:(PTToolManager *)toolManager userId:(NSString *)userId
{
    self = [super init];
    if (self) {
        _toolManager = toolManager;
        
        _externalAnnotManager = [toolManager.pdfViewCtrl EnableAnnotationManager:userId];
        toolManager.annotationAuthor = userId;
                
        // register with toolmanager annotation notifications
        [NSNotificationCenter.defaultCenter addObserver:self
                                               selector:@selector(localAnnotationModified:)
                                                   name:PTToolManagerAnnotationModifiedNotification
                                                 object:toolManager];
        
        [NSNotificationCenter.defaultCenter addObserver:self
                                               selector:@selector(localAnnotationAdded:)
                                                   name:PTToolManagerAnnotationAddedNotification
                                                 object:toolManager];
        
        [NSNotificationCenter.defaultCenter addObserver:self
                                               selector:@selector(localAnnotationRemoved:)
                                                   name:PTToolManagerAnnotationRemovedNotification
                                                 object:toolManager];
        
        _annotationManager = [[PTAnnotationManager alloc] initWithCollaborationManager:self];
    }
    return self;
}

//// bi-directional relationship. Collab manager owns the server, server has weak reference to collabManager
-(void)registerServerCommunicationComponent:(id<PTCollaborationServerCommunication>)server
{
    self.server = server;
    
    [self.server setCollaborationManager:self];
    
    self.annotationManager.documentIdentifier = server.documentID;
    self.annotationManager.userIdentifier = server.userID;
}

#pragma mark - XFDF handling

-(void)takeSnapshot:(NSString*)label
{
    [self.externalAnnotManager TakeSnapshot:label];
}

-(NSString*)GetLastXFDF
{
    return [self.externalAnnotManager GetLastXFDF];
}

-(void)MergeXFDF:(NSString*)inputXFDF
{
    BOOL shouldUnlock = NO;
    @try
    {
        [self.toolManager.pdfViewCtrl DocLock:YES];
        shouldUnlock = YES;
        
        [self.externalAnnotManager MergeXFDF:inputXFDF];
    }
    @catch (NSException *exception) {
        NSLog(@"Exception: %@: %@", exception.name, exception.reason);
    }
    @finally {
        if (shouldUnlock) {
            [self.toolManager.pdfViewCtrl DocUnlock];
        }
    }
}

// Get the unique ID for the given annotation.
- (nullable NSString *)identifierForAnnot:(PTAnnot *)annot error:(NSError * _Nullable *)error
{
    BOOL shouldUnlock = NO;
    @try {
        [self.toolManager.pdfViewCtrl DocLockRead];
        shouldUnlock = YES;
        
        if (![annot IsValid]) {
            if (error) {
                *error = [NSError errorWithDomain:PTErrorDomain code:0 userInfo:
                          @{
                            NSLocalizedDescriptionKey: @"Invalid annotation",
                            NSLocalizedFailureReasonErrorKey: @"The given annotation is not valid",
                            }];
            }
            return nil;
        }
        
        PTObj *uniqueIDObj = [annot GetUniqueID];
        if (![uniqueIDObj IsValid] || ![uniqueIDObj IsString]) {
            if (error) {
                *error = [NSError errorWithDomain:PTErrorDomain code:0 userInfo:
                          @{
                            NSLocalizedDescriptionKey: @"Invalid annotation unique ID",
                            NSLocalizedFailureReasonErrorKey: @"The given annotation does not have a valid unique ID",
                            }];
            }
            return nil;
        }
        
        return [uniqueIDObj GetAsPDFText];
    }
    @catch (NSException *exception) {
        if (error) {
            *error = [NSError errorWithDomain:PTErrorDomain code:0 userInfo:
                      @{
                        NSLocalizedDescriptionKey: @"Failed to get annotation unique ID",
                        NSLocalizedFailureReasonErrorKey: @"The annotation unique ID could not be determined",
                        NSUnderlyingErrorKey: exception.pt_error,
                        }];
        }
        return nil;
    }
    @finally {
        if (shouldUnlock) {
            [self.toolManager.pdfViewCtrl DocUnlockRead];
        }
    }
    
    return nil;
}

- (nullable NSString *)authorForAnnot:(PTAnnot *)annot
{
    BOOL shouldUnlock = NO;
    @try {
        [self.toolManager.pdfViewCtrl DocLockRead];
        shouldUnlock = YES;
        
        if (![annot IsValid]) {
            return nil;
        }
        else if (![annot IsMarkup]) {
            return nil;
        }
        
        PTMarkup *markup = [[PTMarkup alloc] initWithAnn:annot];
        return [markup GetTitle];
    }
    @catch (NSException *exception) {
        return nil;
    }
    @finally {
        if (shouldUnlock) {
            [self.toolManager.pdfViewCtrl DocUnlockRead];
        }
    }
    
    return nil;
}


#pragma mark - Tool manager notifications for local changes

- (void)localAnnotationAdded:(NSNotification *)notification
{
    // Check notification object.
    if (notification.object != self.toolManager) {
        return;
    }
    
    PTAnnot *annotation = notification.userInfo[PTToolManagerAnnotationUserInfoKey];
    
    // Get the annotation's ID.
    NSError *error = nil;
    NSString *annotationID = [self identifierForAnnot:annotation error:&error];
    if (annotationID.length == 0) {
        if (error) {
            NSLog(@"Failed to get ID of added local annotation: %@", error);
        } else {
            NSLog(@"Added local annotation missing a unique ID");
        }
        return;
    }
    
    // Get last XFDF.
    NSString *xfdf = [self GetLastXFDF];
    if (xfdf.length == 0) {
        NSLog(@"Failed to get last XFDF for added local annotation");
        return;
    }
    
    NSString *annotationAuthor = [self authorForAnnot:annotation];
    
    PTCollaborationAnnotation* collaborationAnnotation = [[PTCollaborationAnnotation alloc] init];
    collaborationAnnotation.annotationID = annotationID;
    collaborationAnnotation.userID = annotationAuthor;
    collaborationAnnotation.xfdf = xfdf;
    
    [self.server localAnnotationAdded:collaborationAnnotation];
    
    // Optimistic update.
    [self remoteAnnotationAdded:collaborationAnnotation];
}

- (void)localAnnotationModified:(NSNotification *)notification
{
    // Check notification object.
    if (notification.object != self.toolManager) {
        return;
    }

    PTAnnot* annotation = notification.userInfo[PTToolManagerAnnotationUserInfoKey];
    
    // Get the annotation's ID.
    NSError *error = nil;
    NSString *annotationID = [self identifierForAnnot:annotation error:&error];
    if (annotationID.length == 0) {
        if (error) {
            NSLog(@"Failed to get ID of modified local annotation: %@", error);
        } else {
            NSLog(@"Modified local annotation missing a unique ID");
        }
        return;
    }
    
    // Get last XFDF.
    NSString *xfdf = [self GetLastXFDF];
    if (xfdf.length == 0) {
        NSLog(@"Failed to get last XFDF for modified local annotation");
        return;
    }

    PTCollaborationAnnotation* collaborationAnnotation = [[PTCollaborationAnnotation alloc] init];
    collaborationAnnotation.annotationID = annotationID;
    collaborationAnnotation.xfdf = xfdf;
    
    [self.server localAnnotationModified:collaborationAnnotation];
    
    // Optimistic update.
    [self remoteAnnotationModified:collaborationAnnotation];
}

- (void)localAnnotationRemoved:(NSNotification *)notification
{
    // Check notification object.
    if (notification.object != self.toolManager) {
        return;
    }
    
    PTAnnot *annotation = notification.userInfo[PTToolManagerAnnotationUserInfoKey];
    
    // Get the annotation's ID.
    NSError *error = nil;
    NSString *annotationID = [self identifierForAnnot:annotation error:&error];
    if (annotationID.length == 0) {
        if (error) {
            NSLog(@"Failed to get ID of removed local annotation: %@", error);
        } else {
            NSLog(@"Removed local annotation missing a unique ID");
        }
        return;
    }
    
    NSString *xfdf = [self GetLastXFDF];
    if (xfdf.length == 0) {
        NSLog(@"Failed to get last XFDF for modified local annotation");
        return;
    }
    
    PTCollaborationAnnotation* collaborationAnnotation = [[PTCollaborationAnnotation alloc] init];
    collaborationAnnotation.annotationID = annotationID;
    collaborationAnnotation.xfdf = xfdf;
    
    [self.server localAnnotationRemoved:collaborationAnnotation];
    
    // Optimistic update.
    [self remoteAnnotationRemoved:collaborationAnnotation];
}

#pragma mark - Initial annotations

- (void)loadInitialRemoteAnnotation:(PTCollaborationAnnotation *)collaborationAnnotation
{
    [self loadInitialRemoteAnnotations:@[collaborationAnnotation]];
}

- (void)loadInitialRemoteAnnotations:(NSArray<PTCollaborationAnnotation *> *)collaborationAnnotations
{
    for (PTCollaborationAnnotation *collaborationAnnotation in collaborationAnnotations) {
        // Check if annotation has the required fields.
        if (![collaborationAnnotation isValidForAdd]) {
            NSLog(@"Remote annotation is not valid for an add action");
        }
        
        [self MergeXFDF:collaborationAnnotation.xfdf];
    }
    
    // Post notification.
    [NSNotificationCenter.defaultCenter postNotificationName:PTCollaborationManagerRemoteAnnotationsAddedNotification
                                                      object:self
                                                    userInfo:@{
                                                        PTCollaborationManagerAnnotationsUserInfoKey: collaborationAnnotations,
                                                        PTCollaborationManagerIsInitialUserInfoKey: @YES,
                                                    }];
}

#pragma mark - Importing XFDF

- (void)importAnnotationsWithXFDFString:(NSString *)xfdfString
{
    [self importAnnotationsWithXFDFString:xfdfString isInitial:YES];
}

- (void)importAnnotationsWithXFDFString:(NSString *)xfdfString isInitial:(BOOL)isInitial
{
    @try {
        PTFDFDoc *fdfDoc = [PTFDFDoc CreateFromXFDF:xfdfString];
        
        // Merge XFDF into PDFViewCtrl.
        [self MergeXFDF:xfdfString];
        
        [self PT_importAnnotationsWithFDFDoc:fdfDoc isInitial:isInitial];
        
        [fdfDoc Close];
    }
    @catch (NSException *exception) {
        NSLog(@"Exception: %@, %@", exception.name, exception.reason);
    }
}

- (void)importAnnotationsWithXFDFCommand:(NSString *)xfdfCommand
{
    [self importAnnotationsWithXFDFCommand:xfdfCommand isInitial:YES];
}

- (void)importAnnotationsWithXFDFCommand:(NSString *)xfdfCommand isInitial:(BOOL)isInitial
{
    @try {
        PTFDFDoc *fdfDoc = [[PTFDFDoc alloc] init];
        [fdfDoc MergeAnnots:xfdfCommand permitted_user:@""];
        
        // Merge XFDF into PDFViewCtrl.
        [self MergeXFDF:xfdfCommand];
        
        [self PT_importAnnotationsWithFDFDoc:fdfDoc isInitial:isInitial];
        
        [fdfDoc Close];
    }
    @catch (NSException *exception) {
        NSLog(@"Exception: %@, %@", exception.name, exception.reason);
    }
}

- (void)PT_importAnnotationsWithFDFDoc:(PTFDFDoc *)fdfDoc isInitial:(BOOL)isInitial
{
    NSMutableArray<PTCollaborationAnnotation *> *mutableAnnotations = [NSMutableArray array];
    
    @try {
        for (PTAnnot *fdfAnnot in fdfDoc.annots) {
            if (![fdfAnnot IsValid]) {
                PTLog(@"Skipping invalid FDF annot");
                continue;
            }
            else if ([fdfAnnot GetType] == e_ptPopup) {
                PTLog(@"Skipping Popup FDF annot");
                continue;
            }
            else if (![fdfAnnot IsMarkup]) {
                PTLog(@"Skipping non-markup FDF annot");
                continue;
            }
            
            PTMarkup *fdfMarkup = [[PTMarkup alloc] initWithAnn:fdfAnnot];
            NSAssert([fdfMarkup IsValid], @"FDF markup annotation must be valid");
            
            // Extract the required information from the FDF annot.
            NSString *annotationIdentifier = nil;
            NSString *annotationAuthorIdentifier = nil;
            NSString *xfdf = nil;
            @try {
                annotationIdentifier = fdfMarkup.uniqueID;
                annotationAuthorIdentifier = [fdfMarkup GetTitle];
                
                xfdf = [PTFDFDoc XFDFStringFromAnnot:fdfMarkup];
            }
            @catch (NSException *exception) {
                NSLog(@"Exception: %@, %@", exception.name, exception.reason);
                continue;
            }
            
            PTCollaborationAnnotation *collaborationAnnotation = [[PTCollaborationAnnotation alloc] init];
            
            collaborationAnnotation.annotationID = annotationIdentifier;
            collaborationAnnotation.userID = annotationAuthorIdentifier;
            collaborationAnnotation.xfdf = xfdf;
            
            [mutableAnnotations addObject:collaborationAnnotation];
        }
    }
    @catch (NSException *exception) {
        NSLog(@"Exception: %@, %@", exception.name, exception.reason);
    }
    
    NSArray<PTCollaborationAnnotation *> *annotations = [mutableAnnotations copy];
    
    if (isInitial) {
        [self loadInitialRemoteAnnotations:annotations];
    } else {
        [self remoteAnnotationsAdded:annotations];
    }
}

#pragma mark - Methods by which the manager can be informed of remote changes

- (void)remoteAnnotationAdded:(PTCollaborationAnnotation *)collaborationAnnotation
{
    [self remoteAnnotationsAdded:@[collaborationAnnotation]];
}

-(void)remoteAnnotationsAdded:(NSArray<PTCollaborationAnnotation *> *)collaborationAnnotations
{
    for (PTCollaborationAnnotation *collaborationAnnotation in collaborationAnnotations) {
        // Check if annotation has the required fields.
        if (![collaborationAnnotation isValidForAdd]) {
            NSLog(@"Remote annotation is not valid for an add action");
        }
        
        [self MergeXFDF:collaborationAnnotation.xfdf];
    }

    // Post notification.
    [NSNotificationCenter.defaultCenter postNotificationName:PTCollaborationManagerRemoteAnnotationsAddedNotification
                                                      object:self
                                                    userInfo:@{
                                                        PTCollaborationManagerAnnotationsUserInfoKey: collaborationAnnotations,
                                                    }];
}

- (void)remoteAnnotationModified:(PTCollaborationAnnotation *)collaborationAnnotation
{
    [self remoteAnnotationsModified:@[collaborationAnnotation]];
}

- (void)remoteAnnotationsModified:(NSArray<PTCollaborationAnnotation *> *)collaborationAnnotations
{
    for (PTCollaborationAnnotation *collaborationAnnotation in collaborationAnnotations) {
        // Check if annotation has the required fields.
        if (![collaborationAnnotation isValidForModify]) {
            NSLog(@"Remote annotation is not valid for a modify action");
        }
        
        [self MergeXFDF:collaborationAnnotation.xfdf];
    }
    
    // Post notification.
    [NSNotificationCenter.defaultCenter postNotificationName:PTCollaborationManagerRemoteAnnotationsModifiedNotification
                                                      object:self
                                                    userInfo:@{
                                                               PTCollaborationManagerAnnotationsUserInfoKey: collaborationAnnotations,
                                                               }];
}

- (void)remoteAnnotationRemoved:(PTCollaborationAnnotation *)collaborationAnnotation
{
    [self remoteAnnotationsRemoved:@[collaborationAnnotation]];
}

- (void)remoteAnnotationsRemoved:(NSArray<PTCollaborationAnnotation *> *)collaborationAnnotations
{
    for (PTCollaborationAnnotation *collaborationAnnotation in collaborationAnnotations) {
        // Check if annotation has the required fields.
        if (![collaborationAnnotation isValidForRemove]) {
            NSLog(@"Remote annotation is not valid for a remove action");
        }
        
        NSString *deleteXfdf = [NSString stringWithFormat:@"<delete><id>%@</id></delete>",
                                collaborationAnnotation.annotationID];
        
        [self MergeXFDF:deleteXfdf];
    }
    
    // Post notification.
    [NSNotificationCenter.defaultCenter postNotificationName:PTCollaborationManagerRemoteAnnotationsRemovedNotification
                                                      object:self
                                                    userInfo:@{
                                                               PTCollaborationManagerAnnotationsUserInfoKey: collaborationAnnotations,
                                                               }];
}

@end
