//---------------------------------------------------------------------------------------
// Copyright (c) 2001-2020 by PDFTron Systems Inc. All Rights Reserved.
// Consult legal.txt regarding legal and license information.
//---------------------------------------------------------------------------------------

#import "ToolsConfig.h"
#import "PTSelectableBarButtonItem.h"

#import "NSLayoutConstraint+PTPriority.h"
#import "UIGeometry+PTAdditions.h"

NS_ASSUME_NONNULL_BEGIN

@interface PTSelectableBarButtonView : UIView

@property (nonatomic, strong) UIButton *button;

@property (nonatomic, strong) PTBadgeIndicatorView *badgeIndicatorView;

@property (nonatomic, assign) BOOL constraintsLoaded;

- (instancetype)initWithCoder:(NSCoder *)coder NS_UNAVAILABLE;

@end

NS_ASSUME_NONNULL_END

@implementation PTSelectableBarButtonView

- (void)PTSelectableBarButtonView_commonInit
{
    _button = [UIButton buttonWithType:UIButtonTypeSystem];
    
    // Round button corners and clip to round-bounds.
    // NOTE: The UIButton's imageView draws its background outside the button's bounds, so we
    // need to clip the button's content to control the appearance.
    _button.layer.cornerRadius = 4.0;
    _button.layer.masksToBounds = YES;
    
    _button.contentEdgeInsets = PTUIEdgeInsetsMakeUniform(_button.layer.cornerRadius);
    
    // Increase the button's horizontal content compression resistance priority to counteract
    // the "widthMinimizingConstraint" on the view.
    // (ie. try to keep the view as small as possible, but respect the button's width)
    [_button setContentCompressionResistancePriority:(UILayoutPriorityDefaultHigh + 10)
                                             forAxis:UILayoutConstraintAxisHorizontal];

    [self addSubview:_button];
    
    _badgeIndicatorView = [[PTBadgeIndicatorView alloc] init];
    [self addSubview:_badgeIndicatorView];
    _badgeIndicatorView.hidden = YES;
    
    self.layoutMargins = UIEdgeInsetsMake(0, 10, 0, 10);
    if (@available(iOS 11.0, *)) {
        self.insetsLayoutMarginsFromSafeArea = NO;
    }
            
    // Schedule constaints load.
    [self setNeedsUpdateConstraints];
}

- (instancetype)initWithFrame:(CGRect)frame
{
    self = [super initWithFrame:frame];
    if (self) {
        [self PTSelectableBarButtonView_commonInit];
    }
    return self;
}

#pragma mark - Constraints

- (void)loadConstraints
{
    self.button.translatesAutoresizingMaskIntoConstraints = NO;
    self.badgeIndicatorView.translatesAutoresizingMaskIntoConstraints = NO;
 
    // Constraint setup taken from _UIButtonBarButton & _UIModernBarButton.
    UILayoutGuide *layoutMarginsGuide = self.layoutMarginsGuide;
    
    [NSLayoutConstraint activateConstraints:@[
        [self.button.topAnchor constraintGreaterThanOrEqualToAnchor:self.topAnchor],
        [self.button.bottomAnchor constraintLessThanOrEqualToAnchor:self.bottomAnchor],
        
        [self.button.leadingAnchor constraintGreaterThanOrEqualToAnchor:layoutMarginsGuide.leadingAnchor],
        [self.button.trailingAnchor constraintLessThanOrEqualToAnchor:layoutMarginsGuide.trailingAnchor],
    
        [self.badgeIndicatorView.centerXAnchor constraintEqualToAnchor:self.button.imageView.trailingAnchor],
        [self.badgeIndicatorView.centerYAnchor constraintEqualToAnchor:self.button.imageView.topAnchor],
    ]];
    
    [NSLayoutConstraint pt_activateConstraints:@[
        // Center button in view.
        [self.button.centerXAnchor constraintEqualToAnchor:self.centerXAnchor],
        [self.button.centerYAnchor constraintEqualToAnchor:self.centerYAnchor],
        
        // Match widths, with spacing on both sides.
        [self.button.widthAnchor constraintEqualToAnchor:layoutMarginsGuide.widthAnchor],
    ] withPriority:(UILayoutPriorityRequired - 1) /* 999 */];
    
    // Horizontal constraint to keep view width as small as possible.
    NSLayoutConstraint *widthMinimizingConstraint =
    [self.widthAnchor constraintEqualToConstant:0];
    widthMinimizingConstraint.identifier = @"widthMinimizingConstraint";

    [NSLayoutConstraint pt_activateConstraints:@[
        widthMinimizingConstraint,
    ] withPriority:UILayoutPriorityDefaultLow];
}

- (void)updateConstraints
{
    if (!self.constraintsLoaded) {
        [self loadConstraints];
        
        self.constraintsLoaded = YES;
    }
    
    // Call super implementation as final step.
    [super updateConstraints];
}

+ (BOOL)requiresConstraintBasedLayout
{
    return YES;
}

#pragma mark - Layout

- (void)layoutSubviews
{
    [super layoutSubviews];
    
    [self updateLayoutMargins];
}

- (void)updateLayoutMargins
{
    // Check if the leading or trailing edge of the view is touching the superview.
    // When touching the superview, that edge does not get a "spacer".
    UIEdgeInsets layoutMargins = self.layoutMargins;

    if (CGRectGetMinX(self.frame) == CGRectGetMinX(self.superview.bounds)) {
        // Touching leading end.
        layoutMargins.left = 0;
    }
    else {
        // Not touching leading end.
        layoutMargins.left = 10;
    }
    
    if (CGRectGetMaxX(self.frame) == CGRectGetMaxX(self.superview.bounds)) {
        // Touching trailing end.
        layoutMargins.right = 0;
    }
    else {
        // Not touching trailing end.
        layoutMargins.right = 10;
    }
    
    if (!UIEdgeInsetsEqualToEdgeInsets(self.layoutMargins, layoutMargins)) {
        self.layoutMargins = layoutMargins;
    }
}

#pragma mark - View hierarchy

- (void)didMoveToSuperview
{
    [super didMoveToSuperview];
    
    if (self.superview) {
        // Match the height of the superview's layoutMarginsGuide.
        // NOTE: This effectively makes the view 44pts high.
        [NSLayoutConstraint activateConstraints:@[
           [self.heightAnchor constraintEqualToAnchor:self.superview.layoutMarginsGuide.heightAnchor],
        ]];
    }
}

#pragma mark - Hit testing

- (UIView *)hitTest:(CGPoint)point withEvent:(UIEvent *)event
{
    UIView *hitTestView = [super hitTest:point withEvent:event];
    if (hitTestView == self) {
        hitTestView = self.button;
    }
    return hitTestView;
}

@end

typedef NS_ENUM(NSUInteger, PTSelectableBarButtonType) {
    PTSelectableBarButtonTypeImage,
    PTSelectableBarButtonTypeTitle,
};

NS_ASSUME_NONNULL_BEGIN

@interface PTSelectableBarButtonItem ()

@property (nonatomic, strong, nullable) PTSelectableBarButtonView *view;

@property (nonatomic, assign) PTSelectableBarButtonType buttonType;

@end

NS_ASSUME_NONNULL_END

@implementation PTSelectableBarButtonItem

- (void)PTSelectableBarButtonItem_commonInit
{
    _view = [[PTSelectableBarButtonView alloc] init];
    _view.translatesAutoresizingMaskIntoConstraints = NO;
    
    [_view.button addTarget:self
                     action:@selector(buttonTriggered:)
           forControlEvents:UIControlEventPrimaryActionTriggered];
        
    self.customView = _view;
}

- (instancetype)initWithImage:(UIImage *)image style:(UIBarButtonItemStyle)style target:(id)target action:(SEL)action
{
    self = [super initWithImage:image style:style target:target action:action];
    if (self) {
        [self PTSelectableBarButtonItem_commonInit];
        
        [_view.button setImage:image forState:UIControlStateNormal];
        
        _buttonType = PTSelectableBarButtonTypeImage;
    }
    return self;
}

- (instancetype)initWithTitle:(NSString *)title style:(UIBarButtonItemStyle)style target:(id)target action:(SEL)action
{
    self = [super initWithTitle:title style:style target:target action:action];
    if (self) {
        [self PTSelectableBarButtonItem_commonInit];
        
        [_view.button setTitle:title forState:UIControlStateNormal];
        
        _buttonType = PTSelectableBarButtonTypeTitle;
    }
    return self;
}

- (void)setTitle:(NSString *)title
{
    [super setTitle:title];
    
    if (self.buttonType == PTSelectableBarButtonTypeTitle) {
        [self.view.button setTitle:title forState:UIControlStateNormal];
    }
}

- (void)setImage:(UIImage *)image
{
    [super setImage:image];
    
    if (self.buttonType == PTSelectableBarButtonTypeImage) {
        [self.view.button setImage:image forState:UIControlStateNormal];
    }
}

#pragma mark - Selected

- (void)setSelected:(BOOL)selected
{
    self.view.button.selected = selected;
}

- (BOOL)isSelected
{
    return [self.view.button isSelected];
}

#pragma mark - Badge Hidden

- (BOOL)isBadgeHidden
{
    return [self.view.badgeIndicatorView isHidden];
}

- (void)setBadgeHidden:(BOOL)hidden
{
    self.view.badgeIndicatorView.hidden = hidden;
}

#pragma mark - badgeIndicatorView

- (PTBadgeIndicatorView *)badgeIndicatorView
{
    return self.view.badgeIndicatorView;
}

#pragma mark - Button action

- (void)buttonTriggered:(UIButton *)sender
{
    if (!self.action) {
        return;
    }
    
    [UIApplication.sharedApplication sendAction:self.action
                                             to:self.target
                                           from:self
                                       forEvent:nil];
}

@end
