//---------------------------------------------------------------------------------------
// Copyright (c) 2001-2020 by PDFTron Systems Inc. All Rights Reserved.
// Consult legal.txt regarding legal and license information.
//---------------------------------------------------------------------------------------

#import "ToolsConfig.h"
#import "PTEditToolbar.h"

#import "PTToolsUtil.h"

NS_ASSUME_NONNULL_BEGIN

@interface PTEditToolbar ()

@property (nonatomic, strong) UIBarButtonItem *cancelButton;
@property (nonatomic, strong) UIBarButtonItem *saveButton;
@property (nonatomic, strong) UIBarButtonItem *undoButton;
@property (nonatomic, strong) UIBarButtonItem *redoButton;
@property (nonatomic, strong) UIBarButtonItem *styleButton;

@end

NS_ASSUME_NONNULL_END

@implementation PTEditToolbar

// Synthesized by UIToolbar superclass.
@dynamic delegate;

- (instancetype)initWithFrame:(CGRect)frame
{
    self = [super initWithFrame:frame];
    if (self) {
        _cancelButton = [[UIBarButtonItem alloc] initWithBarButtonSystemItem:UIBarButtonSystemItemCancel
                                                                      target:self
                                                                      action:@selector(cancelEditing:)];
        
        _saveButton = [[UIBarButtonItem alloc] initWithBarButtonSystemItem:UIBarButtonSystemItemDone
                                                                    target:self
                                                                    action:@selector(commitEdits:)];
        
        UIImage* undoImage;
        UIImage* redoImage;

        if (@available(iOS 13.0, *)) {
            undoImage = [UIImage systemImageNamed:@"arrow.uturn.left.circle" withConfiguration:[UIImageSymbolConfiguration configurationWithScale:UIImageSymbolScaleUnspecified]];
            redoImage = [UIImage systemImageNamed:@"arrow.uturn.right.circle" withConfiguration:[UIImageSymbolConfiguration configurationWithScale:UIImageSymbolScaleUnspecified]];
        }
        else {
            undoImage = [PTToolsUtil toolImageNamed:@"glyphicons_435_undo"];
            redoImage = [PTToolsUtil toolImageNamed:@"glyphicons_434_redo"];
        }
        
        _undoButton = [[UIBarButtonItem alloc] initWithImage:undoImage
                                                       style:UIBarButtonItemStylePlain
                                                      target:self
                                                      action:@selector(undoEdit:)];
        
        _redoButton = [[UIBarButtonItem alloc] initWithImage:redoImage
                                                       style:UIBarButtonItemStylePlain
                                                      target:self
                                                      action:@selector(redoEdit:)];
        
        _styleButton = [[UIBarButtonItem alloc] initWithImage:[PTToolsUtil toolImageNamed:@"ic_annotation_freehand_black_24dp"]
                                                       style:UIBarButtonItemStylePlain
                                                      target:self
                                                      action:@selector(toggleStylePicker:)];
        _undoButton.enabled = NO;
        _redoButton.enabled = NO;
        
        _styleButtonHidden = NO;

        [self updateItems];
    }
    return self;
}

- (void)updateItems
{
    NSMutableArray<UIBarButtonItem *> *items = [NSMutableArray array];
    
    [items addObject:self.cancelButton];
    
    UIBarButtonItem* fixedSpaceToSaveButton = [[UIBarButtonItem alloc] initWithBarButtonSystemItem:UIBarButtonSystemItemFixedSpace target:nil action:nil];
    
    if ([UIDevice currentDevice].userInterfaceIdiom == UIUserInterfaceIdiomPhone) {
        fixedSpaceToSaveButton.width = 15;
    } else {
        fixedSpaceToSaveButton.width = 30;
    }
    
    if (!self.styleButtonHidden) {
            [items addObjectsFromArray:
         @[
           fixedSpaceToSaveButton,
           _styleButton
         ]];
    }

    [items addObjectsFromArray:
         @[
       [[UIBarButtonItem alloc] initWithBarButtonSystemItem:UIBarButtonSystemItemFlexibleSpace target:nil action:nil]
       ]];
    
    if (!self.undoRedoHidden) {
        
        UIBarButtonItem* fixedSpaceBetweenUndoRedo = [[UIBarButtonItem alloc] initWithBarButtonSystemItem:UIBarButtonSystemItemFixedSpace target:nil action:nil];
        fixedSpaceBetweenUndoRedo.width = 15;

        [items addObjectsFromArray:
         @[
           _undoButton,
           fixedSpaceBetweenUndoRedo,
           _redoButton,
           fixedSpaceToSaveButton,
           ]];
    }
    
    [items addObject:self.saveButton];
    
    self.items = items;
}

#pragma mark - Button actions

- (void)cancelEditing:(UIBarButtonItem *)button
{
    if ([self.delegate respondsToSelector:@selector(editToolbarDidCancel:)]) {
        [self.delegate editToolbarDidCancel:self];
    }
}

- (void)commitEdits:(UIBarButtonItem *)button
{
    if ([self.delegate respondsToSelector:@selector(editToolbarDidCommit:)]) {
        [self.delegate editToolbarDidCommit:self];
    }
}

- (void)undoEdit:(UIBarButtonItem *)button
{
    if ([self.delegate respondsToSelector:@selector(editToolbarUndoChange:)]) {
        [self.delegate editToolbarUndoChange:self];
    }
}

- (void)redoEdit:(UIBarButtonItem *)button
{
    if ([self.delegate respondsToSelector:@selector(editToolbarRedoChange:)]) {
        [self.delegate editToolbarRedoChange:self];
    }
}

- (void)toggleStylePicker:(UIBarButtonItem *)button
{
    if ([self.delegate respondsToSelector:@selector(editToolbarToggleStylePicker:)]) {
        [self.delegate editToolbarToggleStylePicker:button];
    }
}

- (void)setUndoRedoHidden:(BOOL)hidden
{
    _undoRedoHidden = hidden;
    
    [self updateItems];
}

- (void)setStyleButtonHidden:(BOOL)hidden
{
    _styleButtonHidden = hidden;

    [self updateItems];
}

#pragma mark - Undo/redo enabled

- (void)setUndoEnabled:(BOOL)enabled
{
    self.undoButton.enabled = enabled;
}

- (BOOL)isUndoEnabled
{
    return [self.undoButton isEnabled];
}

- (void)setRedoEnabled:(BOOL)enabled
{
    self.redoButton.enabled = enabled;
}

- (BOOL)isRedoEnabled
{
    return [self.redoButton isEnabled];
}

@end
