//---------------------------------------------------------------------------------------
// Copyright (c) 2001-2020 by PDFTron Systems Inc. All Rights Reserved.
// Consult legal.txt regarding legal and license information.
//---------------------------------------------------------------------------------------

#import "ToolsConfig.h"
#import "PTPageIndicatorViewController.h"

#import "ToolsDefines.h"
#import "PTToolsUtil.h"

@interface PTPageIndicatorViewController ()

@property (nonatomic, strong) PTPDFViewCtrl *pdfViewCtrl;

@property (nonatomic, strong) UIView *contentView;
@property (nonatomic, strong) UILabel *label;

@property (nonatomic, strong) UITapGestureRecognizer *tapGestureRecognizer;

@property (nonatomic, copy, nullable) NSString *pageLabel;
@property (nonatomic, assign) int currentPage;
@property (nonatomic, readonly, assign) int pageCount;

@property (nonatomic, readonly, strong) UIView *defaultBackgroundView;

@end

@implementation PTPageIndicatorViewController

- (instancetype)initWithPDFViewCtrl:(PTPDFViewCtrl *)pdfViewCtrl
{
    self = [super initWithNibName:nil bundle:nil];
    if (self) {
        _pdfViewCtrl = pdfViewCtrl;
    }
    return self;
}

#pragma mark - View lifecycle

- (void)loadView
{
    self.view = [[UIView alloc] init];
    
    // Background view.
    self.backgroundView = self.defaultBackgroundView;
    self.backgroundView.frame = self.view.bounds;
    self.backgroundView.autoresizingMask = UIViewAutoresizingFlexibleWidth | UIViewAutoresizingFlexibleHeight;
    
    [self.view insertSubview:self.backgroundView atIndex:0];
    
    // Content view.
    self.contentView = [[UIView alloc] initWithFrame:self.view.bounds];
    self.contentView.autoresizingMask = UIViewAutoresizingFlexibleWidth | UIViewAutoresizingFlexibleHeight;
    self.contentView.layoutMargins = UIEdgeInsetsMake(4, 8, 4, 8);
    
    if (@available(iOS 11.0, *)) {
        self.contentView.insetsLayoutMarginsFromSafeArea = NO;
    }
    
    [self.view addSubview:self.contentView];
    
    // Label view.
    self.label = [[UILabel alloc] initWithFrame:self.view.bounds];
    self.label.font = [UIFont systemFontOfSize:UIFont.smallSystemFontSize];
    
    [self.contentView addSubview:self.label];

    self.label.translatesAutoresizingMaskIntoConstraints = NO;
    
    UILayoutGuide *marginsGuide = self.contentView.layoutMarginsGuide;
    
    [NSLayoutConstraint activateConstraints:
     @[
       [self.label.leadingAnchor constraintEqualToAnchor:marginsGuide.leadingAnchor],
       [self.label.topAnchor constraintEqualToAnchor:marginsGuide.topAnchor],
       [self.label.widthAnchor constraintEqualToAnchor:marginsGuide.widthAnchor],
       [self.label.heightAnchor constraintEqualToAnchor:marginsGuide.heightAnchor],
       ]];
    
    self.tapGestureRecognizer = [[UITapGestureRecognizer alloc] initWithTarget:self action:@selector(tapActionWithGesture:)];
    [self.view addGestureRecognizer:self.tapGestureRecognizer];
}

- (void)viewDidLoad
{
    [super viewDidLoad];
    
    self.currentPage = [self.pdfViewCtrl GetCurrentPage];

    [self updateLabel];
}

- (void)viewWillAppear:(BOOL)animated
{
    [super viewWillAppear:animated];
    
    self.currentPage = [self.pdfViewCtrl GetCurrentPage];
    
    [self updateLabel];
    
    // Start observing PDFViewCtrl notifications.
    [NSNotificationCenter.defaultCenter addObserver:self
                                           selector:@selector(pdfViewCtrlPageDidChangeNotification:)
                                               name:PTPDFViewCtrlPageDidChangeNotification
                                             object:self.pdfViewCtrl];
    
    [NSNotificationCenter.defaultCenter addObserver:self
                                           selector:@selector(pdfViewCtrlPageCountDidChangeNotification:)
                                               name:PTPDFViewCtrlStreamingEventNotification
                                             object:self.pdfViewCtrl];
}

- (void)viewDidDisappear:(BOOL)animated
{
    [super viewDidDisappear:animated];
    
    // Stop observing PDFViewCtrl notifications.
    [NSNotificationCenter.defaultCenter removeObserver:self
                                                  name:PTPDFViewCtrlPageDidChangeNotification
                                                object:self.pdfViewCtrl];
    
    [NSNotificationCenter.defaultCenter removeObserver:self
                                                  name:PTPDFViewCtrlStreamingEventNotification
                                                object:self.pdfViewCtrl];
}

- (int)pageCount
{
    return [self.pdfViewCtrl GetPageCount];
}

#pragma mark - Background

@synthesize backgroundView = _backgroundView;

- (UIView *)backgroundView
{
    [self loadViewIfNeeded];
    
    return _backgroundView;
}

- (void)setBackgroundView:(UIView *)backgroundView
{
    [self loadViewIfNeeded];

    // Remove old background view.
    if (_backgroundView) {
        [_backgroundView removeFromSuperview];
    }
    
    if (backgroundView) {
        _backgroundView = backgroundView;
    } else {
        _backgroundView = self.defaultBackgroundView;
    }
    
    [self.view insertSubview:_backgroundView atIndex:0];
}

- (UIView *)defaultBackgroundView
{
    UIBlurEffectStyle blurEffectStyle = UIBlurEffectStyleExtraLight;
    if (@available(iOS 13.0, *)) {
        blurEffectStyle = UIBlurEffectStyleSystemMaterial;
    }
    UIView *view = [[UIVisualEffectView alloc] initWithEffect:[UIBlurEffect effectWithStyle:blurEffectStyle]];
    view.layer.cornerRadius = 4.0;
    view.layer.masksToBounds = YES; // Required for UIVisualEffectView cornerRadius.
    return view;
}

#pragma mark - Label

- (void)updateLabel
{
    NSString *shortFormat = PTLocalizedString(@"%d of %d", @"Current page out of page count");
    NSString *shortTitle = [NSString localizedStringWithFormat:shortFormat, self.currentPage, self.pageCount];

    NSString *fullTitle = shortTitle;
    
    if (self.pageLabel) {
        fullTitle = [NSString stringWithFormat:@"%@ (%@)", self.pageLabel, shortTitle];
    }
    
    self.label.text = fullTitle;
}

#pragma mark - Go to Page

- (void)presentGoToPageController
{
    NSString *alertTitle = PTLocalizedString(@"Go to Page", @"Go to Page alert title");
    
    NSString *alertMessageFormat = PTLocalizedString(@"Enter a page number (%d to %d)", @"Go to Page alert message");
    NSString *alertMessage = [NSString localizedStringWithFormat:alertMessageFormat, 1, self.pageCount];
    
    UIAlertController *alertController = [UIAlertController alertControllerWithTitle:alertTitle
                                                                             message:alertMessage
                                                                      preferredStyle:UIAlertControllerStyleAlert];
    
    // Page number text field.
    [alertController addTextFieldWithConfigurationHandler:^(UITextField *textField) {
        textField.keyboardType = UIKeyboardTypeNumberPad;
        textField.returnKeyType = UIReturnKeyGo;

        textField.placeholder = PTLocalizedString(@"Page number", @"Page number placeholder");
    }];
    
    // Cancel action.
    NSString *cancelTitle = PTLocalizedString(@"Cancel", @"Cancel alert label");
    UIAlertAction *cancelAction = [UIAlertAction actionWithTitle:cancelTitle style:UIAlertActionStyleCancel handler:nil];
    [alertController addAction:cancelAction];
    
    // OK action.
    NSString *okTitle = PTLocalizedString(@"OK", @"OK alert label");
    UIAlertAction *okAction = [UIAlertAction actionWithTitle:okTitle style:UIAlertActionStyleDefault handler:^(UIAlertAction *action) {
        // Attempt to jump to the page.
        UITextField *textField = alertController.textFields.firstObject;
        
        // Read int value from string.
        NSScanner *scanner = [NSScanner scannerWithString:textField.text];
        int pageNumber = 0;
        if ([scanner scanInt:&pageNumber] && (pageNumber > 0 && pageNumber <= self.pageCount)) {
            [self.pdfViewCtrl SetCurrentPage:pageNumber];
        } else {
            // Invalid page number.
            [self notifyInvalidPageNumber];
        }
    }];
    [alertController addAction:okAction];
    alertController.preferredAction = okAction;
    
    [self presentViewController:alertController animated:YES completion:nil];
}

- (void)notifyInvalidPageNumber
{
    NSString *alertTitle = PTLocalizedString(@"Invalid page number", @"Invalid page number alert title");
    UIAlertController *alertController = [UIAlertController alertControllerWithTitle:alertTitle
                                                                             message:nil
                                                                      preferredStyle:UIAlertControllerStyleAlert];
    
    NSString *okTitle = PTLocalizedString(@"OK", @"OK alert label");
    [alertController addAction:[UIAlertAction actionWithTitle:okTitle style:UIAlertActionStyleDefault handler:nil]];

    [self presentViewController:alertController animated:YES completion:nil];
}

#pragma mark - Gesture recognizer actions

- (void)tapActionWithGesture:(UITapGestureRecognizer *)tapGestureRecognizer
{
    if (self.pageCount < 2) {
        return;
    }
    
    [self presentGoToPageController];
}

#pragma mark - Notification actions

- (void)pdfViewCtrlPageDidChangeNotification:(NSNotification *)notification
{
    if (notification.object != self.pdfViewCtrl) {
        return;
    }
    
    int currentPageNumber = ((NSNumber *) notification.userInfo[PTPDFViewCtrlCurrentPageNumberUserInfoKey]).intValue;
    if (currentPageNumber == 0) {
        return;
    }
    
    self.currentPage = currentPageNumber;
    
    [self updateLabel];
}

- (void)pdfViewCtrlPageCountDidChangeNotification:(NSNotification *)notification
{
    [self updateLabel];
}

@end
