//---------------------------------------------------------------------------------------
// Copyright (c) 2001-2020 by PDFTron Systems Inc. All Rights Reserved.
// Consult legal.txt regarding legal and license information.
//---------------------------------------------------------------------------------------

#import "ToolsConfig.h"
#import "PTAnnotStyle.h"
#import "PTMeasurementUtil.h"

#import "PTColorDefaults.h"

static NSString * const kAnnotStylePDFTronThicknessKey = @"pdftron_thickness";

#pragma mark - AnnotStyleKey definitions

PTAnnotStyleKey const PTAnnotStyleKeyColor = @"AnnotStyleKeyColor";
PTAnnotStyleKey const PTAnnotStyleKeyStrokeColor = @"AnnotStyleKeyStrokeColor";
PTAnnotStyleKey const PTAnnotStyleKeyFillColor = @"AnnotStyleKeyFillColor";
PTAnnotStyleKey const PTAnnotStyleKeyTextColor = @"AnnotStyleKeyTextColor";

PTAnnotStyleKey const PTAnnotStyleKeyThickness = @"AnnotStyleKeyThickness";
PTAnnotStyleKey const PTAnnotStyleKeyOpacity = @"AnnotStyleKeyOpacity";
PTAnnotStyleKey const PTAnnotStyleKeyTextSize = @"AnnotStyleKeyTextSize";
PTAnnotStyleKey const PTAnnotStyleKeyScale = @"AnnotStyleKeyScale";
PTAnnotStyleKey const PTAnnotStyleKeyPrecision = @"AnnotStyleKeyPrecision";
PTAnnotStyleKey const PTAnnotStyleKeySnapping = @"AnnotStyleKeySnapping";

PTAnnotStyleKey const PTAnnotStyleKeyLabel = @"AnnotStyleKeyLabel";

@interface PTAnnotStyle ()

// Redeclare as read-write internally.
@property (nonatomic, copy, readwrite) NSArray<PTAnnotStyleKey> *availableStyleKeys;

// Annotation loading
- (void)configure;
- (void)configureFreeText;
- (void)configureFreeHand;
- (void)configureMarkup;
- (void)configureHighlight;
- (void)configureTextAnnot;
- (void)configureRedaction;
- (void)configureLengthMeasure;
- (void)configureAreaMeasure;

// Annotation saving
- (void)saveChanges;
- (void)saveFreeText;
- (void)saveFreeHand;
- (void)saveMarkup;
- (void)saveHighlight;
- (void)saveTextAnnot;
- (void)saveRedaction;

// Annotation property loading
- (void)loadTextColorWithFreeText:(PTFreeText *)freeText;
- (void)loadStrokeColor;
- (void)loadFillColorWithMarkup:(PTMarkup *)markup;
- (void)loadBorderColorWithFreeText:(PTFreeText *)freeText;
- (void)loadThickness;
- (void)loadOpacityWithMarkup:(PTMarkup *)markup;
- (void)loadTextSizeWithFreeText:(PTFreeText *)freeText;
- (void)loadMeasurementData;

// Annotation property saving
- (void)saveTextColorWithFreeText:(PTFreeText *)freeText;
- (void)saveStrokeColor;
- (void)saveFillColorWithMarkup:(PTMarkup *)markup;
- (void)saveBorderColorWithFreeText:(PTFreeText *)freeText;
- (void)saveThickness;
- (void)saveOpacityWithMarkup:(PTMarkup *)markup;
- (void)saveTextSizeWithFreeText:(PTFreeText *)freeText;
- (void)saveMeasurementData;

@end

@implementation PTAnnotStyle

// Synthesize annotType for when annot property is nil.
@synthesize annotType = _annotType;

- (instancetype)initWithAnnotType:(PTExtendedAnnotType)annotType
{
    self = [super init];
    if (self) {
        _annotType = annotType;

        _saveValuesAsDefaults = YES;

        [self configure];
    }
    return self;
}

- (instancetype)initWithAnnot:(PTAnnot *)annot
{
    self = [super init];
    if (self) {
        _annotType = [annot extendedAnnotType];
        _annot = annot;

        _saveValuesAsDefaults = YES;

        [self configure];
    }
    return self;
}

#pragma mark - Annotation loading

- (void)configure
{
    switch (self.annotType) {
        case PTExtendedAnnotTypeText:
            break;
        case PTExtendedAnnotTypeLink:
            break;
        case PTExtendedAnnotTypeFreeText:
        case PTExtendedAnnotTypeCallout:
            [self configureFreeText];
            break;
        case PTExtendedAnnotTypeLine:
        case PTExtendedAnnotTypeArrow:
            [self configureFreeHand];
            break;
        case PTExtendedAnnotTypeRuler:
        case PTExtendedAnnotTypePerimeter:
            [self configureLengthMeasure];
            break;
        case PTExtendedAnnotTypeArea:
            [self configureAreaMeasure];
            break;
        case PTExtendedAnnotTypeSquare:
            [self configureMarkup];
            break;
        case PTExtendedAnnotTypeCircle:
            [self configureMarkup];
            break;
        case PTExtendedAnnotTypePolygon:
        case PTExtendedAnnotTypeCloudy:
            [self configureMarkup];
            break;
        case PTExtendedAnnotTypePolyline:
            [self configureFreeHand];
            break;
        case PTExtendedAnnotTypeHighlight:
            [self configureHighlight];
            break;
        case PTExtendedAnnotTypeUnderline:
            [self configureTextAnnot];
            break;
        case PTExtendedAnnotTypeSquiggly:
            [self configureTextAnnot];
            break;
        case PTExtendedAnnotTypeStrikeOut:
            [self configureTextAnnot];
            break;
        case PTExtendedAnnotTypeStamp:
        case PTExtendedAnnotTypeImageStamp:
        case PTExtendedAnnotTypeSignature:
            break;
        case PTExtendedAnnotTypeCaret:
            break;
        case PTExtendedAnnotTypeInk:
        case PTExtendedAnnotTypeFreehandHighlight:
            [self configureFreeHand];
            break;
        case PTExtendedAnnotTypePopup:
            break;
        case PTExtendedAnnotTypeFileAttachment:
            break;
        case PTExtendedAnnotTypeSound:
            break;
        case PTExtendedAnnotTypeMovie:
            break;
        case PTExtendedAnnotTypeWidget:
            break;
        case PTExtendedAnnotTypeScreen:
            break;
        case PTExtendedAnnotTypePrinterMark:
            break;
        case PTExtendedAnnotTypeTrapNet:
            break;
        case PTExtendedAnnotTypeWatermark:
            break;
        case PTExtendedAnnotType3D:
            break;
        case PTExtendedAnnotTypeRedact:
            [self configureRedaction];
            break;
        case PTExtendedAnnotTypeProjection:
            break;
        case PTExtendedAnnotTypeRichMedia:
            break;
        case PTExtendedAnnotTypeUnknown:
            break;
        default:
            break;
    }
}

- (void)configureFreeText
{
    self.availableStyleKeys =
    @[
      PTAnnotStyleKeyTextColor,
      PTAnnotStyleKeyStrokeColor,
      PTAnnotStyleKeyThickness,
      PTAnnotStyleKeyTextSize,
      PTAnnotStyleKeyOpacity,
      ];

    if ([self.annot IsValid]) {
        PTFreeText *freeText = [[PTFreeText alloc] initWithAnn:self.annot];

        [self loadTextColorWithFreeText:freeText];
        [self loadBorderColorWithFreeText:freeText];
        [self loadThickness];
        [self loadTextSizeWithFreeText:freeText];
        [self loadOpacityWithMarkup:freeText];
    } else {
        // Load defaults for annotation type.
        [self loadDefaultTextColor];
        [self loadDefaultBorderColor];
        [self loadDefaultThickness];
        [self loadDefaultTextSize];
        [self loadDefaultOpacity];
    }
}

- (void)configureFreeHand
{
    self.availableStyleKeys =
    @[
      PTAnnotStyleKeyColor,
      PTAnnotStyleKeyThickness,
      PTAnnotStyleKeyOpacity,
      ];

    if ([self.annot IsValid]) {
        PTMarkup *markup = [[PTMarkup alloc] initWithAnn:self.annot];

        [self loadStrokeColor];
        [self loadThickness];
        [self loadOpacityWithMarkup:markup];
    } else {
        // Load defaults for annotation type.
        [self loadDefaultStrokeColor];
        [self loadDefaultThickness];
        [self loadDefaultOpacity];
    }
}

- (void)configureMarkup
{
    self.availableStyleKeys =
    @[
      PTAnnotStyleKeyStrokeColor,
      PTAnnotStyleKeyFillColor,
      PTAnnotStyleKeyThickness,
      PTAnnotStyleKeyOpacity,
      ];

    if ([self.annot IsValid]) {
        PTMarkup *markup = [[PTMarkup alloc] initWithAnn:self.annot];

        [self loadStrokeColor];
        [self loadFillColorWithMarkup:markup];
        [self loadThickness];
        [self loadOpacityWithMarkup:markup];
    } else {
        // Load defaults for annotation type.
        [self loadDefaultStrokeColor];
        [self loadDefaultFillColor];
        [self loadDefaultThickness];
        [self loadDefaultOpacity];
    }
}

- (void)configureHighlight
{
    self.availableStyleKeys =
    @[
      PTAnnotStyleKeyColor,
      PTAnnotStyleKeyOpacity,
      ];

    if ([self.annot IsValid]) {
        PTMarkup *markup = [[PTMarkup alloc] initWithAnn:self.annot];

        [self loadStrokeColor];
        [self loadOpacityWithMarkup:markup];
    } else {
        // Load defaults for annotation type.
        [self loadDefaultStrokeColor];
        [self loadDefaultOpacity];
    }
}

- (void)configureTextAnnot
{
    self.availableStyleKeys =
    @[
      PTAnnotStyleKeyColor,
      PTAnnotStyleKeyThickness,
      PTAnnotStyleKeyOpacity,
      ];

    if ([self.annot IsValid]) {
        PTMarkup *markup = [[PTMarkup alloc] initWithAnn:self.annot];

        [self loadStrokeColor];
        [self loadThickness];
        [self loadOpacityWithMarkup:markup];
    } else {
        // Load defaults for annotation type.
        [self loadDefaultStrokeColor];
        [self loadDefaultThickness];
        [self loadDefaultOpacity];
    }
}

- (void)configureRedaction
{
    self.availableStyleKeys =
    @[
      PTAnnotStyleKeyStrokeColor,
      PTAnnotStyleKeyFillColor,
      PTAnnotStyleKeyThickness,
      PTAnnotStyleKeyOpacity,
      PTAnnotStyleKeyLabel,
      ];

    if ([self.annot IsValid]) {
        PTRedactionAnnot *redaction = [[PTRedactionAnnot alloc] initWithAnn:self.annot];

        [self loadStrokeColor];
        [self loadFillColorWithMarkup:redaction];
        [self loadThickness];
        [self loadOpacityWithMarkup:redaction];
        [self loadOverlayTextWithRedaction:redaction];
    } else {
        // Load defaults for annotation type.
        [self loadDefaultStrokeColor];
        [self loadDefaultFillColor];
        [self loadDefaultThickness];
        [self loadDefaultOpacity];
        [self loadDefaultOverlayText];
    }
}

- (void)configureLengthMeasure
{
    self.availableStyleKeys =
    @[
      PTAnnotStyleKeyColor,
      PTAnnotStyleKeyThickness,
      PTAnnotStyleKeyOpacity,
      PTAnnotStyleKeyScale,
      PTAnnotStyleKeyPrecision,
      PTAnnotStyleKeySnapping
      ];

    if ([self.annot IsValid]) {
        PTMarkup *markup = [[PTMarkup alloc] initWithAnn:self.annot];

        [self loadStrokeColor];
        [self loadThickness];
        [self loadOpacityWithMarkup:markup];
        [self loadMeasurementData];
    } else {
        // Load defaults for annotation type.
        [self loadDefaultStrokeColor];
        [self loadDefaultThickness];
        [self loadDefaultOpacity];
        [self loadDefaultMeasurementData];
    }
}

- (void)configureAreaMeasure
{
    self.availableStyleKeys =
    @[
      PTAnnotStyleKeyStrokeColor,
      PTAnnotStyleKeyFillColor,
      PTAnnotStyleKeyThickness,
      PTAnnotStyleKeyOpacity,
      PTAnnotStyleKeyScale,
      PTAnnotStyleKeyPrecision,
      PTAnnotStyleKeySnapping
      ];

    if ([self.annot IsValid]) {
        PTMarkup *markup = [[PTMarkup alloc] initWithAnn:self.annot];

        [self loadStrokeColor];
        [self loadFillColorWithMarkup:markup];
        [self loadThickness];
        [self loadOpacityWithMarkup:markup];
        [self loadMeasurementData];
    } else {
        // Load defaults for annotation type.
        [self loadDefaultStrokeColor];
        [self loadDefaultFillColor];
        [self loadDefaultThickness];
        [self loadDefaultOpacity];
        [self loadDefaultMeasurementData];
    }
}

#pragma mark - Annotation saving

- (void)saveChanges
{
    switch (self.annotType) {
        case PTExtendedAnnotTypeText:
            break;
        case PTExtendedAnnotTypeLink:
            break;
        case PTExtendedAnnotTypeFreeText:
        case PTExtendedAnnotTypeCallout:
            [self saveFreeText];
            break;
        case PTExtendedAnnotTypeLine:
        case PTExtendedAnnotTypeArrow:
            [self saveFreeHand];
            break;
        case PTExtendedAnnotTypeRuler:
        case PTExtendedAnnotTypePerimeter:
            [self saveFreeHand];
            [self saveMeasurement];
            break;
        case PTExtendedAnnotTypeSquare:
            [self saveMarkup];
            break;
        case PTExtendedAnnotTypeCircle:
            [self saveMarkup];
            break;
        case PTExtendedAnnotTypePolygon:
        case PTExtendedAnnotTypeCloudy:
            [self saveMarkup];
            break;
        case PTExtendedAnnotTypeArea:
            [self saveMarkup];
            [self saveMeasurement];
            break;
        case PTExtendedAnnotTypePolyline:
            [self saveFreeHand];
            break;
        case PTExtendedAnnotTypeHighlight:
            [self saveHighlight];
            break;
        case PTExtendedAnnotTypeUnderline:
            [self saveTextAnnot];
            break;
        case PTExtendedAnnotTypeSquiggly:
            [self saveTextAnnot];
            break;
        case PTExtendedAnnotTypeStrikeOut:
            [self saveTextAnnot];
            break;
        case PTExtendedAnnotTypeStamp:
        case PTExtendedAnnotTypeImageStamp:
        case PTExtendedAnnotTypeSignature:
            break;
        case PTExtendedAnnotTypeCaret:
            break;
        case PTExtendedAnnotTypeInk:
        case PTExtendedAnnotTypeFreehandHighlight:
            [self saveFreeHand];
            break;
        case PTExtendedAnnotTypePopup:
            break;
        case PTExtendedAnnotTypeFileAttachment:
            break;
        case PTExtendedAnnotTypeSound:
            break;
        case PTExtendedAnnotTypeMovie:
            break;
        case PTExtendedAnnotTypeWidget:
            break;
        case PTExtendedAnnotTypeScreen:
            break;
        case PTExtendedAnnotTypePrinterMark:
            break;
        case PTExtendedAnnotTypeTrapNet:
            break;
        case PTExtendedAnnotTypeWatermark:
            break;
        case PTExtendedAnnotType3D:
            break;
        case PTExtendedAnnotTypeRedact:
            [self saveRedaction];
            break;
        case PTExtendedAnnotTypeProjection:
            break;
        case PTExtendedAnnotTypeRichMedia:
            break;
        case PTExtendedAnnotTypeUnknown:
            break;
        default:
            break;
    }

    [self.annot RefreshAppearance];
    
    if (self.saveValuesAsDefaults) {
        [self setCurrentValuesAsDefaults];
    }
}

- (void)saveFreeText
{
    if (![self.annot IsValid]) {
        return;
    }

    PTFreeText *freeText = [[PTFreeText alloc] initWithAnn:self.annot];

    [self saveTextColorWithFreeText:freeText];
    [self saveBorderColorWithFreeText:freeText];
    [self saveThickness];
    [self saveTextSizeWithFreeText:freeText];
    [self saveOpacityWithMarkup:(PTMarkup *)freeText];
}

- (void)saveFreeHand
{
    if (![self.annot IsValid]) {
        return;
    }

    PTMarkup *markup = [[PTMarkup alloc] initWithAnn:self.annot];

    [self saveStrokeColor];
    [self saveThickness];
    [self saveOpacityWithMarkup:markup];
}

- (void)saveMarkup
{
    if (![self.annot IsValid]) {
        return;
    }

    PTMarkup *markup = [[PTMarkup alloc] initWithAnn:self.annot];

    [self saveStrokeColor];
    [self saveFillColorWithMarkup:markup];
    [self saveThickness];
    [self saveOpacityWithMarkup:markup];
}

- (void)saveHighlight
{
    if (![self.annot IsValid]) {
        return;
    }

    PTMarkup *markup = [[PTMarkup alloc] initWithAnn:self.annot];

    [self saveStrokeColor];
    [self saveOpacityWithMarkup:markup];
}

- (void)saveTextAnnot
{
    if (![self.annot IsValid]) {
        return;
    }

    PTMarkup *markup = [[PTMarkup alloc] initWithAnn:self.annot];

    [self saveStrokeColor];
    [self saveThickness];
    [self saveOpacityWithMarkup:markup];
}

- (void)saveRedaction
{
    if (![self.annot IsValid]) {
        return;
    }

    PTRedactionAnnot *redaction = [[PTRedactionAnnot alloc] initWithAnn:self.annot];

    [self saveStrokeColor];
    [self saveFillColorWithMarkup:redaction];
    [self saveOpacityWithMarkup:redaction];
    [self saveOverlayTextWithRedaction:redaction];
}

- (void)saveMeasurement
{
    if (![self.annot IsValid]) {
        return;
    }
    [self saveMeasurementData];
}

- (void)setCurrentValuesAsDefaults
{
    PTExtendedAnnotType annotType = self.annotType;

    for (PTAnnotStyleKey key in self.availableStyleKeys) {
        if ([key isEqualToString:PTAnnotStyleKeyColor]) {
            [PTColorDefaults setDefaultColor:self.color
                              forAnnotType:annotType
                                 attribute:ATTRIBUTE_STROKE_COLOR
                      colorPostProcessMode:e_ptpostprocess_none];
        }
        else if ([key isEqualToString:PTAnnotStyleKeyStrokeColor]) {
            [PTColorDefaults setDefaultColor:self.strokeColor
                              forAnnotType:annotType
                                 attribute:ATTRIBUTE_STROKE_COLOR
                      colorPostProcessMode:e_ptpostprocess_none];
        }
        else if ([key isEqualToString:PTAnnotStyleKeyFillColor]) {
            [PTColorDefaults setDefaultColor:self.fillColor
                              forAnnotType:annotType
                                 attribute:ATTRIBUTE_FILL_COLOR
                      colorPostProcessMode:e_ptpostprocess_none];
        }
        else if ([key isEqualToString:PTAnnotStyleKeyTextColor]) {
            [PTColorDefaults setDefaultColor:self.textColor
                              forAnnotType:annotType
                                 attribute:ATTRIBUTE_TEXT_COLOR
                      colorPostProcessMode:e_ptpostprocess_none];
        }
        else if ([key isEqualToString:PTAnnotStyleKeyThickness]) {
            [PTColorDefaults setDefaultBorderThickness:self.thickness forAnnotType:annotType];
        }
        else if ([key isEqualToString:PTAnnotStyleKeyOpacity]) {
            [PTColorDefaults setDefaultOpacity:self.opacity forAnnotType:annotType];
        }
        else if ([key isEqualToString:PTAnnotStyleKeyTextSize]) {
            [PTColorDefaults setDefaultFreeTextSize:self.textSize];
        }
        else if ([key isEqualToString:PTAnnotStyleKeyScale] || [key isEqualToString:PTAnnotStyleKeyPrecision]) {
            [PTColorDefaults setDefaultMeasurementScale:self.measurementScale forAnnotType:annotType];
        }
    }
}

#pragma mark - Annotation property loading

- (void)loadDefaultTextColor
{
    self.textColor = [PTColorDefaults defaultColorForAnnotType:self.annotType
                                                   attribute:ATTRIBUTE_TEXT_COLOR
                                        colorPostProcessMode:e_ptpostprocess_none];
}

- (void)loadTextColorWithFreeText:(PTFreeText *)freeText
{
    self.textColor = [PTColorDefaults uiColorFromColorPt:[freeText GetTextColor]
                                               compNum:[freeText GetTextColorCompNum]];
}

- (void)loadDefaultTextSize
{
    self.textSize = [PTColorDefaults defaultFreeTextSize];
}

- (void)loadTextSizeWithFreeText:(PTFreeText *)freeText
{
    self.textSize = [freeText GetFontSize];
}

- (void)loadDefaultStrokeColor
{
    self.strokeColor = [PTColorDefaults defaultColorForAnnotType:self.annotType
                                                     attribute:ATTRIBUTE_STROKE_COLOR
                                          colorPostProcessMode:e_ptpostprocess_none];
}

- (void)loadStrokeColor
{
    self.strokeColor = [PTColorDefaults uiColorFromColorPt:[self.annot GetColorAsRGB] compNum:3];
}

- (void)loadDefaultFillColor
{
    self.fillColor = [PTColorDefaults defaultColorForAnnotType:self.annotType
                                                   attribute:ATTRIBUTE_FILL_COLOR
                                        colorPostProcessMode:e_ptpostprocess_none];
}

- (void)loadFillColorWithMarkup:(PTMarkup *)markup
{
    self.fillColor = [PTColorDefaults uiColorFromColorPt:[markup GetInteriorColor]
                                               compNum:[markup GetInteriorColorCompNum]];
}

- (void)loadBorderColorWithFreeText:(PTFreeText *)freeText
{
    self.strokeColor = [PTColorDefaults uiColorFromColorPt:[freeText GetLineColor]
                                                 compNum:[freeText GetLineColorCompNum]];
}

- (void)loadDefaultBorderColor
{
    self.strokeColor = [PTColorDefaults defaultColorForAnnotType:self.annotType
                                                     attribute:ATTRIBUTE_STROKE_COLOR
                                          colorPostProcessMode:e_ptpostprocess_none];
}

- (void)loadDefaultThickness
{
    self.thickness = [PTColorDefaults defaultBorderThicknessForAnnotType:self.annotType];
}

- (void)loadThickness
{
    PTBorderStyle *borderStyle = [self.annot GetBorderStyle];
    self.thickness = [borderStyle GetWidth];
}

- (void)loadDefaultOpacity
{
    self.opacity = [PTColorDefaults defaultOpacityForAnnotType:self.annotType];
}

- (void)loadOpacityWithMarkup:(PTMarkup *)markup
{
    self.opacity = [markup GetOpacity];
}

- (void)loadDefaultOverlayText
{
    self.label = nil;
}

- (void)loadOverlayTextWithRedaction:(PTRedactionAnnot *)redaction
{
    self.label = [redaction GetOverlayText];
}

- (void)loadDefaultMeasurementData
{
    self.measurementScale = [PTColorDefaults defaultMeasurementScaleForAnnotType:self.annotType];
}

- (void)loadMeasurementData
{
    self.measurementScale = [PTMeasurementUtil getMeasurementScaleFromAnnot:self.annot];
}

#pragma mark - Annotation property saving

- (void)saveTextColorWithFreeText:(PTFreeText *)freeText
{
    PTColorPt *colorPt = [PTColorDefaults colorPtFromUIColor:self.textColor];
    int numComps = [PTColorDefaults numCompsInColorPtForUIColor:self.textColor];

    [freeText SetTextColor:colorPt col_comp:numComps];
}

- (void)saveStrokeColor
{
    PTColorPt *colorPt = [PTColorDefaults colorPtFromUIColor:self.strokeColor];
    int numComps = [PTColorDefaults numCompsInColorPtForUIColor:self.strokeColor];

    [self.annot SetColor:colorPt numcomp:numComps];
}

- (void)saveFillColorWithMarkup:(PTMarkup *)markup
{
    PTColorPt *colorPt = [PTColorDefaults colorPtFromUIColor:self.fillColor];
    int numComps = [PTColorDefaults numCompsInColorPtForUIColor:self.fillColor];

    [markup SetInteriorColor:colorPt CompNum:numComps];
}

- (void)saveBorderColorWithFreeText:(PTFreeText *)freeText
{
    PTColorPt *colorPt = [PTColorDefaults colorPtFromUIColor:self.strokeColor];
    int numComps = [PTColorDefaults numCompsInColorPtForUIColor:self.strokeColor];

    [freeText SetLineColor:colorPt col_comp:numComps];
}

- (void)saveThickness
{
    PTBorderStyle *borderStyle = [self.annot GetBorderStyle];
    
    if (self.annotType == PTExtendedAnnotTypeFreeText && CGColorGetAlpha(self.strokeColor.CGColor) == 0.0) {
        self.thickness = 0.0;
    }
    if (self.annotType == PTExtendedAnnotTypeRuler){
        double offset = self.thickness/2.0f;
        PTLineAnnot *line = [[PTLineAnnot alloc] initWithAnn:self.annot];
        [line SetTextVOffset:offset];
    }
    [borderStyle SetWidth:self.thickness];
    [self.annot SetBorderStyle:borderStyle oldStyleOnly:NO];
}

- (void)saveOpacityWithMarkup:(PTMarkup *)markup
{
    [markup SetOpacity:self.opacity];
}

- (void)saveTextSizeWithFreeText:(PTFreeText *)freeText
{
    [freeText SetFontSize:self.textSize];
}

- (void)saveOverlayTextWithRedaction:(PTRedactionAnnot *)redaction
{
    NSString *label = self.label;
    if (!label) {
        label = @"";
    }
    [redaction SetOverlayText:label];
}

- (void)saveMeasurementData
{
    [PTMeasurementUtil setAnnotMeasurementData:self.annot fromMeasurementScale:self.measurementScale];
}

#pragma mark - Property accessors

- (PTExtendedAnnotType)annotType
{
    if (self.annot) {
        return [self.annot extendedAnnotType];
    }

    return _annotType;
}

- (UIColor *)color
{
    return _strokeColor;
}

- (void)setColor:(UIColor *)color
{
    _strokeColor = color;

    // Notify delegate of change.
    if ([self.delegate respondsToSelector:@selector(annotStyle:colorDidChange:)]) {
        [self.delegate annotStyle:self colorDidChange:color];
    }
}

- (void)setStrokeColor:(UIColor *)strokeColor
{
    _strokeColor = strokeColor;

    if ([self.annot IsValid] && self.annotType != e_ptNote) {
        if ([PTColorDefaults numCompsInColorPtForUIColor:strokeColor] == 0) { // Transparent.
            // Set thickness to 0 and store original thickness in SDFObj.
            if (self.thickness > 0) {
                PTObj *obj = [self.annot GetSDFObj];
                [obj PutNumber:kAnnotStylePDFTronThicknessKey value:self.thickness];
            }
            self.thickness = 0;
        } else {
            // Restore thickness from SDFObj.
            PTObj *obj = [self.annot GetSDFObj];
            PTObj *thicknessObj = [obj FindObj:kAnnotStylePDFTronThicknessKey];
            if ([thicknessObj IsValid]) {
                double storedThickness = [thicknessObj GetNumber];
                [obj EraseDictElementWithKey:kAnnotStylePDFTronThicknessKey];

                self.thickness = storedThickness;
            }
        }
    }

    // Notify delegate of change.
    if ([self.delegate respondsToSelector:@selector(annotStyle:strokeColorDidChange:)]) {
        [self.delegate annotStyle:self strokeColorDidChange:strokeColor];
    }
}

- (void)setFillColor:(UIColor *)fillColor
{
    _fillColor = fillColor;

    // Notify delegate of change.
    if ([self.delegate respondsToSelector:@selector(annotStyle:fillColorDidChange:)]) {
        [self.delegate annotStyle:self fillColorDidChange:fillColor];
    }
}

- (void)setTextColor:(UIColor *)textColor
{
    _textColor = textColor;

    // Notify delegate of change.
    if ([self.delegate respondsToSelector:@selector(annotStyle:textColorDidChange:)]) {
        [self.delegate annotStyle:self textColorDidChange:textColor];
    }
}

- (void)setThickness:(CGFloat)thickness
{
    _thickness = thickness;

    // Notify delegate of change.
    if ([self.delegate respondsToSelector:@selector(annotStyle:thicknessDidChange:)]) {
        [self.delegate annotStyle:self thicknessDidChange:thickness];
    }
}

- (void)setOpacity:(CGFloat)opacity
{
    _opacity = opacity;

    // Notify delegate of change.
    if ([self.delegate respondsToSelector:@selector(annotStyle:opacityDidChange:)]) {
        [self.delegate annotStyle:self opacityDidChange:opacity];
    }
}

- (void)setTextSize:(CGFloat)textSize
{
    _textSize = textSize;

    // Notify delegate of change.
    if ([self.delegate respondsToSelector:@selector(annotStyle:textSizeDidChange:)]) {
        [self.delegate annotStyle:self textSizeDidChange:textSize];
    }
}

- (void)setLabel:(NSString *)label
{
    _label = label;

    // Notify delegate of change.
    if ([self.delegate respondsToSelector:@selector(annotStyle:labelDidChange:)]) {
        [self.delegate annotStyle:self labelDidChange:label];
    }
}

- (void)setMeasurementScale:(PTMeasurementScale *)measurementScale
{
    _measurementScale = measurementScale;
    if ([self.delegate respondsToSelector:@selector(annotStyle:measurementScaleDidChange:)]) {
        [self.delegate annotStyle:self measurementScaleDidChange:measurementScale];
    }
}

- (void)setSnappingEnabled:(BOOL)snappingEnabled
{
    _snappingEnabled = snappingEnabled;
    if ([self.delegate respondsToSelector:@selector(annotStyle:snappingToggled:)]) {
        [self.delegate annotStyle:self snappingToggled:snappingEnabled];
    }
}

- (NSString *)thicknessIndicatorString
{
    return [NSString stringWithFormat:@"%ld pt", (long) self.thickness];
}

- (NSString *)opacityIndicatorString
{
    return [NSString stringWithFormat:@"%ld%%", (long) (self.opacity * 100.0)];
}

- (NSString *)textSizeIndicatorString
{
    return [NSString stringWithFormat:@"%ld pt", (long) self.textSize];
}

@end
