//---------------------------------------------------------------------------------------
// Copyright (c) 2001-2020 by PDFTron Systems Inc. All Rights Reserved.
// Consult legal.txt regarding legal and license information.
//---------------------------------------------------------------------------------------

#import "ToolsConfig.h"
#import "PTAnnotStyleViewController.h"

#import "PTAnnotStylePreview.h"
#import "PTAnnotStyleTableViewController.h"
#import "PTColorPickerViewController.h"
#import "PTToolsUtil.h"
#import "PTMeasurementScale.h"

@interface PTAnnotStyleViewController () <UINavigationControllerDelegate, PTAnnotStyleTableViewControllerDelegate, PTColorPickerViewControllerDelegate>

@property (nonatomic, strong, nullable) PTToolManager *toolManager;

@property (nonatomic, strong) UIView *annotPreviewContainer;
@property (nonatomic, strong) PTAnnotStylePreview *annotPreviewView;

@property (nonatomic, strong) UINavigationController *embeddedNavigationController;
@property (nonatomic, strong) PTAnnotStyleTableViewController *tableViewController;

@property (nonatomic, strong) PTColorPickerViewController *colorPickerViewController;

@property (nonatomic) BOOL didSetupConstraints;

@property (nonatomic) BOOL needsPreferredContentSizeUpdate;

@property (nonatomic) BOOL navigationControllerTransitioning;

- (void)configureWithAnnotStyle:(PTAnnotStyle *)annotStyle;

@end

@implementation PTAnnotStyleViewController

- (instancetype)initWithAnnotStyle:(PTAnnotStyle *)annotStyle
{
    self = [super initWithNibName:nil bundle:nil];
    if (self) {
        _annotStyle = annotStyle;

        _annotStyle.delegate = self;

        // Use half-modal presentation as default.
        _presentationManager = [[PTHalfModalPresentationManager alloc] init];

        self.modalPresentationStyle = UIModalPresentationCustom;
        self.transitioningDelegate = _presentationManager;
    }
    return self;
}

- (instancetype)initWithToolManager:(PTToolManager *)toolManager annotStyle:(PTAnnotStyle *)annotStyle
{
    self = [self initWithAnnotStyle:annotStyle];
    if (self) {
        _toolManager = toolManager;
    }
    return self;
}

- (void)configureWithAnnotStyle:(PTAnnotStyle *)annotStyle
{
    NSMutableArray *sections = [NSMutableArray array];
    NSMutableArray *items = [NSMutableArray array];

    for (PTAnnotStyleKey key in annotStyle.availableStyleKeys) {
        PTAnnotStyleTableViewItem *item = nil;

        if ([key isEqualToString:PTAnnotStyleKeyColor]) {
            NSString *title = PTLocalizedString(@"Color", nil);

            item = [[PTAnnotStyleColorTableViewItem alloc] initWithTitle:title
                                                                 color:annotStyle.color
                                                         annotStyleKey:PTAnnotStyleKeyColor];

            self.annotPreviewView.color = annotStyle.color;
        }
        if ([key isEqualToString:PTAnnotStyleKeyStrokeColor]) {
            NSString *title = PTLocalizedString(@"Stroke Color", nil);

            if (self.annotStyle.annotType == PTExtendedAnnotTypeFreeText) {
                title = PTLocalizedString(@"Border Color", nil);
            }

            item = [[PTAnnotStyleColorTableViewItem alloc] initWithTitle:title
                                                                 color:annotStyle.strokeColor
                                                         annotStyleKey:PTAnnotStyleKeyStrokeColor];

            self.annotPreviewView.color = annotStyle.strokeColor;
        }
        if ([key isEqualToString:PTAnnotStyleKeyFillColor]) {
            NSString *title = PTLocalizedString(@"Fill Color", nil);

            item = [[PTAnnotStyleColorTableViewItem alloc] initWithTitle:title
                                                                 color:annotStyle.fillColor
                                                         annotStyleKey:PTAnnotStyleKeyFillColor];

            self.annotPreviewView.fillColor = annotStyle.fillColor;
        }
        if ([key isEqualToString:PTAnnotStyleKeyTextColor]) {
            NSString *title = PTLocalizedString(@"Text Color", nil);

            item = [[PTAnnotStyleColorTableViewItem alloc] initWithTitle:title
                                                                 color:annotStyle.textColor
                                                         annotStyleKey:PTAnnotStyleKeyTextColor];

            self.annotPreviewView.textColor = annotStyle.textColor;
        }

        if ([key isEqualToString:PTAnnotStyleKeyThickness]) {
            // HACK: Don't show the thickness slider for redaction annotations.
            if (annotStyle.annotType != PTExtendedAnnotTypeRedact) {
                NSString *title = PTLocalizedString(@"Thickness", nil);
                
                CGFloat minimumValue = 1.0;
                
                CGFloat maximumValue = 12.0;
                if (annotStyle.annotType == PTExtendedAnnotTypeFreehandHighlight) {
                    maximumValue = 20.0;
                }
                
                // Consult delegate for minimum value.
                if ([self.delegate respondsToSelector:@selector(annotStyleViewController:minimumValue:forStyle:key:)]) {
                    [self.delegate annotStyleViewController:self minimumValue:&minimumValue forStyle:annotStyle key:key];
                }

                // Consult delegate for maximum value.
                if ([self.delegate respondsToSelector:@selector(annotStyleViewController:maximumValue:forStyle:key:)]) {
                    [self.delegate annotStyleViewController:self maximumValue:&maximumValue forStyle:annotStyle key:key];
                }

                item = [[PTAnnotStyleSliderTableViewItem alloc] initWithTitle:title
                                                                 minimumValue:minimumValue
                                                                 maximumValue:maximumValue
                                                                        value:annotStyle.thickness
                                                                indicatorText:annotStyle.thicknessIndicatorString
                                                                annotStyleKey:PTAnnotStyleKeyThickness];
            }

            self.annotPreviewView.thickness = annotStyle.thickness;
        }
        if ([key isEqualToString:PTAnnotStyleKeyOpacity]) {
            NSString *title = PTLocalizedString(@"Opacity", nil);

            CGFloat minimumValue = 10.0;
            CGFloat maximumValue = 100.0;
            
            // Consult delegate for minimum value.
            if ([self.delegate respondsToSelector:@selector(annotStyleViewController:minimumValue:forStyle:key:)]) {
                [self.delegate annotStyleViewController:self minimumValue:&minimumValue forStyle:annotStyle key:key];
            }

            // Consult delegate for maximum value.
            if ([self.delegate respondsToSelector:@selector(annotStyleViewController:maximumValue:forStyle:key:)]) {
                [self.delegate annotStyleViewController:self maximumValue:&maximumValue forStyle:annotStyle key:key];
            }
            
            item = [[PTAnnotStyleSliderTableViewItem alloc] initWithTitle:title
                                                           minimumValue:minimumValue
                                                           maximumValue:maximumValue
                                                                  value:(annotStyle.opacity * 100.0)
                                                          indicatorText:annotStyle.opacityIndicatorString
                                                          annotStyleKey:PTAnnotStyleKeyOpacity];

            self.annotPreviewView.opacity = annotStyle.opacity;
        }
        if ([key isEqualToString:PTAnnotStyleKeyTextSize]) {
            NSString *title = PTLocalizedString(@"Text Size", nil);
            
            CGFloat minimumValue = 4.0;
            CGFloat maximumValue = 72.0;
            
            // Consult delegate for minimum value.
            if ([self.delegate respondsToSelector:@selector(annotStyleViewController:minimumValue:forStyle:key:)]) {
                [self.delegate annotStyleViewController:self minimumValue:&minimumValue forStyle:annotStyle key:key];
            }

            // Consult delegate for maximum value.
            if ([self.delegate respondsToSelector:@selector(annotStyleViewController:maximumValue:forStyle:key:)]) {
                [self.delegate annotStyleViewController:self maximumValue:&maximumValue forStyle:annotStyle key:key];
            }

            item = [[PTAnnotStyleSliderTableViewItem alloc] initWithTitle:title
                                                           minimumValue:minimumValue
                                                           maximumValue:maximumValue
                                                                  value:annotStyle.textSize
                                                          indicatorText:annotStyle.textSizeIndicatorString
                                                          annotStyleKey:PTAnnotStyleKeyTextSize];

            self.annotPreviewView.textSize = annotStyle.textSize;
        }
        if ([key isEqualToString:PTAnnotStyleKeyLabel]) {
            NSString *title = nil;
            if (annotStyle.annotType == PTExtendedAnnotTypeRedact) {
                title = PTLocalizedString(@"Overlay text", @"Overlay text for redaction annot");
            } else {
                title = PTLocalizedString(@"Label", nil);
            }

            item = [[PTAnnotStyleTextFieldTableViewItem alloc] initWithTitle:title
                                                                        text:annotStyle.label
                                                               annotStyleKey:PTAnnotStyleKeyLabel];

            self.annotPreviewView.textSize = annotStyle.textSize;
        }
        if ([key isEqualToString:PTAnnotStyleKeyScale]) {
            NSString *title = PTLocalizedString(@"Scale", @"Scale style view controller cell title");

            item = [[PTAnnotStyleScaleTableViewItem alloc] initWithTitle:title
                                                                         measurementScale:annotStyle.measurementScale
                                                                 annotStyleKey:PTAnnotStyleKeyScale];
        }
        if ([key isEqualToString:PTAnnotStyleKeyPrecision]) {
            NSString *title = PTLocalizedString(@"Precision", @"Precision style view controller cell title");
            
            item = [[PTAnnotStylePrecisionTableViewItem alloc] initWithTitle:title
                                                                   measurementScale:annotStyle.measurementScale
                                                           annotStyleKey:PTAnnotStyleKeyPrecision];
        }
        if ([key isEqualToString:PTAnnotStyleKeySnapping]) {
            NSString *title = PTLocalizedString(@"Snapping", @"Measurement tool snapping switch label");

            item = [[PTAnnotStyleSwitchTableViewItem alloc] initWithTitle:title
                                                          snappingEnabled:self.toolManager.snapToDocumentGeometryEnabled
                                                            annotStyleKey:PTAnnotStyleKeySnapping];
        }
        if (item) {
            [items addObject:item];
        }
    }

    // Add style controls section.
    [sections addObject:items];

    self.annotPreviewView.annotType = annotStyle.annotType;

    self.tableViewController.items = sections;
}

- (void)selectStyle
{
    // Notify delegate of current style selection.
    if ([self.delegate respondsToSelector:@selector(annotStyleViewController:didCommitStyle:)]) {
        [self.delegate annotStyleViewController:self didCommitStyle:self.annotStyle];
    }
}

#pragma mark - View lifecycle

- (void)viewDidLoad
{
    [super viewDidLoad];

//    self.view.backgroundColor = UIColor.whiteColor;
//    self.view.layer.cornerRadius = 10;
//    self.view.layer.masksToBounds = YES;

    // Initialize style preview.
    self.annotPreviewContainer = [[UIView alloc] init];
    UIColor *annotPreviewContainerColor = [UIColor colorWithWhite:0.97 alpha:1.0];
    if (@available(iOS 11.0, *)) {
        annotPreviewContainerColor = [UIColor colorNamed:@"annotPreviewContainerColor" inBundle:[PTToolsUtil toolsBundle] compatibleWithTraitCollection:self.traitCollection];
    }

    self.annotPreviewContainer.backgroundColor = annotPreviewContainerColor;
    [self.view addSubview:self.annotPreviewContainer];

    self.annotPreviewView = [[PTAnnotStylePreview alloc] init];
    self.annotPreviewView.textColor = [UIColor blackColor];

    [self.annotPreviewContainer addSubview:self.annotPreviewView];

    // Initialize style list view controller.
    self.tableViewController = [[PTAnnotStyleTableViewController alloc] init];
    self.tableViewController.extendedLayoutIncludesOpaqueBars = YES;
    self.tableViewController.delegate = self;
    self.tableViewController.tableView.tableFooterView = [[UIView alloc] init];

    // Initialize embedded navigation controller with the style list as the root.
    self.embeddedNavigationController = [[UINavigationController alloc] initWithRootViewController:self.tableViewController];
    self.embeddedNavigationController.delegate = self;

    // Add embedded navigation controller as child.
    [self addChildViewController:self.embeddedNavigationController];

    [self.view addSubview:self.embeddedNavigationController.view];

    [self.embeddedNavigationController didMoveToParentViewController:self];

    // Schedule constraints update.
    [self.view setNeedsUpdateConstraints];

    [self configureWithAnnotStyle:self.annotStyle];
}

- (void)updateViewConstraints
{
    if (!self.didSetupConstraints) {
        self.annotPreviewContainer.translatesAutoresizingMaskIntoConstraints = NO;
        self.annotPreviewView.translatesAutoresizingMaskIntoConstraints = NO;
        self.embeddedNavigationController.view.translatesAutoresizingMaskIntoConstraints = NO;

        UIView *navigationControllerView = self.embeddedNavigationController.view;

        NSLayoutConstraint *annotPreviewContainerCenterXConstraint, *annotPreviewContainerWidthConstraint, *annotPreviewContainerTopConstraint;
        NSLayoutConstraint *navigationControllerLeadingConstraint, *navigationControllerWidthConstraint, *navigationControllerBottomConstraint;

        if (@available(iOS 11, *)) {
            annotPreviewContainerCenterXConstraint = [self.annotPreviewContainer.centerXAnchor constraintEqualToAnchor:self.annotPreviewContainer.superview.safeAreaLayoutGuide.centerXAnchor];
            annotPreviewContainerWidthConstraint = [self.annotPreviewContainer.widthAnchor constraintEqualToAnchor:self.annotPreviewContainer.superview.safeAreaLayoutGuide.widthAnchor];
            annotPreviewContainerTopConstraint = [self.annotPreviewContainer.topAnchor constraintEqualToAnchor:self.annotPreviewContainer.superview.safeAreaLayoutGuide.topAnchor];

            navigationControllerLeadingConstraint = [navigationControllerView.leadingAnchor constraintEqualToAnchor:navigationControllerView.superview.safeAreaLayoutGuide.leadingAnchor];
            navigationControllerWidthConstraint = [navigationControllerView.widthAnchor constraintEqualToAnchor:navigationControllerView.superview.safeAreaLayoutGuide.widthAnchor];
            navigationControllerBottomConstraint = [navigationControllerView.bottomAnchor constraintEqualToAnchor:navigationControllerView.superview.safeAreaLayoutGuide.bottomAnchor];
        } else {
            annotPreviewContainerCenterXConstraint = [self.annotPreviewContainer.centerXAnchor constraintEqualToAnchor:self.annotPreviewContainer.superview.centerXAnchor];
            annotPreviewContainerWidthConstraint = [self.annotPreviewContainer.widthAnchor constraintEqualToAnchor:self.annotPreviewContainer.superview.widthAnchor];
            annotPreviewContainerTopConstraint = [self.annotPreviewContainer.topAnchor constraintEqualToAnchor:self.annotPreviewContainer.superview.topAnchor];

            navigationControllerLeadingConstraint = [navigationControllerView.leadingAnchor constraintEqualToAnchor:navigationControllerView.superview.leadingAnchor];
            navigationControllerWidthConstraint = [navigationControllerView.widthAnchor constraintEqualToAnchor:navigationControllerView.superview.widthAnchor];
            navigationControllerBottomConstraint = [navigationControllerView.bottomAnchor constraintEqualToAnchor:navigationControllerView.superview.bottomAnchor];
        }

        [NSLayoutConstraint activateConstraints:
         @[
             annotPreviewContainerCenterXConstraint,
             annotPreviewContainerWidthConstraint,
             annotPreviewContainerTopConstraint,
             [self.annotPreviewContainer.heightAnchor constraintEqualToConstant:88.0],

             [self.annotPreviewView.centerXAnchor constraintEqualToAnchor:self.annotPreviewView.superview.centerXAnchor],
             [self.annotPreviewView.widthAnchor constraintEqualToConstant:150.0],

             [self.annotPreviewView.centerYAnchor constraintEqualToAnchor:self.annotPreviewView.superview.centerYAnchor],
             [self.annotPreviewView.heightAnchor constraintEqualToAnchor:self.annotPreviewView.superview.heightAnchor constant:-20.0],

             navigationControllerLeadingConstraint,
             navigationControllerWidthConstraint,
             [navigationControllerView.topAnchor constraintEqualToAnchor:self.annotPreviewContainer.bottomAnchor],
             navigationControllerBottomConstraint
         ]];

        // Constraints are set up.
        self.didSetupConstraints = YES;
    }

    // Call super implementation as final step.
    [super updateViewConstraints];
}

- (CGSize)topViewControllerPreferredContentSize
{
    UIViewController *topViewController = self.embeddedNavigationController.topViewController;
    if (topViewController) {
        CGSize size = topViewController.preferredContentSize;
        
        if (@available(iOS 11.0, *)) {
            UIEdgeInsets safeAreaInsets = topViewController.view.safeAreaInsets;
            
            size.width += safeAreaInsets.left + safeAreaInsets.right;
            size.height += safeAreaInsets.top + safeAreaInsets.bottom;
        }
        
        return size;
    } else {
        return self.embeddedNavigationController.preferredContentSize;
    }
}

- (void)setNeedsPreferredContentSizeUpdate:(BOOL)needsPreferredContentSizeUpdate
{
    _needsPreferredContentSizeUpdate = needsPreferredContentSizeUpdate;
    
    if (needsPreferredContentSizeUpdate) {
        [self.view setNeedsLayout];
    }
}

- (void)setNavigationControllerTransitioning:(BOOL)transitioning
{
    _navigationControllerTransitioning = transitioning;
    
    if (!transitioning) {
        [self.view setNeedsLayout];
    }
}

- (void)viewDidLayoutSubviews
{
    [super viewDidLayoutSubviews];
    
    if (self.needsPreferredContentSizeUpdate && !self.navigationControllerTransitioning) {
        self.needsPreferredContentSizeUpdate = NO;
        
        CGSize previewSize = self.annotPreviewContainer.frame.size;
        CGSize navigationControllerPreferredContentSize = [self topViewControllerPreferredContentSize];
        
        self.preferredContentSize = CGSizeMake(375,
                                               previewSize.height + navigationControllerPreferredContentSize.height);
    }
}

- (void)setModalPresentationStyle:(UIModalPresentationStyle)modalPresentationStyle
{
    [super setModalPresentationStyle:modalPresentationStyle];
    
    // Remove presentation manager when modal presentation style is changed.
    if (self.presentationManager == self.transitioningDelegate) {
        self.transitioningDelegate = nil;
        self.presentationManager = nil;
    }
}

#pragma mark - <UINavigationControllerDelegate>

- (void)navigationController:(UINavigationController *)navigationController willShowViewController:(UIViewController *)viewController animated:(BOOL)animated
{
    if (self.embeddedNavigationController == navigationController) {
        self.navigationControllerTransitioning = YES;

        viewController.navigationItem.rightBarButtonItem = [[UIBarButtonItem alloc] initWithBarButtonSystemItem:UIBarButtonSystemItemDone target:self action:@selector(doneAction)];
    }
}

- (void)navigationController:(UINavigationController *)navigationController didShowViewController:(UIViewController *)viewController animated:(BOOL)animated
{
    if (self.embeddedNavigationController == navigationController) {
        self.navigationControllerTransitioning = NO;
        
        if (!CGSizeEqualToSize(viewController.preferredContentSize, CGSizeZero)) {
            // Schedule preferredContentSize update.
            self.needsPreferredContentSizeUpdate = YES;
            [self.view setNeedsLayout];
        }
        
        UIView *view = viewController.view;
        UIView *superview = view.superview;
        
        // In some cases, the top view controller (viewController) view's frame is not correct.
        // (It should be the same as its superview's bounds).
        
        if (!CGRectEqualToRect(view.frame, superview.bounds)
            && (view.autoresizingMask & (UIViewAutoresizingFlexibleWidth | UIViewAutoresizingFlexibleHeight))) {
            // Manually update the view controller's frame after the transition is finished.
            view.frame = superview.bounds;
        }
    }
}

- (void)preferredContentSizeDidChangeForChildContentContainer:(id<UIContentContainer>)container
{
    if (self.embeddedNavigationController == container) {
        // Schedule preferredContentSize update.
        self.needsPreferredContentSizeUpdate = YES;
        
        if (!self.navigationControllerTransitioning) {
            [self.view setNeedsLayout];
        }
    }

    [super preferredContentSizeDidChangeForChildContentContainer:container];
}

#pragma mark - <AnnotStyleTableViewControllerDelegate>

- (void)tableViewController:(PTAnnotStyleTableViewController *)tableViewController didSelectItemAtIndexPath:(NSIndexPath *)indexPath
{
    PTAnnotStyleTableViewItem *item = self.tableViewController.items[indexPath.section][indexPath.row];
    if (item.type != PTAnnotStyleTableViewItemTypeColor) {
        return;
    }

    PTAnnotStyleColorTableViewItem *colorItem = (PTAnnotStyleColorTableViewItem *) item;

    // Show color picker (prevent pushing onto stack more than once with fast taps).
    if (![self.embeddedNavigationController.viewControllers containsObject:self.colorPickerViewController]) {
        self.colorPickerViewController.color = colorItem.color;
        self.colorPickerViewController.title = colorItem.title;
        self.colorPickerViewController.tag = colorItem.annotStyleKey;
        
        self.navigationControllerTransitioning = YES;
        [self.embeddedNavigationController pushViewController:self.colorPickerViewController animated:YES];
    } else {
        NSLog(@"Color picker view controller already on navigation stack");
    }
}

- (void)tableViewController:(PTAnnotStyleTableViewController *)tableViewController sliderDidBeginSlidingForItemAtIndexPath:(NSIndexPath *)indexPath
{
    // Do nothing.
}

- (float)tableViewController:(PTAnnotStyleTableViewController *)tableViewController sliderValueDidChange:(float)value forItemAtIndexPath:(NSIndexPath *)indexPath
{
    PTAnnotStyleTableViewItem *item = self.tableViewController.items[indexPath.section][indexPath.row];
    if (item.type != PTAnnotStyleTableViewItemTypeSlider ||
        ![item isKindOfClass:[PTAnnotStyleSliderTableViewItem class]]) {
        return value;
    }

    PTAnnotStyleSliderTableViewItem *sliderItem = (PTAnnotStyleSliderTableViewItem *) item;

    if ([item.annotStyleKey isEqualToString:PTAnnotStyleKeyThickness]) {
        // Round to nearest integer.
        NSInteger integerValue = (NSInteger) roundf(value);
        self.annotStyle.thickness = integerValue;

        sliderItem.value = integerValue;
        sliderItem.indicatorText = self.annotStyle.thicknessIndicatorString;
        return integerValue;
    }
    else if ([item.annotStyleKey isEqualToString:PTAnnotStyleKeyOpacity]) {
        // Round to nearest multiple of 5.
        NSInteger integerValue = (NSInteger) (5.0 * roundf(value / 5.0));
        self.annotStyle.opacity = integerValue / 100.0; // Convert to range [0.0, 1.0].

        sliderItem.value = integerValue;
        sliderItem.indicatorText = self.annotStyle.opacityIndicatorString;
        return integerValue;
    }
    else if ([item.annotStyleKey isEqualToString:PTAnnotStyleKeyTextSize]) {
        // Round to nearest integer.
        NSInteger integerValue = (NSInteger) roundf(value);
        self.annotStyle.textSize = integerValue;

        sliderItem.value = integerValue;
        sliderItem.indicatorText = self.annotStyle.textSizeIndicatorString;
        return integerValue;
    }
    else {
        return value;
    }
}

- (void)tableViewController:(PTAnnotStyleTableViewController *)tableViewController sliderDidEndSlidingForItemAtIndexPath:(NSIndexPath *)indexPath
{
    // Notify delegate of change.
    if ([self.delegate respondsToSelector:@selector(annotStyleViewController:didChangeStyle:)]) {
        [self.delegate annotStyleViewController:self didChangeStyle:self.annotStyle];
    }
}

- (void)tableViewController:(PTAnnotStyleTableViewController *)tableViewController textFieldContentsDidChange:(NSString *)text forItemAtIndexPath:(NSIndexPath *)indexPath
{
    PTAnnotStyleTableViewItem *item = self.tableViewController.items[indexPath.section][indexPath.row];
    if (item.type != PTAnnotStyleTableViewItemTypeTextField ||
        ![item isKindOfClass:[PTAnnotStyleTextFieldTableViewItem class]]) {
        return;
    }

    if ([item.annotStyleKey isEqualToString:PTAnnotStyleKeyLabel]) {
        self.annotStyle.label = text;
    }

    // Notify delegate of change.
    if ([self.delegate respondsToSelector:@selector(annotStyleViewController:didChangeStyle:)]) {
        [self.delegate annotStyleViewController:self didChangeStyle:self.annotStyle];
    }
}

- (PTMeasurementScale *)tableViewController:(PTAnnotStyleTableViewController *)tableViewController scaleDidChange:(PTMeasurementScale *)measurementScale forItemAtIndexPath:(NSIndexPath *)indexPath
{
   PTAnnotStyleTableViewItem *item = self.tableViewController.items[indexPath.section][indexPath.row];
    if (item.type != PTAnnotStyleTableViewItemTypeScale ||
        ![item isKindOfClass:[PTAnnotStyleScaleTableViewItem class]]) {
        return measurementScale;
    }

    PTAnnotStyleScaleTableViewItem *scaleItem = (PTAnnotStyleScaleTableViewItem *) item;
    scaleItem.measurementScale = measurementScale;
    self.annotStyle.measurementScale = measurementScale;

    if ([self.delegate respondsToSelector:@selector(annotStyleViewController:didChangeStyle:)]) {
        [self.delegate annotStyleViewController:self didChangeStyle:self.annotStyle];
    }
    return measurementScale;
}

- (void)tableViewController:(PTAnnotStyleTableViewController *)tableViewController snappingToggled:(BOOL)snappingEnabled forItemAtIndexPath:(nonnull NSIndexPath *)indexPath
{
    PTAnnotStyleTableViewItem *item = self.tableViewController.items[indexPath.section][indexPath.row];
    if (item.type != PTAnnotStyleTableViewItemTypeSwitch ||
        ![item isKindOfClass:[PTAnnotStyleSwitchTableViewItem class]]) {
        return;
    }

    PTAnnotStyleSwitchTableViewItem *switchItem = (PTAnnotStyleSwitchTableViewItem *) item;
    switchItem.snappingEnabled = snappingEnabled;
    self.annotStyle.snappingEnabled = snappingEnabled;

    if ([self.delegate respondsToSelector:@selector(annotStyleViewController:didChangeStyle:)]) {
        [self.delegate annotStyleViewController:self didChangeStyle:self.annotStyle];
    }
    return;
}

- (PTMeasurementScale *)tableViewController:(PTAnnotStyleTableViewController *)tableViewController precisionDidChange:(PTMeasurementScale *)measurementScale forItemAtIndexPath:(NSIndexPath *)indexPath
{
    PTAnnotStyleTableViewItem *item = self.tableViewController.items[indexPath.section][indexPath.row];
    if (item.type != PTAnnotStyleTableViewItemTypePrecision ||
        ![item isKindOfClass:[PTAnnotStylePrecisionTableViewItem class]]) {
        return measurementScale;
    }
    
    PTAnnotStylePrecisionTableViewItem *precisionItem = (PTAnnotStylePrecisionTableViewItem *) item;
    precisionItem.measurementScale = measurementScale;
    self.annotStyle.measurementScale = measurementScale;
    
    if ([self.delegate respondsToSelector:@selector(annotStyleViewController:didChangeStyle:)]) {
        [self.delegate annotStyleViewController:self didChangeStyle:self.annotStyle];
    }
    return measurementScale;
}
#pragma mark - <PTColorPickerViewControllerDelegate>

- (void)colorPickerController:(PTColorPickerViewController *)colorPickerController didSelectColor:(UIColor *)color
{
    PTAnnotStyleKey key = colorPickerController.tag;

    if ([key isEqualToString:PTAnnotStyleKeyColor]) {
        self.annotStyle.color = color;
    }
    else if ([key isEqualToString:PTAnnotStyleKeyStrokeColor]) {
        self.annotStyle.strokeColor = color;
    }
    else if ([key isEqualToString:PTAnnotStyleKeyFillColor]) {
        self.annotStyle.fillColor = color;
    }
    else if ([key isEqualToString:PTAnnotStyleKeyTextColor]) {
        self.annotStyle.textColor = color;
    }

    // Notify delegate of change.
    if ([self.delegate respondsToSelector:@selector(annotStyleViewController:didChangeStyle:)]) {
        [self.delegate annotStyleViewController:self didChangeStyle:self.annotStyle];
    }
}

# pragma mark - <AnnotStyleDelegate>

- (void)annotStyle:(PTAnnotStyle *)annotStyle colorDidChange:(UIColor *)color
{
    self.annotPreviewView.color = color;

    PTAnnotStyleTableViewItem *item = [self.tableViewController tableViewItemForAnnotStyleKey:PTAnnotStyleKeyColor];
    if (item && item.type == PTAnnotStyleTableViewItemTypeColor) {
        PTAnnotStyleColorTableViewItem *colorItem = (PTAnnotStyleColorTableViewItem *) item;

        colorItem.color = color;

        [self.tableViewController.tableView reloadData];
    }
}

- (void)annotStyle:(PTAnnotStyle *)annotStyle strokeColorDidChange:(UIColor *)strokeColor
{
    self.annotPreviewView.color = strokeColor;

    PTAnnotStyleTableViewItem *item = [self.tableViewController tableViewItemForAnnotStyleKey:PTAnnotStyleKeyStrokeColor];
    if (item && item.type == PTAnnotStyleTableViewItemTypeColor) {
        PTAnnotStyleColorTableViewItem *colorItem = (PTAnnotStyleColorTableViewItem *) item;

        colorItem.color = strokeColor;

        [self.tableViewController.tableView reloadData];
    }

}

- (void)annotStyle:(PTAnnotStyle *)annotStyle fillColorDidChange:(UIColor *)fillColor
{
    self.annotPreviewView.fillColor = fillColor;

    PTAnnotStyleTableViewItem *item = [self.tableViewController tableViewItemForAnnotStyleKey:PTAnnotStyleKeyFillColor];
    if (item && item.type == PTAnnotStyleTableViewItemTypeColor) {
        PTAnnotStyleColorTableViewItem *colorItem = (PTAnnotStyleColorTableViewItem *) item;

        colorItem.color = fillColor;

        [self.tableViewController.tableView reloadData];
    }
}

- (void)annotStyle:(PTAnnotStyle *)annotStyle textColorDidChange:(UIColor *)textColor
{
    self.annotPreviewView.textColor = textColor;

    PTAnnotStyleTableViewItem *item = [self.tableViewController tableViewItemForAnnotStyleKey:PTAnnotStyleKeyTextColor];
    if (item && item.type == PTAnnotStyleTableViewItemTypeColor) {
        PTAnnotStyleColorTableViewItem *colorItem = (PTAnnotStyleColorTableViewItem *) item;

        colorItem.color = textColor;

        [self.tableViewController.tableView reloadData];
    }
}

- (void)annotStyle:(PTAnnotStyle *)annotStyle thicknessDidChange:(CGFloat)thickness
{
    self.annotPreviewView.thickness = thickness;
}

- (void)annotStyle:(PTAnnotStyle *)annotStyle opacityDidChange:(CGFloat)opacity
{
    self.annotPreviewView.opacity = opacity;
}

- (void)annotStyle:(PTAnnotStyle *)annotStyle textSizeDidChange:(CGFloat)textSize
{
    self.annotPreviewView.textSize = textSize;
}

- (void)annotStyle:(PTAnnotStyle *)annotStyle labelDidChange:(NSString *)label
{
    PTAnnotStyleTableViewItem *item = [self.tableViewController tableViewItemForAnnotStyleKey:PTAnnotStyleKeyLabel];
    if (item && item.type == PTAnnotStyleTableViewItemTypeTextField) {
        PTAnnotStyleTextFieldTableViewItem *textItem = (PTAnnotStyleTextFieldTableViewItem *)item;

        textItem.text = label;

//        [self.tableViewController.tableView reloadData];
    }
}

- (void)annotStyle:(PTAnnotStyle *)annotStyle measurementScaleDidChange:(PTMeasurementScale *)measurementScale
{
    PTAnnotStyleTableViewItem *item =  [self.tableViewController tableViewItemForAnnotStyleKey:PTAnnotStyleKeyScale];
    if (item && item.type == PTAnnotStyleTableViewItemTypeScale) {
        PTAnnotStyleScaleTableViewItem *scaleItem = (PTAnnotStyleScaleTableViewItem *) item;

        scaleItem.measurementScale = measurementScale;

//        [self.tableViewController.tableView reloadData];
    }
}

- (void)annotStyle:(PTAnnotStyle *)annotStyle snappingToggled:(BOOL)snappingEnabled
{
    PTAnnotStyleTableViewItem *item =  [self.tableViewController tableViewItemForAnnotStyleKey:PTAnnotStyleKeySnapping];
    if (item && item.type == PTAnnotStyleTableViewItemTypeSwitch) {
        PTAnnotStyleSwitchTableViewItem *switchItem = (PTAnnotStyleSwitchTableViewItem *) item;
        self.toolManager.snapToDocumentGeometry = snappingEnabled;
        switchItem.snappingEnabled = snappingEnabled;
    }
}

#pragma mark - Navigation bar button actions

- (void)doneAction
{
    // Notify delegate of current style selection.
    if ([self.delegate respondsToSelector:@selector(annotStyleViewController:didCommitStyle:)]) {
        [self.delegate annotStyleViewController:self didCommitStyle:self.annotStyle];
    } else {
        // Dismiss manually.
        [self.presentingViewController dismissViewControllerAnimated:YES completion:nil];
    }
}

#pragma mark - Property accessors

- (void)setAnnotStyle:(PTAnnotStyle *)annotStyle
{
    _annotStyle = annotStyle;
    _annotStyle.delegate = self;

    [self configureWithAnnotStyle:annotStyle];
}

- (PTColorPickerViewController *)colorPickerViewController
{
    // Lazy load color picker.
    if (!_colorPickerViewController) {
        _colorPickerViewController = [[PTColorPickerViewController alloc]
                                      initWithTransitionStyle:UIPageViewControllerTransitionStyleScroll
                                      navigationOrientation:UIPageViewControllerNavigationOrientationHorizontal
                                      options:nil];

        _colorPickerViewController.colorPickerDelegate = self;
        _colorPickerViewController.extendedLayoutIncludesOpaqueBars = YES;
    }
    return _colorPickerViewController;
}

@end
