//---------------------------------------------------------------------------------------
// Copyright (c) 2001-2020 by PDFTron Systems Inc. All Rights Reserved.
// Consult legal.txt regarding legal and license information.
//---------------------------------------------------------------------------------------

#import "ToolsConfig.h"
#import "PTFreeHandCreate.h"
#import "PTPanTool.h"
#import "PTTimer.h"

static const NSTimeInterval PTFreeHandCreateStylusAppendInterval = 2.0; // seconds
static const CGFloat PTFreeHandCreateStylusAppendDistance = 200; // points in page space

@interface PTFreeHandCreate ()
{
	int m_startPageNum;
    NSMutableArray* m_redo_strokes;
    NSMutableArray* m_free_hand_strokes;
    NSMutableArray* m_free_hand_points;
    
    NSMutableArray* m_page_space_redo_strokes;
    NSMutableArray* m_page_space_free_hand_strokes;
    NSMutableArray* m_page_space_free_hand_points;
    
    CGContextRef context;

    PTTimer *stylusAppendTimeout;
}
@end

@implementation PTFreeHandCreate

@synthesize multistrokeMode;
@synthesize pencilMode;

-(void)ignorePinch:(UIGestureRecognizer *)gestureRecognizer
{
    return;
}

-(BOOL)pdfViewCtrl:(PTPDFViewCtrl*)pdfViewCtrl handleDoubleTap:(UITapGestureRecognizer*)gestureRecognizer
{
	// ignore
	return YES;
}

static CGPoint midPoint(CGPoint p1, CGPoint p2)
{
	
    return CGPointMake((p1.x + p2.x) * 0.5, (p1.y + p2.y) * 0.5);
	
}

- (instancetype)initWithPDFViewCtrl:(PTPDFViewCtrl*)in_pdfViewCtrl
{
    self = [super initWithPDFViewCtrl:in_pdfViewCtrl];
    if (self) {

        self.opaque = NO;
        
		m_startPageNum = 0; // non-existant page in PDF
		
		UIPinchGestureRecognizer* pgr = [[UIPinchGestureRecognizer alloc] initWithTarget:self action:@selector(ignorePinch:)];
  
        [self addGestureRecognizer:pgr];

        self.frame = CGRectMake([self.pdfViewCtrl GetHScrollPos], [self.pdfViewCtrl GetVScrollPos], self.pdfViewCtrl.frame.size.width, self.pdfViewCtrl.frame.size.height);
    }

    return self;
}

-(Class)annotClass
{
    return [PTInk class];
}

+(PTExtendedAnnotType)annotType
{
	return PTExtendedAnnotTypeInk;
}

-(BOOL)canEditStyle
{
    return YES;
}

- (void)pdfViewCtrl:(PTPDFViewCtrl *)pdfViewCtrl pageNumberChangedFrom:(int)oldPageNumber To:(int)newPageNumber
{
    [super pdfViewCtrl:pdfViewCtrl pageNumberChangedFrom:oldPageNumber To:newPageNumber];
    
    [self endDrawing];
    [self.toolManager changeTool:[PTPanTool class]];
}

- (BOOL)pdfViewCtrl:(PTPDFViewCtrl*)pdfViewCtrl touchesShouldCancelInContentView:(UIView *)view
{
    if( self.multistrokeMode == YES )
		return NO;
	else
		return [super pdfViewCtrl:pdfViewCtrl touchesShouldCancelInContentView:view];
}

+(BOOL)createsAnnotation
{
	return YES;
}

- (void)drawRect:(CGRect)rect
{
    if( m_free_hand_points && (self.multistrokeMode == YES || !self.allowScrolling))
    {
        CGPoint previousPoint1 = CGPointZero;
        CGPoint previousPoint2 = CGPointZero;
        CGPoint currentPoint;
        
        context = UIGraphicsGetCurrentContext();
        
		if( ! context )
			return;
		
        [super setupContext:context];
        
        CGContextSetLineJoin(context, kCGLineJoinRound);
        
		for (NSArray<NSValue*>* point_array in m_free_hand_strokes)
		{
            previousPoint1 = CGPointZero;
            previousPoint2 = CGPointZero;
            
            CGContextBeginPath(context);
			
			for (NSValue* val in point_array)
			{
				currentPoint = val.CGPointValue;
				
				if( CGPointEqualToPoint(previousPoint1, CGPointZero))
					previousPoint1 = currentPoint;
				
				if( CGPointEqualToPoint(previousPoint2, CGPointZero))
					previousPoint2 = currentPoint;
				
				CGPoint mid1 = midPoint(previousPoint1, previousPoint2);
				CGPoint	mid2 = midPoint(currentPoint, previousPoint1);
				
				CGContextMoveToPoint(context, mid1.x, mid1.y);
				
				CGContextAddQuadCurveToPoint(context, previousPoint1.x, previousPoint1.y, mid2.x, mid2.y);
			
				previousPoint2 = previousPoint1;
				previousPoint1 = currentPoint;
			}
            
            CGContextStrokePath(context);
		}
    }
}

-(void)addFreehandPoint:(NSSet*)touches
{
    UITouch *touch = touches.allObjects[0];
    
    CGPoint touchPoint = [touch preciseLocationInView:self.pdfViewCtrl];
    
    self.endPoint = [super boundToPageScreenPoint:touchPoint withThicknessCorrection:_thickness/2];
    
    [m_free_hand_points addObject:[NSValue valueWithCGPoint:self.endPoint]];
    
    CGPoint pagePoint = [self convertScreenPtToPagePt:self.endPoint onPageNumber:self.pageNumber];
    [m_page_space_free_hand_points addObject:[NSValue valueWithCGPoint:pagePoint]];
    
    CGPoint lastDrawPoint = [[m_free_hand_points objectAtIndex:MAX(0,(int)[m_free_hand_points count]-3)] CGPointValue];

    [self setNeedsDisplayInRect:CGRectMake(MIN(self.endPoint.x,lastDrawPoint.x)-10-_thickness, MIN(self.endPoint.y,lastDrawPoint.y)-10-_thickness, ABS(self.endPoint.x-lastDrawPoint.x)+20+_thickness*2, ABS(self.endPoint.y-lastDrawPoint.y)+20+_thickness*2)];

}

- (BOOL)pdfViewCtrl:(PTPDFViewCtrl*)pdfViewCtrl onTouchesBegan:(NSSet *)touches withEvent:(UIEvent *)event
{

    UITouch *touch = touches.allObjects[0];

    self.startPoint = [touch preciseLocationInView:self.pdfViewCtrl];
    CGPoint pageStartPoint = [self convertScreenPtToPagePt:self.startPoint onPageNumber:self.pageNumber];

    PTPDFRect *radiusRect = [self rectFromPoints];
    [radiusRect SetX1:radiusRect.GetX1 - PTFreeHandCreateStylusAppendDistance];
    [radiusRect SetX2:radiusRect.GetX2 + PTFreeHandCreateStylusAppendDistance];
    [radiusRect SetY1:radiusRect.GetY1 - PTFreeHandCreateStylusAppendDistance];
    [radiusRect SetY2:radiusRect.GetY2 + PTFreeHandCreateStylusAppendDistance];

    // If the touch point is within 200 page-points of the current rect, append to the drawing
    BOOL continueAnnotation = pageStartPoint.x > radiusRect.GetX1 && pageStartPoint.x < radiusRect.GetX2 && pageStartPoint.y > radiusRect.GetY1 && pageStartPoint.y < radiusRect.GetY2;

    if (touch.type == UITouchTypePencil && self.backToPanToolAfterUse) {
        pencilMode = YES;
        multistrokeMode = YES;
        [stylusAppendTimeout invalidate];
    }

    if (pencilMode && m_free_hand_points.count > 0) { // In pencil mode
        if (touch.type != UITouchTypePencil) {
            [self endDrawing];
            return YES;
        } else if (!continueAnnotation) {
            [self commitAnnotation];
            return NO;
        }
    }
    
    _pageNumber = [self.pdfViewCtrl GetPageNumberFromScreenPt:self.startPoint.x y:self.startPoint.y];
    
    if (self.pageNumber < 1 || (m_startPageNum > 0 && m_startPageNum != self.pageNumber) ) {
        if (touch.type == UITouchTypePencil) { // If the pencil drawing is on another page, commit the existing annotation
            [self commitAnnotation];
            return NO;
        }
        return YES;
    }
    context = UIGraphicsGetCurrentContext();
	if( context )
		CGContextSetLineJoin(context, kCGLineJoinRound);

    CGPoint pagePoint = CGPointMake(self.startPoint.x, self.startPoint.y);
    
	
	m_free_hand_points = [[NSMutableArray alloc] initWithCapacity:50];
	
	[m_free_hand_points addObject:[NSValue valueWithCGPoint:pagePoint]];

	if( !m_free_hand_strokes )
	{
		m_startPageNum = self.pageNumber;
		
		m_free_hand_strokes = [[NSMutableArray alloc] init];
		
		m_redo_strokes = [[NSMutableArray alloc] init];
		
		self.frame = CGRectMake([self.pdfViewCtrl GetHScrollPos], [self.pdfViewCtrl GetVScrollPos], self.pdfViewCtrl.bounds.size.width, self.pdfViewCtrl.bounds.size.height);
	}
    
    if( !m_page_space_redo_strokes )
    {
        m_page_space_redo_strokes = [[NSMutableArray alloc] init];
    }
	
	[m_free_hand_strokes addObject:m_free_hand_points];
    
    if( !m_page_space_free_hand_strokes )
        m_page_space_free_hand_strokes = [[NSMutableArray alloc] init];
    
    if( !m_page_space_free_hand_points )
        m_page_space_free_hand_points = [[NSMutableArray alloc] init];
    
    [m_page_space_free_hand_strokes addObject:m_page_space_free_hand_points];
    
    return YES;
}

- (BOOL)pdfViewCtrl:(PTPDFViewCtrl*)pdfViewCtrl onTouchesMoved:(NSSet *)touches withEvent:(UIEvent *)event
{
	if (self.pageNumber < 1 || m_startPageNum != self.pageNumber) {
        return YES;
    }
	
    [self addFreehandPoint:touches];
    
    return YES;
}

-(void)saveRedoStrokesInPageSpace
{

    m_page_space_free_hand_points = [[NSMutableArray alloc] init];

    NSArray<NSValue*>* point_array = m_redo_strokes.lastObject;

    for (NSValue* val in point_array)
    {
        CGPoint screenPoint = val.CGPointValue;
        CGPoint pagePoint = [self convertScreenPtToPagePt:screenPoint onPageNumber:self.pageNumber];
        [m_page_space_free_hand_points addObject:[NSValue valueWithCGPoint:pagePoint]];
    }
    
    [m_page_space_redo_strokes addObject:m_page_space_free_hand_points];
    
    m_page_space_free_hand_points = [[NSMutableArray alloc] init];
    
}

-(void)restorePointsInScreenSpace
{
    [m_free_hand_points removeAllObjects];
    [m_free_hand_strokes removeAllObjects];
    
    for (NSArray<NSValue*>* point_array in m_page_space_free_hand_strokes)
    {
        for (NSValue* val in point_array)
        {
            CGPoint pagePoint = val.CGPointValue;
            CGPoint screenPoint = [self convertPagePtToScreenPt:pagePoint onPageNumber:self.pageNumber];
            [m_free_hand_points addObject:[NSValue valueWithCGPoint:screenPoint]];
        }
        
        [m_free_hand_strokes addObject:m_free_hand_points];
        
        m_free_hand_points = [[NSMutableArray alloc] init];
    }
    
    [m_redo_strokes removeAllObjects];
    
    NSMutableArray* redoPointArray = [[NSMutableArray alloc] init];
    for (NSArray<NSValue*>* point_array in m_page_space_redo_strokes)
    {
        for (NSValue* val in point_array)
        {
            CGPoint pagePoint = val.CGPointValue;
            CGPoint screenPoint = [self convertPagePtToScreenPt:pagePoint onPageNumber:self.pageNumber];
            [redoPointArray addObject:[NSValue valueWithCGPoint:screenPoint]];
        }
        
        [m_redo_strokes addObject:redoPointArray];
        
        redoPointArray = [[NSMutableArray alloc] init];
    }
}

- (BOOL)pdfViewCtrl:(PTPDFViewCtrl*)pdfViewCtrl onTouchesEnded:(NSSet *)touches withEvent:(UIEvent *)event
{

    if (pencilMode) {
        [stylusAppendTimeout invalidate];
        stylusAppendTimeout = [PTTimer scheduledTimerWithTimeInterval:PTFreeHandCreateStylusAppendInterval target:self selector:@selector(endDrawing) userInfo:nil repeats:NO];
    }
	if (self.pageNumber < 1 || m_startPageNum != self.pageNumber) {
        return YES;
    }
    
    if( m_free_hand_points.count == 1)
    {
        [self addFreehandPoint:touches];
    }
    
    [self saveRedoStrokesInPageSpace];
    
	if( !multistrokeMode && !pencilMode)
	{
		[self commitAnnotation];
		
		if( self.backToPanToolAfterUse )
		{
			self.nextToolType = [PTPanTool class];
			return NO;
		}
		else
			return YES;
	}
	else
	{
		[m_redo_strokes removeAllObjects];
        [m_page_space_redo_strokes removeAllObjects];
		
		if( [self.delegate respondsToSelector:@selector(strokeAdded:)] )
		{
			[self.delegate strokeAdded:self];
		}
		
		return YES;
	}
}

-(PTPDFRect *)rectFromPoints
{
    PTPDFRect* myRect = [[PTPDFRect alloc] init];;

    double minX = RAND_MAX, minY = RAND_MAX, maxX = -RAND_MAX, maxY = -RAND_MAX;

    int numStrokes = 0;

    for (NSMutableArray* point_array in m_free_hand_strokes)
    {

        for (NSValue* point in point_array)
        {
            CGPoint aPoint = point.CGPointValue;

            PTPDFPoint* pdfPoint = [[PTPDFPoint alloc] init];

            [self ConvertScreenPtToPagePtX:&aPoint.x Y:&aPoint.y PageNumber:m_startPageNum];

            [pdfPoint setX:aPoint.x];
            [pdfPoint setY:aPoint.y];

            minX = MIN(minX, aPoint.x);
            minY = MIN(minY, aPoint.y);

            maxX = MAX(maxX, aPoint.x);
            maxY = MAX(maxY, aPoint.y);
        }
        numStrokes++;
    }

    [myRect SetX1:minX];
    [myRect SetY1:minY];

    [myRect SetX2:maxX];
    [myRect SetY2:maxY];

    return myRect;
}

-(void)endDrawing
{
    [self commitAnnotation];
    pencilMode = NO;
    multistrokeMode = NO;
    if (self.backToPanToolAfterUse) {
        [self.toolManager changeTool:[PTPanTool class]];
    }
}

- (PTInk *)createAnnotation
{
    PTPDFDoc *doc = [self.pdfViewCtrl GetDoc];
    PTInk *annot = [PTInk Create:(PTSDFDoc*)doc pos:[[PTPDFRect alloc] init]];

    [super setPropertiesFromAnnotation:annot];

    return annot;
}

-(void)commitAnnotation
{
    if (m_free_hand_strokes.count < 1) {
        return;
    }

	if( self.pageNumber > 0 )
	{
		[self keepToolAppearanceOnScreen];

        PTPDFDoc* doc = [self.pdfViewCtrl GetDoc];

		PTInk* inkAnnot;
		
		@try
		{
			[self.pdfViewCtrl DocLock:YES];
			
			PTPage* pg = [doc GetPage:m_startPageNum];
			
			if( [pg IsValid] )
			{
				
				PTPDFRect* myRect = [[PTPDFRect alloc] init];
				
				inkAnnot = [self createAnnotation];
				
				int i = 0;
				
				double minX = RAND_MAX, minY = RAND_MAX, maxX = -RAND_MAX, maxY = -RAND_MAX;
				
				int numStrokes = 0;
				
				for (NSMutableArray* point_array in m_free_hand_strokes)
				{

					for (NSValue* point in point_array)
					{
						CGPoint aPoint = point.CGPointValue;
						
						PTPDFPoint* pdfPoint = [[PTPDFPoint alloc] init];
						
						[self ConvertScreenPtToPagePtX:&aPoint.x Y:&aPoint.y PageNumber:m_startPageNum];
						
						[pdfPoint setX:aPoint.x];
						[pdfPoint setY:aPoint.y];
						
						[inkAnnot SetPoint:numStrokes pointindex:i++ pt:pdfPoint];
						
						minX = MIN(minX, aPoint.x);
						minY = MIN(minY, aPoint.y);

						maxX = MAX(maxX, aPoint.x);
						maxY = MAX(maxY, aPoint.y);
					}
					numStrokes++;
					i = 0;
				}

				[myRect SetX1:minX];
				[myRect SetY1:minY];
				
				[myRect SetX2:maxX];
				[myRect SetY2:maxY];
				
				[inkAnnot SetRect:myRect];
				
				[inkAnnot RefreshAppearance];
				
				if( self.annotationAuthor && self.annotationAuthor.length > 0 && [inkAnnot isKindOfClass:[PTMarkup class]]	)
				{
					[(PTMarkup*)inkAnnot SetTitle:self.annotationAuthor];
				}
			
				[pg AnnotPushBack:inkAnnot];
			}
			
		}
		@catch (NSException *exception) {
			NSLog(@"Exception: %@: %@",exception.name, exception.reason);
		}
		@finally {
			[self.pdfViewCtrl DocUnlock];
		}
		
		m_free_hand_points = 0;
        
		
		[self.pdfViewCtrl UpdateWithAnnot:inkAnnot page_num:m_startPageNum];
		[self.pdfViewCtrl RequestRendering];
		
		[self annotationAdded:inkAnnot onPageNumber:self.pageNumber];
		
	}
	
    [m_page_space_free_hand_strokes removeAllObjects];
	[m_free_hand_strokes removeAllObjects];
	[m_redo_strokes removeAllObjects];
	[m_free_hand_points removeAllObjects];
	[self setNeedsDisplay];
}

- (BOOL)pdfViewCtrl:(PTPDFViewCtrl*)pdfViewCtrl handleTap:(UITapGestureRecognizer *)sender
{
	if( !multistrokeMode )
	{
		return [super pdfViewCtrl:pdfViewCtrl handleTap:sender];
	}
	
	[self setNeedsDisplay];
	return YES;
}

-(void)pdfViewCtrlOnLayoutChanged:(PTPDFViewCtrl*)pdfViewCtrl
{
    [self restorePointsInScreenSpace];
    
    if( !multistrokeMode )
    {
        [self commitAnnotation];
    }

    [self setNeedsDisplay];
    
    [super pdfViewCtrlOnLayoutChanged:pdfViewCtrl];
}

- (BOOL)requiresEditSupport
{
    
    return YES;
}

#pragma mark - multistroke undo/redo methods

-(void)undoStroke
{
	if( m_free_hand_strokes.count )
	{
		[m_redo_strokes addObject:m_free_hand_strokes.lastObject];
        [m_page_space_redo_strokes addObject:m_page_space_free_hand_strokes.lastObject];
		[m_free_hand_strokes removeLastObject];
        [m_page_space_free_hand_strokes removeLastObject];
		[self setNeedsDisplay];
	}
}

-(void)redoStroke
{
	if( m_redo_strokes.count )
	{
		[m_free_hand_strokes addObject:m_redo_strokes.lastObject];
        [m_page_space_free_hand_strokes addObject:m_page_space_redo_strokes.lastObject];
		[m_redo_strokes removeLastObject];
        [m_page_space_redo_strokes removeLastObject];
		[self setNeedsDisplay];
	}
}

-(BOOL)canUndoStroke
{
	return m_free_hand_strokes.count > 0 ? YES : NO;
}

-(BOOL)canRedoStroke
{
	return m_redo_strokes.count > 0 ? YES : NO;
}

- (BOOL)onSwitchToolEvent:(id)userData
{
    if (pencilMode && m_free_hand_points.count > 0) {
        [self endDrawing];
    }
    return YES;
}

@end
