//---------------------------------------------------------------------------------------
// Copyright (c) 2001-2020 by PDFTron Systems Inc. All Rights Reserved.
// Consult legal.txt regarding legal and license information.
//---------------------------------------------------------------------------------------

#import "ToolsConfig.h"
#import "PTFreeTextCreate.h"

#import "PTPanTool.h"
#import "PTColorDefaults.h"
#import "CGGeometry+PTAdditions.h"

static const CGFloat PTFreeTextDragThreshold = 10.0;

@interface PTFreeTextCreate () <UIGestureRecognizerDelegate>
{
    CGPoint firstTouchPoint;
    CGPoint startPoint;
    CGPoint endPoint;
    UITextView* textView;
    UIScrollView* sv;
    BOOL wroteAnnot;
    BOOL created;
    BOOL keyboardOnScreen;
}

@property (assign, nonatomic) BOOL isDrag;

@property (strong, nonatomic) NSDate* touchesEndedTime;

@property (nonatomic, strong) PTPDFRect* writeRect;
@property (nonatomic, strong) PTPage* writePage;
/**
 * Defines if the text box should orient itself for right-to-left language input.
 * Switches based on language that is first used when creating the annotation
 */
@property (assign, nonatomic) BOOL isRTL;

@end

@implementation PTFreeTextCreate


- (instancetype)initWithPDFViewCtrl:(PTPDFViewCtrl*)in_pdfViewCtrl
{
    
    self = [super initWithPDFViewCtrl:in_pdfViewCtrl];
    if (self) {
		self.backgroundColor = [UIColor clearColor];
        
        self.isRTL = false;
        
        // use a temporary textView to see if the user is about to write in a RTL or LTR language
        UITextView* textView = [[UITextView alloc] initWithFrame:CGRectZero];
        [in_pdfViewCtrl addSubview:textView];
        [textView becomeFirstResponder];


        if( [NSLocale characterDirectionForLanguage:[textView textInputMode].primaryLanguage] == NSLocaleLanguageDirectionRightToLeft)
        {
            self.isRTL = true;
        }
        
        [textView resignFirstResponder];
        
        [textView removeFromSuperview];
        
        [[NSNotificationCenter defaultCenter] addObserver:self
                                                 selector:@selector (UITextInputCurrentInputModeDidChangeNotification:)
                                                     name: UITextInputCurrentInputModeDidChangeNotification object:nil];
        
    }
    
    return self;
}

-(Class)annotClass
{
    return [PTFreeText class];
}

+(PTExtendedAnnotType)annotType
{
    return PTExtendedAnnotTypeFreeText;
}

-(BOOL)canEditStyle
{
    return YES;
}

+(BOOL)createsAnnotation
{
	return YES;
}

- (void)dealloc
{
    [[NSNotificationCenter defaultCenter] removeObserver:self];
}

-(void)UITextInputCurrentInputModeDidChangeNotification:(NSNotification*)notification
{
    if( textView.text.length == 0 )
    {
        
        bool prior = self.isRTL;
        
        if( [NSLocale characterDirectionForLanguage:[textView textInputMode].primaryLanguage] == NSLocaleLanguageDirectionRightToLeft)
        {
            self.isRTL = true;
        }
        else
        {
            self.isRTL = false;
        }
        
        if( prior != self.isRTL )
        {
            [self setTextAreaFrame];
        }
    }
}

@synthesize textView = textView;

-(void)setUpTextEntry
{
	PTColorPostProcessMode mode = [self.pdfViewCtrl GetColorPostProcessMode];
    textView = [[UITextView alloc] init];
    sv = [[UIScrollView alloc] init];
    
    textView.contentInset = UIEdgeInsetsZero;
    textView.textContainerInset = UIEdgeInsetsZero;
    textView.textContainer.lineFragmentPadding = 0;
    
    sv.autoresizingMask = UIViewAutoresizingFlexibleWidth | UIViewAutoresizingFlexibleHeight;
    textView.autoresizingMask = UIViewAutoresizingFlexibleWidth | UIViewAutoresizingFlexibleHeight;
    
    [sv addSubview:textView];
    [self addSubview:sv];
    
    [[NSNotificationCenter defaultCenter] addObserver:self
                                             selector:@selector (keyboardWillShow:)
                                                 name: UIKeyboardWillShowNotification object:nil];
    
    sv.backgroundColor = [UIColor clearColor];
	
    self.userInteractionEnabled = YES;
    sv.userInteractionEnabled = YES;
    textView.userInteractionEnabled = YES;
    textView.delegate = self;
    textView.textColor = [PTColorDefaults defaultColorForAnnotType:self.annotType attribute:ATTRIBUTE_TEXT_COLOR colorPostProcessMode:mode];
    
    NSString* fontName = self.toolManager.freeTextAnnotationOptions.defaultFontName ? self.toolManager.freeTextAnnotationOptions.defaultFontName : @"Helvetica";
    
    textView.font = [UIFont fontWithName:fontName size:[PTColorDefaults defaultFreeTextSize]*[self.pdfViewCtrl GetZoom]];
    textView.backgroundColor = [UIColor clearColor];

	// one day add background colour to live editing too?
	//http://stackoverflow.com/questions/15438869/uitextview-text-background-colour
	
    created = NO;
    
    wroteAnnot = NO;

    return;
}

+(void)setRectForFreeText:(PTFreeText*)freeText withRect:(PTPDFRect*)rect pdfViewCtrl:(PTPDFViewCtrl*)pdfViewCtrl isRTL:(BOOL)isRTL
{
    // Get the annotation's content stream
    PTObj* contentStream = [[[freeText GetSDFObj] FindObj:@"AP"] FindObj:@"N"];
    
    if( !contentStream )
    {
        // unexpected failure
        // bbox will not be tight
        return;
    }
    
    // use element reader to iterate through elements and union their bounding boxes
    PTElementReader* er = [[PTElementReader alloc] init];
    
    PTObj* dict = [[PTObj alloc] init];
    
    PTContext* context = [[PTContext alloc] init];
    
    [er ReaderBeginWithSDFObj:contentStream resource_dict:dict ocg_context:context];
    
    PTElement *element;
    
    PTPDFRect* unionRect = 0;
    
    for (element=[er Next]; element != NULL; element = [er Next])
    {
        PTPDFRect* elementRect = [element GetBBox];
        
        if([element GetType] == e_pttext_obj)
        {
            if( [elementRect Width] && [elementRect Height] )
            {
                if( unionRect == 0 )
                    unionRect = elementRect;
                unionRect = [PTTool GetRectUnion:unionRect Rect2:elementRect];
            }
        }
        
    }
    
    double width = fabs([unionRect GetX2] - [unionRect GetX1])+10;
    double height = fabs([unionRect GetY2] - [unionRect GetY1])+10;
    
    PTRotate ctrlRotation = [pdfViewCtrl GetRotation];
    PTRotate pageRotation = [[freeText GetPage] GetRotation];
    int annotRotation = ((pageRotation + ctrlRotation) % 4) * 90;

    if(  !isRTL )
    {
        if (ctrlRotation == e_pt90 || ctrlRotation == e_pt270) {
            // Swap width and height if pdfViewCtrl is rotated 90 or 270 degrees
            width = width + height;
            height = width - height;
            width = width - height;
        }

        [unionRect SetX1:[rect GetX1]];
        [unionRect SetY1:[rect GetY1]];
        
        if( annotRotation == 0 )
        {
            [unionRect SetX2:[rect GetX1]+width];
            [unionRect SetY2:[rect GetY1]-height];
        }
        else if( annotRotation == 90 )
        {
            [unionRect SetX2:[rect GetX1]+width];
            [unionRect SetY2:[rect GetY1]+height];
        }
        else if(annotRotation == 180)
        {
            [unionRect SetX2:[rect GetX1]-width];
            [unionRect SetY2:[rect GetY1]+height];
        }
        else if(annotRotation == 270)
        {
            [unionRect SetX2:[rect GetX1]-width];
            [unionRect SetY2:[rect GetY1]-height];
        }
        
        
        
    }
    else
    {
        const int rightAligned = 2;
        
        [freeText SetQuaddingFormat:rightAligned];
        if (pageRotation == e_pt90 || pageRotation == e_pt270) {
            // Swap width and height if page is rotated 90 or 270 degrees
            width = width + height;
            height = width - height;
            width = width - height;
        }

        if( annotRotation == 0 )
        {
            [unionRect SetX1:[rect GetX2]-width];
            [unionRect SetY1:[rect GetY1]-height];
            
            [unionRect SetX2:[rect GetX2]];
            [unionRect SetY2:[rect GetY1]];
        }
        else if( annotRotation == 90 )
        {
            
            [unionRect SetX1:[rect GetX1]];
            [unionRect SetY1:[rect GetY2]-width];
            
            [unionRect SetX2:[rect GetX1]+height];
            [unionRect SetY2:[rect GetY2]];
            
        }
        else if( annotRotation == 180 )
        {
            [unionRect SetX1:[rect GetX2]+width];
            [unionRect SetY1:[rect GetY1]+height];
            
            [unionRect SetX2:[rect GetX2]];
            [unionRect SetY2:[rect GetY1]];
        }
        else if( annotRotation == 270 )
        {
            [unionRect SetX1:[rect GetX1]-height];
            [unionRect SetY1:[rect GetY2]];
            
            [unionRect SetX2:[rect GetX1]];
            [unionRect SetY2:[rect GetY2]+width];
        }
        
        
    }
    
    [unionRect Normalize];
    
    [freeText Resize:unionRect];
    
    [freeText SetRotation:annotRotation];
}

-(BOOL)setTextAreaFrame
{
    startPoint = self.longPressPoint;
    
    int pageNumber = [self.pdfViewCtrl GetPageNumberFromScreenPt:startPoint.x y:startPoint.y];
	
    if (pageNumber < 1 )
      return NO;

    assert(startPoint.x > 0);
    assert(startPoint.y > 0);
	
    PTPDFRect* testRect = [[[self.pdfViewCtrl GetDoc] GetPage:pageNumber] GetCropBox];
	
    CGFloat cbx1 = [testRect GetX1];
    CGFloat cbx2 = [testRect GetX2];
    CGFloat cby1 = [testRect GetY1];
    CGFloat cby2 = [testRect GetY2];
	
    CGFloat pageLeft;
    CGFloat pageTop;
	
    [self ConvertPagePtToScreenPtX:&cbx1 Y:&cby1 PageNumber:pageNumber];
    [self ConvertPagePtToScreenPtX:&cbx2 Y:&cby2 PageNumber:pageNumber];
    
    //double thickness = [PTColorDefaults defaultBorderThicknessForAnnotType:self.annotType];
    double offset;// = thickness * [self.pdfViewCtrl GetZoom];
    offset = 0;
    
    if( self.isRTL == false )
    {
        pageLeft = MIN(cbx1, cbx2);
        pageTop =  MIN(cby1, cby2);
        endPoint.x = MAX(cbx1, cbx2);
        endPoint.y = MAX(cby1, cby2);
        
        startPoint.x = MAX(pageLeft, startPoint.x-offset);
        startPoint.y = MAX(pageTop, startPoint.y-offset);
    }
    else
    {
        pageLeft = MIN(cbx1, cbx2);
        pageTop =  MIN(cby1, cby2);
        
        endPoint.x = startPoint.x+offset;
        endPoint.y = MAX(cby1, cby2);
        
        startPoint.x = pageLeft;
        startPoint.y = MAX(pageTop, startPoint.y+offset);
        
    }
	
	
	self.frame = CGRectMake([self.pdfViewCtrl GetHScrollPos]+startPoint.x, [self.pdfViewCtrl GetVScrollPos]+startPoint.y, fabs(endPoint.x-startPoint.x), fabs(endPoint.y-startPoint.y));

	return YES;
}

-(void)activateTextEntry
{
    assert([NSThread isMainThread]);
    
    CGRect cursorRect = [textView caretRectForPosition:textView.selectedTextRange.start];
    double thickness = [PTColorDefaults defaultBorderThicknessForAnnotType:self.annotType] * [self.pdfViewCtrl GetZoom];
    double inset = thickness*2;
    double topOffset = textView.font.lineHeight-textView.font.pointSize;
    textView.textContainerInset = UIEdgeInsetsMake(inset-topOffset, inset, inset, inset);
    //Move and resize the frame by the width of the border (and the additional padding inside the border)
    sv.frame = CGRectMake(-inset, -inset-cursorRect.size.height/2, self.frame.size.width-inset, self.frame.size.height-inset-cursorRect.size.height/2);
    textView.frame = CGRectMake(0, 0, sv.frame.size.width, sv.frame.size.height);
    
    if ([textView canBecomeFirstResponder]) {

        
        [textView becomeFirstResponder];

    }
    
    created = YES;
    
    UITapGestureRecognizer *tgr = [[UITapGestureRecognizer alloc] initWithTarget:self action:@selector(textTap:)];
    tgr.delegate = self;
    
    NSUInteger numTaps = 1;
    
    tgr.numberOfTapsRequired = numTaps;
    
    [tgr setCancelsTouchesInView:YES];
    [tgr setDelaysTouchesEnded:NO];
    
    [textView addGestureRecognizer:tgr];
    
    PTPDFDoc* doc = [self.pdfViewCtrl GetDoc];
    
    [self.pdfViewCtrl DocLock:YES];
    
    @try {
        int pageNumber = [self.pdfViewCtrl GetPageNumberFromScreenPt:startPoint.x y:startPoint.y];
        
        if (pageNumber < 1)
            pageNumber = [self.pdfViewCtrl GetPageNumberFromScreenPt:endPoint.x y:endPoint.y];
        
        if (pageNumber < 1)
        {
            self.nextToolType = [PTPanTool class];
            
            return;
        }
        
        self.annotationPageNumber = pageNumber;
        
        self.writePage = [doc GetPage:pageNumber];

        
    }
    @catch (NSException *exception) {
        NSLog(@"activateTextEntry Exception: %@: %@",exception.name, exception.reason);
    }
    @finally {
        [self.pdfViewCtrl DocUnlock];
    }
}

- (PTFreeText *)createFreeText
{
    PTPDFDoc* doc = [self.pdfViewCtrl GetDoc];
    
    return [PTFreeText Create:[doc GetSDFDoc] pos:self.writeRect];
}

- (void)commitAnnotation
{
    if(wroteAnnot || !textView || [textView.text isEqualToString:@""])
        return;
    
    PTFreeText* annotation;
	    
    @try
    {
        [self.pdfViewCtrl DocLock:YES];
        
        CGRect screenRect = [self.pdfViewCtrl convertRect:textView.frame fromView:textView.superview];
        
        CGFloat x1 = screenRect.origin.x;
        CGFloat y1 = screenRect.origin.y;
        CGFloat x2 = screenRect.origin.x+screenRect.size.width;
        CGFloat y2 = screenRect.origin.y+screenRect.size.height;
        
        [self ConvertScreenPtToPagePtX:&x1 Y:&y1 PageNumber:self.annotationPageNumber];
        [self ConvertScreenPtToPagePtX:&x2 Y:&y2 PageNumber:self.annotationPageNumber];
        
        self.writeRect = [[PTPDFRect alloc] initWithX1:x1 y1:y1 x2:x2 y2:y2];

        // Create a new PTFreeText annotation.
        annotation = [self createFreeText];
		
		[annotation SetFontSize:[PTColorDefaults defaultFreeTextSize]];

        [annotation SetContents:textView.text];
        
        if( self.toolManager.freeTextAnnotationOptions.defaultFontName )
        {
            [annotation setFontWithName:self.toolManager.freeTextAnnotationOptions.defaultFontName pdfDoc:[self.pdfViewCtrl GetDoc]];
        }
        [annotation RefreshAppearance];
		
		if (self.annotationAuthor.length > 0) {
			[annotation SetTitle:self.annotationAuthor];
		}
        
        // push back annotation now in case of rotated page
        [self.writePage AnnotPushBack:annotation];
        [annotation RefreshAppearance];
        
        [self setRectForFreeText:annotation];
        
        // Apply annotation colors, border, etc.
        [self setPropertiesForFreeText:annotation];
        
        [annotation RefreshAppearance];
        
        wroteAnnot = YES;
    }
    @catch (NSException *exception) {
        NSLog(@"Exception: %@: %@",exception.name, exception.reason);
    }
    @finally {
        [self.pdfViewCtrl DocUnlock];
    }
    
    if( self.annotationPageNumber > 0 )
        [self.pdfViewCtrl UpdateWithAnnot:annotation page_num:self.annotationPageNumber];
	
	[self annotationAdded:annotation onPageNumber:self.annotationPageNumber];
    
    [textView removeFromSuperview];
    textView = nil;
    [sv removeFromSuperview];
    sv = nil;
}

- (void)setRectForFreeText:(PTFreeText *)freeText
{
    [PTFreeTextCreate setRectForFreeText:freeText
                                withRect:self.writeRect
                             pdfViewCtrl:self.pdfViewCtrl
                                   isRTL:self.isRTL];
}

- (void)setPropertiesForFreeText:(PTFreeText *)freeText
{
    // write text colour
    PTColorPt* cp = [PTColorDefaults defaultColorPtForAnnotType:self.annotType attribute:ATTRIBUTE_TEXT_COLOR  colorPostProcessMode:e_ptpostprocess_none];
    int numCompsText = [PTColorDefaults numCompsInColorPtForAnnotType:self.annotType attribute:ATTRIBUTE_TEXT_COLOR];
    
    if( numCompsText > 0 ) {
        [freeText SetTextColor:cp col_comp:3];
    } else {
        [freeText SetTextColor:cp col_comp:0];
    }
    
    PTColorPt* strokeColor = [PTColorDefaults defaultColorPtForAnnotType:self.annotType attribute:ATTRIBUTE_STROKE_COLOR colorPostProcessMode:e_ptpostprocess_none];
    int numCompsStroke = [PTColorDefaults numCompsInColorPtForAnnotType:self.annotType attribute:ATTRIBUTE_STROKE_COLOR];
    [freeText SetLineColor:strokeColor col_comp:numCompsStroke];
    
    // write border colour
    double thickness = [PTColorDefaults defaultBorderThicknessForAnnotType:self.annotType];
    PTBorderStyle* bs = [freeText GetBorderStyle];
    
    
    if(numCompsStroke == 0){
        thickness = 0.0;
    }
    
    [bs SetWidth:thickness];
    [freeText SetBorderStyle:bs oldStyleOnly:NO];

    // write fill colour
    int numCompsFill = [PTColorDefaults numCompsInColorPtForAnnotType:self.annotType attribute:ATTRIBUTE_FILL_COLOR];
    cp = [PTColorDefaults defaultColorPtForAnnotType:self.annotType attribute:ATTRIBUTE_FILL_COLOR colorPostProcessMode:e_ptpostprocess_none];
    
    if( numCompsFill > 0 ) {
        [freeText SetColor:cp numcomp:3];
    } else {
        [freeText SetColor:cp numcomp:0];
    }
}

- (void)textViewDidEndEditing:(UITextView *)textView
{
	self.isDrag = NO;
    [self commitAnnotation];
}

- (BOOL)textFieldShouldBeginEditing:(UITextField *)textField
{
    return YES;
}

-(void)keyboardWillShow:(NSNotification *)notification
{
    CGRect cursorRect = [textView caretRectForPosition:textView.selectedTextRange.start];
    CGRect cursorScreenRect = [self.pdfViewCtrl convertRect:cursorRect fromView:textView];
    cursorScreenRect.origin.x += [self.pdfViewCtrl GetHScrollPos];
    cursorScreenRect.origin.y += [self.pdfViewCtrl GetVScrollPos];

    CGFloat topEdge = 0.0;
    if (@available(iOS 11.0, *)) {
        topEdge = self.pdfViewCtrl.safeAreaInsets.top;
    }
    
    [self.pdfViewCtrl keyboardWillShow:notification rectToNotOverlapWith:cursorScreenRect topEdge:topEdge];
    [[NSNotificationCenter defaultCenter] addObserver:self
                                             selector:@selector (keyboardWillHide:)
                                                 name: UIKeyboardWillHideNotification object:nil];

    keyboardOnScreen = true;
}

- (BOOL)onSwitchToolEvent:(id)userData
{
	NSString* start = (NSString*)userData;
    if( userData && ((![start isEqualToString:@"Start"] && self.backToPanToolAfterUse) || (self.backToPanToolAfterUse && [start isEqualToString:@"BackToPan"])))
    {
        self.nextToolType = [PTPanTool class];
        return NO;
    }
	else if( [start isEqualToString:@"BackToPan"] && !self.backToPanToolAfterUse )
	{
		self.nextToolType = [self class];
		return NO;
    }else if( [start isEqualToString:@"CloseAnnotationToolbar"] && !self.backToPanToolAfterUse )
    {
        return NO;
    }
    else
    {
        created = NO;

		startPoint = self.longPressPoint;

		[self setTextAreaFrame];
        [self setUpTextEntry];
		[self activateTextEntry];

        return YES;
    }
}

-(void)keyboardWillHide:(NSNotification *)notification
{
    
    created = NO;
    
	[self.pdfViewCtrl keyboardWillHide:notification];
    
    keyboardOnScreen = false;

    [NSNotificationCenter.defaultCenter removeObserver:self name:UIKeyboardWillShowNotification object:nil];
    
    [NSNotificationCenter.defaultCenter removeObserver:self name:UIKeyboardWillHideNotification object:nil];
    
	[self.toolManager createSwitchToolEvent:@"BackToPan"];
}

-(BOOL)createEntryAtPoint:(CGPoint)point
{
    assert([NSThread isMainThread]);
	[self.superview bringSubviewToFront:self];
	startPoint = point;
    self.longPressPoint = point;
	BOOL couldSetFrame = [self setTextAreaFrame];
	
	if( !couldSetFrame )
	{
		self.nextToolType = [PTPanTool class];
        return NO;
	}
	else
	{
		if( textView == 0 && sv == 0 )
			[self setUpTextEntry];
		
		[self activateTextEntry];
        
        return YES;
	}
}

- (BOOL)pdfViewCtrl:(PTPDFViewCtrl*)pdfViewCtrl handleTap:(UITapGestureRecognizer *)sender
{
    if( created == NO )
    {
        self.isDrag = NO;
        [self createEntryAtPoint:[sender locationOfTouch:0 inView:self.pdfViewCtrl]];
	
        return YES;
    }
    else
    {
        if( textView != nil )
		{
            // UI gesture recognizer may also come in after the touches ended. This is to prevent
            // one from firing after the other, causing the keyboard to popup and then immediately
            // dismiss.
            if( self.isDrag && [[NSDate date] timeIntervalSinceDate:self.touchesEndedTime] < 0.85 )
                return YES;

            [textView resignFirstResponder];
            self.frame = CGRectZero;
            
		}
		
		if( self.backToPanToolAfterUse )
		{
			self.nextToolType = [PTPanTool class];
			return NO;
		}
		else
		{
			created = NO;
			return YES;
		}
    }
}


- (void)textTap:(UITapGestureRecognizer *)gestureRecognizer { 

	CGSize textSize = [textView.text boundingRectWithSize:textView.bounds.size options:(NSStringDrawingUsesLineFragmentOrigin|NSStringDrawingUsesFontLeading) attributes:@{ NSFontAttributeName : [UIFont fontWithName:textView.font.fontName size:[PTColorDefaults defaultFreeTextSize]*[self.pdfViewCtrl GetZoom]] } context:nil].size;
	
    CGPoint down = [gestureRecognizer locationInView:textView];
    
    // 6 for textview inset
    if (down.x-6 < textSize.width && down.y-6 < textSize.height) {
        // tap was on text so ignore
    }
    else
    {
        // tap was outside of text, dismiss textview and commit annotation
        [self pdfViewCtrl:self.pdfViewCtrl handleTap:gestureRecognizer];
    }
    
}

- (BOOL)gestureRecognizer:(UIGestureRecognizer *)gestureRecognizer shouldRecognizeSimultaneouslyWithGestureRecognizer:(UIGestureRecognizer *)otherGestureRecognizer
{
    // required for the textview gesture recognizer to fire
    return YES;
}

- (BOOL)pdfViewCtrl:(PTPDFViewCtrl*)pdfViewCtrl touchesShouldCancelInContentView:(UIView *)view
{
    return [super pdfViewCtrl:pdfViewCtrl touchesShouldCancelInContentView: view];
}

- (BOOL)pdfViewCtrl:(PTPDFViewCtrl *)pdfViewCtrl onTouchesBegan:(NSSet<UITouch *> *)touches withEvent:(UIEvent *)event
{
    UITouch *touch = touches.allObjects[0];
    firstTouchPoint = [touch locationInView:self.pdfViewCtrl];
    return YES;
}

- (BOOL)pdfViewCtrl:(PTPDFViewCtrl*)pdfViewCtrl onTouchesMoved:(NSSet *)touches withEvent:(UIEvent *)event
{
    UITouch *touch = touches.allObjects[0];
    CGPoint down = [touch locationInView:self.pdfViewCtrl];
    // Only classify touch as a drag if the touch point is over a certain distance threshold from the initial touch point
    self.isDrag = PTCGPointDistanceToPoint(down, firstTouchPoint) > PTFreeTextDragThreshold;
    self.touchesEndedTime = [NSDate date];
    return YES;
}


- (BOOL)pdfViewCtrl:(PTPDFViewCtrl*)pdfViewCtrl onTouchesEnded:(NSSet *)touches withEvent:(UIEvent *)event
{
	if( !self.isDrag )
		return YES;
	
    UITouch *touch = touches.allObjects[0];

    
	if( created == NO )
        [self createEntryAtPoint:[touch locationInView:self.pdfViewCtrl]];

    created = YES;

    if( self.backToPanToolAfterUse )
    {
        self.nextToolType = [PTPanTool class];
        return NO;
    }
    else
        return YES;
}

- (BOOL)pdfViewCtrl:(PTPDFViewCtrl*)pdfViewCtrl onTouchesCancelled:(NSSet *)touches withEvent:(UIEvent *)event
{
	return YES;
}



@end
