//---------------------------------------------------------------------------------------
// Copyright (c) 2001-2020 by PDFTron Systems Inc. All Rights Reserved.
// Consult legal.txt regarding legal and license information.
//---------------------------------------------------------------------------------------

#import "ToolsConfig.h"
#import "PTToolsUtil.h"
#import "PTAnnotStyle.h"



static NSString * const PTToolsBundleIdentifier = @"com.pdftron.Tools";

/**
 * This key is used as the default localized string value when checking if a string definition has
 * been overridden in the main (app) bundle, in the "PTLocalizable" string table. If this value is
 * returned from that check then it means that the string definition was not found and the Tools
 * bundle should be searched next.
 *
 * NOTE: When the NSShowNonLocalizedStrings user default is set, the return value is capitalized when
 * the key is not found in the "PTLocalizable" string table. To handle this case, the value of this
 * constant is always capitalized so that the missing-key case can still be detected.
 */
static NSString * const PT_LocalizedStringNotFound = @"PT_LOCALIZEDSTRINGNOTFOUND";

@implementation PTToolsUtil

static NSBundle *PTToolsUtil_toolsBundle;

static UIImage* makeImageNegative(UIImage* originalImage)
{
    UIGraphicsBeginImageContext(originalImage.size);
    CGContextSetBlendMode(UIGraphicsGetCurrentContext(), kCGBlendModeCopy);
    CGRect imageRect = CGRectMake(0, 0, originalImage.size.width, originalImage.size.height);
    [originalImage drawInRect:imageRect];
    
    
    CGContextSetBlendMode(UIGraphicsGetCurrentContext(), kCGBlendModeDifference);
    // translate/flip the graphics context (for transforming from CG* coords to UI* coords
    CGContextTranslateCTM(UIGraphicsGetCurrentContext(), 0, originalImage.size.height);
    CGContextScaleCTM(UIGraphicsGetCurrentContext(), 1.0, -1.0);
    //mask the image
    CGContextClipToMask(UIGraphicsGetCurrentContext(), imageRect,  originalImage.CGImage);
    CGContextSetFillColorWithColor(UIGraphicsGetCurrentContext(),[UIColor whiteColor].CGColor);
    CGContextFillRect(UIGraphicsGetCurrentContext(), CGRectMake(0, 0, originalImage.size.width, originalImage.size.height));
    UIImage *returnImage = UIGraphicsGetImageFromCurrentImageContext();
    UIGraphicsEndImageContext();
    
    return returnImage;
}

+ (UIImage*)toolImageNamed:(NSString*)name inverted:(BOOL)inverted
{
    if( !inverted )
        return [PTToolsUtil toolImageNamed:name];
    else
        return makeImageNegative([PTToolsUtil toolImageNamed:name]);
}

+ (UIImage*)toolImageNamed:(NSString*)name
{
    
    UIImage *image = [UIImage imageNamed:name
                                inBundle:[PTToolsUtil toolsBundle]
           compatibleWithTraitCollection:Nil];
    
    if( image )
    {
        return image;
    }
    
    
	NSArray<NSString*>* imageExtensions = @[@"png", @"jpg", @"jpeg", @"tif", @"tiff", @"gif", @"bmp", @"BMP", @"ico", @"cur", @"xbm"];
	
	NSString* nameNoExt = [name stringByDeletingPathExtension];
    

	for (NSString* ext in imageExtensions) {
		
		
		NSString* path = [[PTToolsUtil toolsBundle] pathForResource:[@"Images" stringByAppendingPathComponent:nameNoExt] ofType:ext];
		
		if( path )
		{
            UIImage* img = [[UIImage imageWithContentsOfFile:path] imageWithRenderingMode:UIImageRenderingModeAlwaysTemplate];
			if( img )
                
				return img;
		}

	}
    
    NSLog(@"Could not find %@", name);


	return nil;
}


+ (NSBundle *)findToolsBundle
{
    // Bundle of the enclosing class (for Tools framework bundle).
    NSBundle *classBundle = [NSBundle bundleForClass:[self class]];
    if ([classBundle.bundleIdentifier isEqualToString:PTToolsBundleIdentifier]) {
        return classBundle;
    }
    
    // Tools.bundle in the enclosing bundle.
    NSURL *bundleURL = [classBundle URLForResource:@"Tools" withExtension:@"bundle"];
    if (bundleURL) {
        NSBundle *bundle = [NSBundle bundleWithURL:bundleURL];
        if ([bundle.bundleIdentifier isEqualToString:PTToolsBundleIdentifier]) {
            return bundle;
        }
    }
    
    // Bundle with Tools identifier located in the enclosing bundle.
    NSArray<NSURL *> *bundleURLs = [classBundle URLsForResourcesWithExtension:@"bundle" subdirectory:nil];
    for (NSURL *url in bundleURLs) {
        NSBundle *bundle = [NSBundle bundleWithURL:url];
        if ([bundle.bundleIdentifier isEqualToString:PTToolsBundleIdentifier]) {
            return bundle;
        }
    }
    
    return nil;
}

+ (NSBundle*)toolsBundle
{
	if( !PTToolsUtil_toolsBundle )
	{
        NSBundle *bundle = [self findToolsBundle];
        if (!bundle) {
            // Failed to find Tools bundle: Using main bundle.
            bundle = [NSBundle mainBundle];
        }
        
        PTToolsUtil_toolsBundle = bundle;
	}
	
	return PTToolsUtil_toolsBundle;
}

@end

NSString *PTLocalizedString(NSString *key, NSString *comment)
{
    NSString * string = [[NSBundle mainBundle] localizedStringForKey:key value:PT_LocalizedStringNotFound table:@"PTLocalizable"];
    if (![string isEqualToString:PT_LocalizedStringNotFound]){
        return string;
    }
    return [PTToolsUtil.toolsBundle localizedStringForKey:key value:@"" table:nil];
}

NSString *PTLocalizedStringFromTable(NSString *key, NSString *table, NSString *comment)
{
    NSString * string = [[NSBundle mainBundle] localizedStringForKey:key value:PT_LocalizedStringNotFound table:table];
    if (![string isEqualToString:PT_LocalizedStringNotFound]){
        return string;
    }
    return [PTToolsUtil.toolsBundle localizedStringForKey:key value:@"" table:table];
}
