//---------------------------------------------------------------------------------------
// Copyright (c) 2001-2020 by PDFTron Systems Inc. All Rights Reserved.
// Consult legal.txt regarding legal and license information.
//---------------------------------------------------------------------------------------

#import "ToolsConfig.h"
#import "NSObject+PTAdditions.h"

#import "Runtime.h"

@implementation NSObject (PTAdditions)

+ (nullable instancetype)pt_castAsKindFromObject:(nullable id)object
{
    if ([object isKindOfClass:self]) {
        return object;
    }
    return nil;
}

+ (nullable instancetype)pt_castAsMemberFromObject:(nullable id)object
{
    if ([object isMemberOfClass:self]) {
        return object;
    }
    return nil;
}

#pragma mark - Property names

+ (void)pt_enumeratePropertiesForClass:(Class)cls usingBlock:(void (^)(objc_property_t property, BOOL *stop))block
{
    BOOL stop = NO;
    
    unsigned int propertyCount = 0;
    objc_property_t *properties = class_copyPropertyList(cls, &propertyCount);
    
    cleanup({
        free(properties);
    })
    
    for (unsigned int i = 0; i < propertyCount; i++) {
        objc_property_t property = properties[i];
        
        block(property, &stop);
        
        if (stop) {
            break;
        }
    }
}

+ (NSDictionary<NSString *, Class> *)pt_generateLocalPropertyClassesForClass:(Class)cls
{
    NSMutableDictionary<NSString *, Class> *mutableClasses = [NSMutableDictionary dictionary];
    
    [self pt_enumeratePropertiesForClass:cls usingBlock:^(objc_property_t property, BOOL *stop) {
        NSString *key = [NSString stringWithUTF8String:property_getName(property)];
        
        Class propertyClass = pt_property_getClass(property);
        if (propertyClass) {
            mutableClasses[key] = propertyClass;
        }
    }];
    
    return [mutableClasses copy];
}

static void * NSObject_pt_LocalPropertyClassesKey = &NSObject_pt_LocalPropertyClassesKey;

+ (NSDictionary<NSString *, Class> *)pt_localPropertyClassesForClass:(Class)cls
{
    if ([cls isEqual:[NSObject class]]) {
        return nil;
    }
    
    // Check for cached property classes.
    NSDictionary<NSString *, Class> *cachedClasses = objc_getAssociatedObject(cls, NSObject_pt_LocalPropertyClassesKey);
    if (cachedClasses) {
        return cachedClasses;
    }
    
    // Generate property classes.
    NSDictionary<NSString *, Class> *classes = [self pt_generateLocalPropertyClassesForClass:cls];
    
    // Cache property classes.
    objc_setAssociatedObject(cls, NSObject_pt_LocalPropertyClassesKey, classes, OBJC_ASSOCIATION_COPY_NONATOMIC);
    
    return classes;
}

+ (NSSet<NSString *> *)pt_propertyNamesForKindOfClass:(Class)aClass
{
    // Collect the classes to query, with the least derived class first.
    // This is done to support subclasses who override properties from their superclass with a different
    // type.
    NSMutableArray<Class> *mutableClassesToQuery = [NSMutableArray array];
    for (Class cls = self; ![cls isEqual:[NSObject class]]; cls = [cls superclass]) {
        [mutableClassesToQuery insertObject:cls atIndex:0];
    }
    NSArray<Class> *classesToQuery = [mutableClassesToQuery copy];
    
    NSMutableSet<NSString *> *allNames = [NSMutableSet set];
    
    for (Class cls in classesToQuery) {
        NSDictionary<NSString *, Class> *classes = [self pt_localPropertyClassesForClass:cls];
        
        [classes enumerateKeysAndObjectsUsingBlock:^(NSString *key, Class class, BOOL *stop) {
            if ([class isSubclassOfClass:aClass]) {
                [allNames addObject:key];
            }
        }];
    }
    
    return [allNames copy];
}

@end

PT_DEFINE_CATEGORY_SYMBOL(NSObject, PTAdditions)
