//---------------------------------------------------------------------------------------
// Copyright (c) 2001-2020 by PDFTron Systems Inc. All Rights Reserved.
// Consult legal.txt regarding legal and license information.
//---------------------------------------------------------------------------------------

#import "ToolsConfig.h"
#import "PTAutoCoding.h"

#import "Runtime.h"

#include <objc/runtime.h>

@implementation PTAutoCoding

+ (void)PT_encodeIvar:(Ivar)ivar ofObject:(id)object forKey:(NSString *)key withCoder:(NSCoder *)coder
{
    NSParameterAssert(key != nil);
    
    const char *typeEncoding = ivar_getTypeEncoding(ivar);
    if (!typeEncoding) {
        // No type encoding information.
        return;
    }
    
    switch (typeEncoding[0]) {
        case _C_ID:
        {
            id value = object_getIvar(object, ivar);
            if (value) {
                if ([value conformsToProtocol:@protocol(NSCoding)]) {
                    [coder encodeObject:value forKey:key];
                }
            }
        }
            break;
        case _C_CLASS:
        {
            Class value = object_getIvar(object, ivar);
            if (value) {
                [coder encodeObject:NSStringFromClass(value) forKey:key];
            }
        }
            break;
        case _C_SEL:
        {
            SEL value = *((SEL *)pt_object_getIvarValue(object, ivar));
            if (value) {
                [coder encodeObject:NSStringFromSelector(value) forKey:key];
            }
        }
            break;
        case _C_INT:
        {
            int value = *((int *)pt_object_getIvarValue(object, ivar));
            [coder encodeInt:value forKey:key];
        }
            break;
        default:
            break;
    }
}

+ (void)PT_decodeIvar:(Ivar)ivar ofObject:(id)object forKey:(NSString *)key withCoder:(NSCoder *)coder
{
    NSParameterAssert(key != nil);
    
    // Skip keys not present in coder.
    if (![coder containsValueForKey:key]) {
        return;
    }
    
    const char *typeEncoding = ivar_getTypeEncoding(ivar);
    if (!typeEncoding) {
        // No type encoding information.
        return;
    }
    
    switch (typeEncoding[0]) {
        case _C_ID:
        {
            id value = [coder decodeObjectForKey:key];
            object_setIvar(object, ivar, value);
        }
            break;
        case _C_CLASS:
        {
            NSString *stringValue = [coder decodeObjectForKey:key];
            Class value = NSClassFromString(stringValue);
            object_setIvar(object, ivar, value);
        }
            break;
        case _C_SEL:
        {
            SEL value = nil;
            id stringValue = [coder decodeObjectForKey:key];
            if ([stringValue isKindOfClass:[NSString class]]) {
                value = NSSelectorFromString((NSString *)stringValue);
            }
            pt_object_setIvarValue(object, ivar, &value, sizeof(value));
        }
            break;
        case _C_INT:
        {
            int value = [coder decodeIntForKey:key];
            pt_object_setIvarValue(object, ivar, &value, sizeof(value));
        }
            break;
        case _C_BOOL:
        {
            BOOL value = [coder decodeBoolForKey:key];
            pt_object_setIvarValue(object, ivar, &value, sizeof(value));
        }
            break;
        case _C_LNG_LNG:
        {
            long long value = 0;
            size_t size = sizeof(value);
            if (size == sizeof(int64_t)) {
                value = [coder decodeInt64ForKey:key];
            }
            else if (size == sizeof(int32_t)) {
                value = [coder decodeInt32ForKey:key];
            }
            else {
                // Unknown size.
                break;
            }
            pt_object_setIvarValue(object, ivar, &value, size);
        }
            break;
        default:
            break;
    }
}

+ (void)PT_enumeratePropertyIvarsForClass:(Class)cls usingBlock:(void (^)(objc_property_t property, Ivar ivar))block
{
    // Get the list of all properties for the class.
    unsigned int propertyCount = 0;
    objc_property_t *properties = class_copyPropertyList(cls, &propertyCount);
    cleanup({
        free(properties);
    })
    
    for (unsigned int propertyIndex = 0; propertyIndex < propertyCount; propertyIndex++) {
        objc_property_t property = properties[propertyIndex];
        
        // Check if the property is backed by an ivar.
        Ivar ivar = pt_class_getPropertyIvar(cls, property);
        if (!ivar) {
            // No ivar found for the property.
            continue;
        }
        
        if (block) {
            block(property, ivar);
        }
    }
}

+ (void)autoArchiveObject:(id)object withCoder:(NSCoder *)coder
{
    // Walk up the object's class hierarchy, enumerating the list of (instance) properties at each level.
    for (Class cls = [object class]; cls && ![cls isEqual:[NSObject class]]; cls = [cls superclass]) {
        // Enumerate all properties backed by an ivar.
        [self PT_enumeratePropertyIvarsForClass:cls usingBlock:^(objc_property_t property, Ivar ivar) {
            // Encode the ivar, with the property name as the key.
            NSString *key = [NSString stringWithUTF8String:property_getName(property)];
            [self PT_encodeIvar:ivar ofObject:object forKey:key withCoder:coder];
        }];
    }
}

+ (void)autoUnarchiveObject:(id)object withCoder:(NSCoder *)coder
{
    // Walk up the object's class hierarchy, enumerating the list of (instance) properties at each level.
    for (Class cls = [object class]; cls && ![cls isEqual:[NSObject class]]; cls = [cls superclass]) {
        // Enumerate all properties backed by an ivar.
        [self PT_enumeratePropertyIvarsForClass:cls usingBlock:^(objc_property_t property, Ivar ivar) {
            // Decode the ivar, with the property name as the key.
            NSString *key = [NSString stringWithUTF8String:property_getName(property)];
            [self PT_decodeIvar:ivar ofObject:object forKey:key withCoder:coder];
        }];
    }
}

@end
