//---------------------------------------------------------------------------------------
// Copyright (c) 2001-2020 by PDFTron Systems Inc. All Rights Reserved.
// Consult legal.txt regarding legal and license information.
//---------------------------------------------------------------------------------------

#import "ToolsConfig.h"
#import "PTDocumentTabBar.h"

#import "PTDocumentTabBarFlowLayout.h"
#import "PTDocumentTabBarCell.h"

#import "UICollectionView+PTAdditions.h"
#import "CGGeometry+PTAdditions.h"

#import <tgmath.h>

const CGFloat PTDocumentTabBarDefaultMinimumTabWidth = 120;

static const CGFloat PTDocumentTabBar_IntrinsicHeight = 32.0;

static NSString * const PTDocumentTabBar_CellReuseIdentifier = @"PTDocumentTabBarCell";

@interface PTDocumentTabBar ()

@property (nonatomic, assign) CGVector interactiveMovementOffset;

@end

@implementation PTDocumentTabBar

- (instancetype)initWithFrame:(CGRect)frame
{
    self = [super initWithFrame:frame];
    if (self) {
        // Set up collection view layout.
        UICollectionViewFlowLayout *collectionViewLayout = [[PTDocumentTabBarFlowLayout alloc] init];
        collectionViewLayout.scrollDirection = UICollectionViewScrollDirectionHorizontal;
        
        // No margins.
        collectionViewLayout.sectionInset = UIEdgeInsetsZero;
        // No horizontal spacing.
        collectionViewLayout.minimumLineSpacing = 0.0;
        
        _collectionView = [[UICollectionView alloc] initWithFrame:frame collectionViewLayout:collectionViewLayout];
        _collectionView.dataSource = self;
        _collectionView.delegate = self;

        [_collectionView registerClass:[PTDocumentTabBarCell class] forCellWithReuseIdentifier:PTDocumentTabBar_CellReuseIdentifier];
        
        _interactiveMovementOffset = PTCGVectorZero;
        
        [_collectionView addGestureRecognizer:[[UILongPressGestureRecognizer alloc] initWithTarget:self action:@selector(handleCollectionViewLongPress:)]];
        
        _collectionView.autoresizingMask = UIViewAutoresizingFlexibleWidth | UIViewAutoresizingFlexibleHeight;
        
        [self addSubview:_collectionView];
        
        _collectionView.allowsSelection = YES;
        
        // Disable scroll indicators.
        _collectionView.showsHorizontalScrollIndicator = NO;
        _collectionView.showsVerticalScrollIndicator = NO;
        
        _collectionView.backgroundColor = [UIColor clearColor];
        
        // Default background view.
        self.backgroundView = [[UIToolbar alloc] init];
        self.minimumTabWith = PTDocumentTabBarDefaultMinimumTabWidth;
        
        // 1px bottom border.
        UIView *shadowView = [[UIView alloc] init];
        shadowView.backgroundColor = [UIColor colorWithWhite:0.0 alpha:0.3];
        
        [self addSubview:shadowView];
        
        shadowView.translatesAutoresizingMaskIntoConstraints = NO;
        
        [NSLayoutConstraint activateConstraints:
         @[
           [shadowView.centerXAnchor constraintEqualToAnchor:self.centerXAnchor],
           [shadowView.bottomAnchor constraintEqualToAnchor:self.bottomAnchor],
           [shadowView.widthAnchor constraintEqualToAnchor:self.widthAnchor],
           [shadowView.heightAnchor constraintEqualToConstant:(1 / UIScreen.mainScreen.nativeScale)],
           ]];
    }
    return self;
}

- (void)layoutSubviews
{
    [super layoutSubviews];
    
    [self.collectionView.collectionViewLayout invalidateLayout];
}

- (CGSize)intrinsicContentSize
{
    return CGSizeMake(UIViewNoIntrinsicMetric, PTDocumentTabBar_IntrinsicHeight);
}

#pragma mark - Background view

- (UIView *)backgroundView
{
    return self.collectionView.backgroundView;
}

- (void)setBackgroundView:(UIView *)backgroundView
{
    self.collectionView.backgroundView = backgroundView;
}

#pragma mark - <UICollectionViewDataSource>

- (NSInteger)numberOfSectionsInCollectionView:(UICollectionView *)collectionView
{
    return 1;
}

- (NSInteger)collectionView:(UICollectionView *)collectionView numberOfItemsInSection:(NSInteger)section
{
    return [self.dataSource documentTabBarNumberOfItems:self];
}

- (UICollectionViewCell *)collectionView:(UICollectionView *)collectionView cellForItemAtIndexPath:(NSIndexPath *)indexPath
{
    UICollectionViewCell *collectionViewCell = [collectionView dequeueReusableCellWithReuseIdentifier:PTDocumentTabBar_CellReuseIdentifier forIndexPath:indexPath];
    
    NSString *title = [self.dataSource documentTabBar:self titleForItemAtIndex:indexPath.item];
    
    if ([collectionViewCell isKindOfClass:[PTDocumentTabBarCell class]]) {
        PTDocumentTabBarCell *tabBarCell = (PTDocumentTabBarCell *)collectionViewCell;
        
        [tabBarCell configureWithTitle:title];
        
        [tabBarCell.button addTarget:self
                              action:@selector(closeDocumentTab:)
                    forControlEvents:UIControlEventPrimaryActionTriggered];
    }
    
    return collectionViewCell;
}

#pragma mark Interactive movement

- (BOOL)collectionView:(UICollectionView *)collectionView canMoveItemAtIndexPath:(NSIndexPath *)indexPath
{
    // Ask data source.
    if ([self.dataSource respondsToSelector:@selector(documentTabBar:canMoveItemAtIndex:)]) {
        return [self.dataSource documentTabBar:self canMoveItemAtIndex:indexPath.item];
    }
    
    // Disable movement by default.
    return NO;
}

- (void)collectionView:(UICollectionView *)collectionView moveItemAtIndexPath:(NSIndexPath *)sourceIndexPath toIndexPath:(NSIndexPath *)destinationIndexPath
{
    NSUInteger sourceIndex = sourceIndexPath.item;
    NSUInteger destinationIndex = destinationIndexPath.item;
    
    // Update data source.
    if ([self.dataSource respondsToSelector:@selector(documentTabBar:moveItemAtIndex:toIndex:)]) {
        [self.dataSource documentTabBar:self moveItemAtIndex:sourceIndex toIndex:destinationIndex];
    }
}

#pragma mark - <UICollectionViewDelegate>

// NOTE: Consider highlights as selections.

- (BOOL)collectionView:(UICollectionView *)collectionView shouldHighlightItemAtIndexPath:(NSIndexPath *)indexPath
{
    // Ask delegate.
    if ([self.delegate respondsToSelector:@selector(documentTabBar:shouldSelectItemAtIndex:)]) {
        return [self.delegate documentTabBar:self shouldSelectItemAtIndex:indexPath.item];
    }
    
    // Allow selection (highlight) by default.
    return YES;
}

- (void)collectionView:(UICollectionView *)collectionView didHighlightItemAtIndexPath:(NSIndexPath *)indexPath
{
    [collectionView selectItemAtIndexPath:indexPath animated:NO scrollPosition:UICollectionViewScrollPositionNone];
    
    // Scroll to make item visible.
    [collectionView pt_scrollItemToVisible:indexPath animated:YES];
    
    // Notify delegate of selection (highlight).
    if ([self.delegate respondsToSelector:@selector(documentTabBar:didSelectItemAtIndex:)]) {
        [self.delegate documentTabBar:self didSelectItemAtIndex:indexPath.item];
    }
}

- (BOOL)collectionView:(UICollectionView *)collectionView shouldSelectItemAtIndexPath:(NSIndexPath *)indexPath
{
    // Selection is handled via highlighting.
    return NO;
}

- (void)collectionView:(UICollectionView *)collectionView didSelectItemAtIndexPath:(NSIndexPath *)indexPath
{
    // Selection is handled via highlighting.
}

#pragma mark - <UICollectionViewDelegateFlowLayout>

- (CGSize)collectionView:(UICollectionView *)collectionView layout:(UICollectionViewLayout *)collectionViewLayout sizeForItemAtIndexPath:(NSIndexPath *)indexPath
{
    CGRect bounds = collectionView.bounds;
    
    UIEdgeInsets inset = collectionView.contentInset;
    if (@available(iOS 11, *)) {
        inset = collectionView.adjustedContentInset;
    }
    
    bounds = UIEdgeInsetsInsetRect(bounds, inset);
    
    NSUInteger itemCount = [self.dataSource documentTabBarNumberOfItems:self];
    
    return CGSizeMake(fmax(CGRectGetWidth(bounds) / itemCount, self.minimumTabWith), CGRectGetHeight(bounds));
}

#pragma mark - Selection

- (NSUInteger)selectedIndex
{
    NSArray<NSIndexPath *> *selectedIndexPaths = self.collectionView.indexPathsForSelectedItems;
    
    if ([self.dataSource documentTabBarNumberOfItems:self] > 0) {
//        NSAssert(selectedIndexPaths.count == 1, @"Tab bar does not contain a selected tab");
    }
    
    NSIndexPath *indexPath = selectedIndexPaths.firstObject;
    return (indexPath ? indexPath.item : NSNotFound);
}

- (void)setSelectedIndex:(NSUInteger)selectedIndex
{
    [self setSelectedIndex:selectedIndex animated:NO];
}

- (void)setSelectedIndex:(NSUInteger)selectedIndex animated:(BOOL)animated
{
    if (selectedIndex >= [self.dataSource documentTabBarNumberOfItems:self]) {
        return;
    }
    
    NSIndexPath *indexPath = [NSIndexPath indexPathForItem:selectedIndex inSection:0];
    
    [self.collectionView selectItemAtIndexPath:indexPath animated:animated scrollPosition:UICollectionViewScrollPositionNone];
    
    [self.collectionView pt_scrollItemToVisible:indexPath animated:animated];
}

- (void)removeTabAtIndex:(NSUInteger)index animated:(BOOL)animated
{
    if (self.dataSource && ![self.dataSource respondsToSelector:@selector(documentTabBar:removeItemAtIndex:)]) {
        return;
    }
    
    // Allow delegate to update selection etc. before removal.
    if ([self.delegate respondsToSelector:@selector(documentTabBar:willRemoveItemAtIndex:)]) {
        [self.delegate documentTabBar:self willRemoveItemAtIndex:index];
    }
    
    [self.collectionView pt_performBatchUpdates:^{
        // Remove item from data source.
        [self.dataSource documentTabBar:self removeItemAtIndex:index];
        
        NSIndexPath *indexPath = [NSIndexPath indexPathForItem:index inSection:0];
        
        // Delete item from collection view.
        [self.collectionView deleteItemsAtIndexPaths:@[indexPath]];
    } completion:^(BOOL finished) {
        // Notify delegate of removal.
        if ([self.delegate respondsToSelector:@selector(documentTabBar:didRemoveItemAtIndex:)]) {
            [self.delegate documentTabBar:self didRemoveItemAtIndex:index];
        }
    } animated:animated];
}

- (void)reloadItems
{
    NSUInteger selected = self.selectedIndex;
    [self.collectionView reloadData];
    self.selectedIndex = selected;
}

#pragma mark - Actions

- (void)closeDocumentTab:(id)sender
{
    if ([sender isKindOfClass:[UIButton class]]) {
        UIButton *button = (UIButton *)sender;
        
        // Get location of button in collection view.
        CGPoint location = [self.collectionView convertPoint:PTCGRectGetCenter(button.bounds) fromView:button];
        
        NSIndexPath *indexPath = [self.collectionView indexPathForItemAtPoint:location];
        if (!indexPath) {
            return;
        }
        
        // Notify delegate of close button action.
        if ([self.delegate respondsToSelector:@selector(documentTabBar:closeButtonTappedForItemWithIndex:)]) {
            [self.delegate documentTabBar:self closeButtonTappedForItemWithIndex:indexPath.item];
        }
    }
}

- (void)handleCollectionViewLongPress:(UILongPressGestureRecognizer *)gestureRecognizer
{
    if (gestureRecognizer.view != self.collectionView) {
        return;
    }
    
    switch (gestureRecognizer.state) {
        case UIGestureRecognizerStateBegan:
        {
            CGPoint position = [gestureRecognizer locationInView:self.collectionView];
            NSIndexPath *indexPath = [self.collectionView indexPathForItemAtPoint:position];
            if (indexPath && [self.collectionView beginInteractiveMovementForItemAtIndexPath:indexPath]) {
                
                UICollectionViewCell *cell = [self.collectionView cellForItemAtIndexPath:indexPath];
                if (cell) {
                    // Save offset from cell center to touch position.
                    self.interactiveMovementOffset = PTCGPointOffsetFromPoint(cell.center, position);
                    
                    // Adjust and apply initial target position.
                    CGPoint targetPosition = PTCGVectorOffsetPoint(position, self.interactiveMovementOffset);
                    [self.collectionView updateInteractiveMovementTargetPosition:targetPosition];
                }
                
                self.interactivelyMoving = YES;
            }
            
            break;
        }
        case UIGestureRecognizerStateChanged:
        {
            CGPoint position = [gestureRecognizer locationInView:self.collectionView];

            // Offset touch position by saved offset.
            CGPoint targetPosition = PTCGVectorOffsetPoint(position, self.interactiveMovementOffset);
            [self.collectionView updateInteractiveMovementTargetPosition:targetPosition];
            
            break;
        }
        case UIGestureRecognizerStateEnded:
        {
            [self.collectionView endInteractiveMovement];
            
            // Reset offset.
            self.interactiveMovementOffset = PTCGVectorZero;
            
            self.interactivelyMoving = NO;
            break;
        }
        case UIGestureRecognizerStateCancelled:
            // Fall through.
        default:
        {
            [self.collectionView cancelInteractiveMovement];
            
            // Reset offset.
            self.interactiveMovementOffset = PTCGVectorZero;
            
            self.interactivelyMoving = NO;
            break;
        }
    }
}

@end
