//---------------------------------------------------------------------------------------
// Copyright (c) 2001-2020 by PDFTron Systems Inc. All Rights Reserved.
// Consult legal.txt regarding legal and license information.
//---------------------------------------------------------------------------------------

#import "ToolsConfig.h"
#import "PTDocumentTabBarCell.h"

#import "PTDocumentTabBar.h"
#import "PTToolsUtil.h"

#include <tgmath.h>

const CGFloat PTDocumentTabBarCellInnerSpacing = 4.0;

@interface PTDocumentTabBarCell ()

@property (nonatomic, strong) UIStackView *stackView;

@end

@implementation PTDocumentTabBarCell

- (instancetype)initWithFrame:(CGRect)frame
{
    self = [super initWithFrame:frame];
    if (self) {
        // Set up label.
        _label = [[UILabel alloc] init];
        _label.textAlignment = NSTextAlignmentCenter;
        _label.lineBreakMode = NSLineBreakByTruncatingMiddle;
        
        _label.font = [UIFont systemFontOfSize:UIFont.smallSystemFontSize weight:UIFontWeightSemibold];
        
        // Compress (and truncate) label if necessary.
        [_label setContentCompressionResistancePriority:UILayoutPriorityDefaultLow forAxis:UILayoutConstraintAxisHorizontal];
        
        // Set up button.
        _button = [UIButton buttonWithType:UIButtonTypeSystem];
        
        UIImage *buttonImage = [UIImage imageNamed:@"ic_ios_cancel_black_16px" inBundle:[PTToolsUtil toolsBundle] compatibleWithTraitCollection:self.traitCollection];
        [_button setImage:buttonImage forState:UIControlStateNormal];
        
        _button.tintAdjustmentMode = UIViewTintAdjustmentModeDimmed;
                
        _button.imageView.contentMode = UIViewContentModeScaleAspectFit;
        
        // Hide by default.
        _button.hidden = YES;
        
        // Resist being made larger than intrinsic content size.
        [_button setContentHuggingPriority:UILayoutPriorityDefaultHigh forAxis:UILayoutConstraintAxisHorizontal];
        
#pragma mark Stack view
        
        // Set up stack view container.
        _stackView = [[UIStackView alloc] initWithFrame:self.contentView.bounds];
        _stackView.axis = UILayoutConstraintAxisHorizontal;
        
        _stackView.alignment = UIStackViewAlignmentCenter;
        
        // Use content hugging/compression resistance priorities to resolve layout.
        _stackView.distribution = UIStackViewDistributionFill;
        
        _stackView.spacing = PTDocumentTabBarCellInnerSpacing;
        
        // Arrange subviews relative to margins.
        _stackView.layoutMarginsRelativeArrangement = YES;

        // Use manual layout margins, and don't adjust for safe area.
        _stackView.layoutMargins = UIEdgeInsetsMake(8, 8, 8, 8);
        if (@available(iOS 11, *)) {
            _stackView.insetsLayoutMarginsFromSafeArea = NO;
        }
        
        _stackView.autoresizingMask = UIViewAutoresizingFlexibleWidth | UIViewAutoresizingFlexibleHeight;
        
        [self.contentView addSubview:_stackView];
        
        [_stackView addArrangedSubview:_button];
        [_stackView addArrangedSubview:_label];
        
        self.selectedBackgroundView = [[UIToolbar alloc] initWithFrame:self.bounds];
        
        [self resetState];
    }
    return self;
}

- (void)configureWithTitle:(NSString *)title
{
    self.label.text = title;
    
    [self updateState];
}

- (void)resetState
{
    self.label.text = nil;
    self.label.enabled = NO;
    self.button.hidden = YES;
}

- (void)updateState
{
    BOOL selected = self.selected;
    
    self.label.enabled = selected;
    self.button.hidden = !selected;
}

- (void)prepareForReuse
{
    [super prepareForReuse];
    
    [self resetState];
}

- (void)setSelected:(BOOL)selected
{
    [super setSelected:selected];
    
    [self updateState];
}

- (UIView *)hitTest:(CGPoint)point withEvent:(UIEvent *)event
{
    UIView *view = [super hitTest:point withEvent:event];
    if (view) {
        CGPoint contentViewPoint = [self convertPoint:point toView:self.contentView];
        
        CGFloat touchTargetDimension = fmin(CGRectGetHeight(self.contentView.bounds), 44.0);
        
        CGRect fakeButtonFrame = CGRectMake(0, 0,
                                            touchTargetDimension, touchTargetDimension);
        
        if (CGRectContainsPoint(fakeButtonFrame, contentViewPoint)) {
            return self.button;
        }
    }
    return view;
}

@end
