//---------------------------------------------------------------------------------------
// Copyright (c) 2001-2020 by PDFTron Systems Inc. All Rights Reserved.
// Consult legal.txt regarding legal and license information.
//---------------------------------------------------------------------------------------

#import "ToolsConfig.h"
#import "PTTabbedDocumentViewController.h"

#import "PTToolsUtil.h"
#import "PTKeyValueObserving.h"
#import "PTDocumentTabBar.h"
#import "PTDocumentViewController+Protected.h"
#import "PTForwardingNavigationItem.h"

#import "UIViewController+PTAdditions.h"
#import "UINavigationBar+PTAdditions.h"

#import <objc/runtime.h>
#import <tgmath.h>

#define UIViewInheritParentAnimationDuration 0

@interface PTTabbedDocumentViewController () <PTDocumentTabBarDataSource, PTDocumentTabBarDelegate, PTDocumentViewControllerToolbarDelegate>

#pragma mark Subviews

@property (nonatomic, strong) PTDocumentTabBar *tabBar;

@property (nonatomic, strong) UIView *contentContainerView;

#pragma mark Layout

@property (nonatomic, assign) UIEdgeInsets additionalChildViewControllerSafeAreaInsets NS_AVAILABLE_IOS(11_0);

@property (nonatomic, assign) BOOL didSetupViewConstraints;

#pragma mark Animation

@property (nonatomic, assign) NSUInteger activeTabBarTransitionCount;

#pragma mark View controllers

@property (nonatomic, readonly, copy, nullable) NSArray<PTDocumentViewController *> *viewControllers;

@property (nonatomic, strong) NSMutableDictionary<NSURL *, PTDocumentViewController *> *urlInfo;

#pragma mark State persistence

@property (nonatomic, readonly, copy) NSString *modelRestorationIdentifier;

#pragma mark Other

@property (nonatomic, readonly, assign) NSUInteger effectiveMaximumTabCount;

@end

@implementation PTTabbedDocumentViewController

- (instancetype)init
{
    self = [super initWithNibName:nil bundle:nil];
    if (self) {
        _tabsEnabled = YES;
        _maximumTabCount = NSUIntegerMax;
        
        _tabURLs = [NSArray array];
        
        _urlInfo = [NSMutableDictionary dictionary];
        
        _additionalChildViewControllerSafeAreaInsets = UIEdgeInsetsZero;
                
        [self restoreModel];
    }
    return self;
}

#pragma mark - View controller lifecycle

- (void)willMoveToParentViewController:(UIViewController *)parent
{
    [super willMoveToParentViewController:parent];
    
    if (parent) {
        // Will be added to parent view controller.
    } else {
        // Will be removed from parent view controller.
        [self.selectedViewController stopAutomaticControlHidingTimer];
    }
}

#pragma mark - View lifecycle

// NOTE: Do not call super implementation.
- (void)loadView
{
    self.view = [[UIView alloc] init];
    
    // Create container view for child view controller.
    self.contentContainerView = [[UIView alloc] init];
    
    [self.view addSubview:self.contentContainerView];
    
    // Create document tab bar.
    self.tabBar = [[PTDocumentTabBar alloc] init];
    self.tabBar.dataSource = self;
    self.tabBar.delegate = self;
    
    [self.view addSubview:self.tabBar];
}

- (void)viewDidLoad {
    [super viewDidLoad];
    
    self.extendedLayoutIncludesOpaqueBars = YES;
    
    if (@available(iOS 11, *)) {
        // Do nothing.
    } else {
        self.automaticallyAdjustsScrollViewInsets = NO;
    }
    
    self.view.backgroundColor = [UIColor colorWithWhite:(180.0 / 255) alpha:1.0];
    
    // Set standard root view resizing mask.
    self.view.autoresizingMask = UIViewAutoresizingFlexibleWidth | UIViewAutoresizingFlexibleHeight;
    
    self.contentContainerView.frame = self.view.bounds;
    self.contentContainerView.autoresizingMask = UIViewAutoresizingFlexibleWidth | UIViewAutoresizingFlexibleHeight;
        
    self.tabBar.backgroundView = [[UIToolbar alloc] init];
    UIColor *tabBarBGColor = [UIColor colorWithWhite:0.9 alpha:1.0];
    if (@available(iOS 11.0, *)) {
        tabBarBGColor = [UIColor colorNamed:@"tabBarBGColor" inBundle:[PTToolsUtil toolsBundle] compatibleWithTraitCollection:self.traitCollection];
    }
    ((UIToolbar*)self.tabBar.backgroundView).barTintColor = tabBarBGColor;
    
    // Schedule view constraints update.
    [self.view setNeedsUpdateConstraints];
}

- (void)updateViewConstraints
{
    if (!self.didSetupViewConstraints) {
        self.tabBar.translatesAutoresizingMaskIntoConstraints = NO;
                
        [NSLayoutConstraint activateConstraints:
         @[
           // Attach tab view to the top of the view controller.
           [self.tabBar.centerXAnchor constraintEqualToAnchor:self.view.centerXAnchor],
           [self.tabBar.topAnchor constraintEqualToAnchor:self.pt_safeTopAnchor],
           [self.tabBar.widthAnchor constraintEqualToAnchor:self.view.widthAnchor],
           /* Use intrinsic PTDocumentTabBar height. */
           ]];
        
        // View constraints are set up.
        self.didSetupViewConstraints = YES;
    }
    
    // Call super implementation as final step.
    [super updateViewConstraints];
}

- (void)viewDidLayoutSubviews
{
    [super viewDidLayoutSubviews];
    
    if (@available(iOS 11, *)) {
        [self updateChildViewControllerAdditionalSafeAreaInsets];
    }

    // Adjust tab bar frame.
    if (self.navigationController.navigationBar && self.navigationController.navigationBarHidden) {
        // The iPhone X animates the (hiding) navigation bar much further than the top of the
        // navigation controller, so the tab bar is left behind during the animation.
        UINavigationBar *navigationBar = self.navigationController.navigationBar;
        CGRect navigationBarFrame = [self.view convertRect:navigationBar.bounds fromView:navigationBar];
        
        CGRect tabBarFrame = self.tabBar.frame;
        tabBarFrame.origin.y = fmin(CGRectGetMaxY(navigationBarFrame), tabBarFrame.origin.y);
        if (!PT_ToolsMacCatalyst) {
            self.tabBar.frame = tabBarFrame;
        }
    }
}

#pragma mark - View appearance

- (void)viewWillAppear:(BOOL)animated
{
    [super viewWillAppear:animated];
    
//    [self restoreModel];
    [self.tabBar reloadItems];
    self.tabBar.selectedIndex = self.selectedIndex;
    
    // Hide the navigation bar shadow (bottom line) when tabs are enabled.
    self.navigationController.navigationBar.pt_shadowHidden = self.tabsEnabled;
}

- (void)viewDidAppear:(BOOL)animated
{
    [super viewDidAppear:animated];
    
    if (@available(iOS 11, *)) {
        [self updateChildViewControllerAdditionalSafeAreaInsets];
    }
}

- (void)viewWillDisappear:(BOOL)animated
{
    [super viewWillDisappear:animated];
    
    // Restore the navigation bar shadow (bottom line).
    self.navigationController.navigationBar.pt_shadowHidden = NO;
}

- (void)didReceiveMemoryWarning
{
    [super didReceiveMemoryWarning];
    
    // Remove offscreen view controllers.
    PTDocumentViewController *selectedViewController = self.selectedViewController;

    for (NSURL *url in self.tabURLs) {
        PTDocumentViewController *viewController = self.urlInfo[url];
        if (!viewController || viewController == selectedViewController) {
            continue;
        }

        // Remove child view controller.
        [self PT_removeDocumentViewController:viewController];
    }
}

#pragma mark Safe area

- (void)updateChildViewControllerAdditionalSafeAreaInsets NS_AVAILABLE_IOS(11_0)
{
    UIEdgeInsets insets = UIEdgeInsetsZero;
    
    if (!self.tabBarHidden) {
        insets = UIEdgeInsetsMake(CGRectGetHeight(self.tabBar.bounds), 0, 0, 0);
    }
    
    if (UIEdgeInsetsEqualToEdgeInsets(insets, self.additionalChildViewControllerSafeAreaInsets)) {
        // No change to insets.
        return;
    }
    
    self.additionalChildViewControllerSafeAreaInsets = insets;
}

- (void)setAdditionalChildViewControllerSafeAreaInsets:(UIEdgeInsets)additionalChildSafeAreaInsets
{
    _additionalChildViewControllerSafeAreaInsets = additionalChildSafeAreaInsets;
    
    for (UIViewController *viewController in self.viewControllers) {
        viewController.additionalSafeAreaInsets = additionalChildSafeAreaInsets;
    }
}

#pragma mark - Status bar

- (UIViewController *)childViewControllerForStatusBarHidden
{
    // Selected view controller controls status bar hidden/unhidden state.
    return self.selectedViewController;
}

- (UIViewController *)childViewControllerForStatusBarStyle
{
    // Selected view controller controls status bar style.
    return self.selectedViewController;
}

- (UIViewController *)childViewControllerForScreenEdgesDeferringSystemGestures
{
    // Selected view controller controls screen edge gesture deferring.
    return self.selectedViewController;
}

#pragma mark - Tab settings

- (void)setTabsEnabled:(BOOL)tabsEnabled
{
    if (_tabsEnabled == tabsEnabled) {
        // No change.
        return;
    }
    
    _tabsEnabled = tabsEnabled;
    
    if (!tabsEnabled) {
        // Trim tabs list.
        [self adjustTabURLs];
    }
    
    // Hide tab bar when tabs are disabled.
    BOOL tabBarHidden = !tabsEnabled;
    
    // Hide tab bar only when currently shown.
    if (tabBarHidden && self.tabBarHidden != tabBarHidden) {
         [self setTabBarHidden:YES animated:NO];
    }
    else if (!tabBarHidden && !self.navigationController.navigationBarHidden) {
        [self setTabBarHidden:NO animated:NO];
    }
}

- (void)setMaximumTabCount:(NSUInteger)maximumTabCount
{
    if (maximumTabCount < 1) {
        PTLog(@"Maximum tab count must be greater than zero");
        return;
    }
    
    if ((!self.tabsEnabled && maximumTabCount > 1) ||
        _maximumTabCount == maximumTabCount) {
        return;
    }
    
    _maximumTabCount = maximumTabCount;
    
    if (self.tabURLs.count <= maximumTabCount) {
        return;
    }
    
    // Adjust tab URLs array.
    [self adjustTabURLs];
}

- (NSUInteger)effectiveMaximumTabCount
{
    return (self.tabsEnabled) ? self.maximumTabCount : 1;
}

- (void)adjustTabURLs
{
    NSMutableArray<NSURL *> *tabURLs = [self.tabURLs mutableCopy];
    NSArray<NSURL *> *urlsToKeep = self.selectedURL ? @[self.selectedURL] : nil;
    
    NSArray<NSURL *> *removedURLs = [self trimItems:tabURLs itemsToKeep:urlsToKeep withLimit:self.effectiveMaximumTabCount];
    if (removedURLs.count < 1) {
        // No adjustment necessary.
        return;
    }
    
    for (NSURL *url in removedURLs) {
        PTDocumentViewController *viewController = self.urlInfo[url];
        if (!viewController) {
            continue;
        }
        
        // Remove child view controller.
        [self PT_removeDocumentViewController:viewController];
    }
    
    self.tabURLs = tabURLs;
    
    [self.tabBar reloadItems]; // legit use
    self.tabBar.selectedIndex = self.selectedIndex;
}

- (NSArray *)trimItems:(NSMutableArray *)items itemsToKeep:(NSArray *)itemsToKeep withLimit:(NSUInteger)limit
{
    NSAssert(itemsToKeep.count <= limit,
             @"Cannot keep more items than specified limit");
 
    NSUInteger itemCount = items.count;

    if (itemCount <= limit) {
        return nil;
    }
    
    NSMutableArray *itemsToRemove = [NSMutableArray array];
    
    for (id item in items) {
        if (![itemsToKeep containsObject:item]) {
            [itemsToRemove addObject:item];
            
            // Check if enough items have been removed.
            if ((itemCount - itemsToRemove.count) <= limit) {
                break;
            }
        }
    }
    
    [items removeObjectsInArray:itemsToRemove];
    
    return [itemsToRemove copy];
}

#pragma mark - Tab bar

- (PTDocumentTabBar *)tabBar
{
    [self loadViewIfNeeded];
    
    return _tabBar;
}

- (void)setTabBarHidden:(BOOL)hidden
{
    if (self.tabBarHidden == hidden) {
        return;
    }
    
    _tabBarHidden = hidden;
    
    // Animation pre-amble:
    if (self.activeTabBarTransitionCount == 0) {
        if (hidden) {
            // No pre-amble.
        } else {
            // Show tab bar in preparation for animation.
            self.tabBar.hidden = NO;
        }
    }
    
    [UIView animateWithDuration:UIViewInheritParentAnimationDuration delay:0 options:UIViewAnimationOptionBeginFromCurrentState animations:^{
        // Animate layout changes.
        [self.view layoutIfNeeded];
        
        // Animate tab bar alpha: requires UIViewAnimationOptionBeginFromCurrentState.
        if (hidden && CGRectIntersectsRect(self.view.bounds, self.tabBar.frame)) {
            self.tabBar.alpha = 0.0;
        } else {
            self.tabBar.alpha = 1.0;
        }
    } completion:^(BOOL finished) {
        self.activeTabBarTransitionCount--;
        
        // Animation post-amble:
        if (self.activeTabBarTransitionCount == 0) {
            // Check state at time of completion.
            if (self.tabBarHidden) {
                // Hide tab bar.
                self.tabBar.hidden = YES;
            } else {
                // No post-amble.
            }
        }
    }];
    
    self.activeTabBarTransitionCount++;
}

- (void)setTabBarHidden:(BOOL)hidden animated:(BOOL)animated
{
    if (animated) {
        [UIView animateWithDuration:UINavigationControllerHideShowBarDuration animations:^{
            self.tabBarHidden = hidden;
        }];
    } else {
        self.tabBarHidden = hidden;
    }
}

#pragma mark - Tab urls

- (void)addTabWithURL:(NSURL *)url selected:(BOOL)selected
{
    [self insertTabWithURL:url atIndex:self.tabURLs.count selected:selected];
}

- (void)insertTabWithURL:(NSURL *)url atIndex:(NSUInteger)index selected:(BOOL)selected
{
    // Need access to the security scoped resource before anything else is done.
    if ([url isFileURL]) {
        BOOL success = [url startAccessingSecurityScopedResource];
        if (!success) {
            PTLog(@"Failed to access security scoped resource with URL: %@", url);
        }
    }
    
    // Create view controller for url, if necessary.
    PTDocumentViewController *viewController = self.urlInfo[url];
    if (!viewController) {
        viewController = [self createDocumentViewController];
        
        if ([self.delegate respondsToSelector:@selector(tabbedDocumentViewController:willAddDocumentViewController:)]) {
            [self.delegate tabbedDocumentViewController:self willAddDocumentViewController:viewController];
        }
        
        [viewController openDocumentWithURL:url];
        url = viewController.localDocumentURL ?: viewController.documentUrl ?: url;
        
        
        
        // Add as child view controller.
        [self PT_addDocumentViewController:viewController forURL:url];
    }
    
    if (![self.tabURLs containsObject:url]) {
        if (self.viewIfLoaded.window) {
            // Update tab URLs via the tab bar when the view is added to the window to animate the change.
            [self.tabBar.collectionView performBatchUpdates:^{
                // Update data source.
                NSMutableArray<NSURL *> *tabURLs = [self.tabURLs mutableCopy];
                [tabURLs insertObject:url atIndex:index];
                self.tabURLs = tabURLs;
                
                NSIndexPath *indexPath = [NSIndexPath indexPathForItem:index inSection:0];
                
                [self.tabBar.collectionView insertItemsAtIndexPaths:@[indexPath]];
            } completion:nil];
        } else {
            // Note: the tab bar will be updated when the view is added to the window.
            NSMutableArray<NSURL *> *tabURLs = [self.tabURLs mutableCopy];
            [tabURLs insertObject:url atIndex:index];
            self.tabURLs = tabURLs;
        }
    }
    
    if (selected) {
        self.selectedURL = url;
    }
    
    // Adjust tab URLs for maximumTabCount and tabsEnabled.
    // MUST be done after (optionally) selecting the new URL.
    [self adjustTabURLs];
    
    
    return;
}

- (void)removeTabWithURL:(NSURL *)url
{
    [self removeTabAtIndex:[self.tabURLs indexOfObject:url]];
}

- (void)removeTabAtIndex:(NSUInteger)index
{
    if (index > self.tabURLs.count) {
        // Invalid index.
        return;
    }

    [self.tabBar removeTabAtIndex:index animated:NO];
}

- (void)setTabURLs:(NSArray<NSURL *> *)tabURLs
{
    if ([_tabURLs isEqualToArray:tabURLs]) {
        // No change.
        return;
    }
    
    NSURL *selectedURL = _selectedURL;

    _tabURLs = [tabURLs copy];
    
    for (NSURL *url in _tabURLs) {
        if ([url isEqual:selectedURL]) {
            _selectedURL = url;
            break;
        }
    }
    
    [self persistModel];
}

#pragma mark - View controllers

- (PTDocumentViewController *)createDocumentViewController
{
    return [[PTDocumentViewController allocOverridden] init];
}

- (PTDocumentViewController *)documentViewControllerAtIndex:(NSUInteger)index
{
    if (index >= self.tabURLs.count) {
        // Invalid index.
        return nil;
    }
    
    // Get the URL at the specified index.
    NSURL *url = self.tabURLs[index];
    if (!url) {
        return nil;
    }
    
    // Get the document view controller for the URL, if it is loaded.
    return self.urlInfo[url];
}

- (NSArray<PTDocumentViewController *> *)viewControllers
{
    return self.urlInfo.allValues;
}

- (void)transitionFromViewController:(PTDocumentViewController *)fromViewController toViewController:(PTDocumentViewController *)toViewController
{
    
    [self loadViewIfNeeded];
    
    // Remove old view controller view if loaded and attached to the view hierarchy.
    if ([fromViewController.viewIfLoaded isDescendantOfView:self.contentContainerView]) {
        [fromViewController.view removeFromSuperview];
    }
    
    if (toViewController) {
        UIView *toView = toViewController.view;
        
        // Child view expands to fill entire content view.
        toView.frame = self.contentContainerView.bounds;
        toView.autoresizingMask = UIViewAutoresizingFlexibleWidth | UIViewAutoresizingFlexibleHeight;
        
        [self.contentContainerView addSubview:toView];
    }
}

- (void)PT_addDocumentViewController:(PTDocumentViewController *)documentViewController forURL:(NSURL *)url
{
    // Remove any existing mappings for view controller.
    NSSet<NSURL *> *keys = [self.urlInfo keysOfEntriesPassingTest:^BOOL(NSURL *key, PTDocumentViewController *obj, BOOL *stop) {
        return [obj isEqual:documentViewController];
    }];
    [self.urlInfo removeObjectsForKeys:keys.allObjects];
    
    if (!documentViewController.parentViewController) {
        // View controller containment.
        [self addChildViewController:documentViewController];
        [documentViewController didMoveToParentViewController:self];
        
        // Start observing child view controller properties.
        [self pt_observeObject:documentViewController
                    forKeyPath:PT_KEY(documentViewController, localDocumentURL)
                      selector:@selector(documentViewControllerLocalDocumentURLDidChange:)
                       options:(NSKeyValueObservingOptionOld)];
    }
    
    // Associate view controller with given URL.
    self.urlInfo[url] = documentViewController;
}

- (void)PT_removeDocumentViewController:(PTDocumentViewController *)documentViewController
{
    // Remove any mappings for view controller.
    NSSet<NSURL *> *keys = [self.urlInfo keysOfEntriesPassingTest:^BOOL(NSURL *key, PTDocumentViewController *obj, BOOL *stop) {
        return [obj isEqual:documentViewController];
    }];
    
    // Stop observing child view controller properties.
    [documentViewController pt_removeObserver:self
                                   forKeyPath:PT_KEY(documentViewController, localDocumentURL)];
    
    // Remove child view controller.
    [documentViewController willMoveToParentViewController:nil];
    [documentViewController.viewIfLoaded removeFromSuperview];
    [documentViewController removeFromParentViewController];
    
    [self.urlInfo removeObjectsForKeys:keys.allObjects];
}

#pragma mark - Selection

- (void)setSelectedURL:(NSURL *)selectedURL
{
    if ([_selectedURL isEqual:selectedURL]) {
        // No change.
        return;
    }
    
    if (selectedURL && ![self.tabURLs containsObject:selectedURL]) {
        @throw [NSException exceptionWithName:NSInternalInconsistencyException
                                       reason:@"Selected URL is not in list"
                                     userInfo:nil];
        return;
    }
    
    // The incoming selectedURL is not retained and could be deallocated
    // so get a strong reference to an equivalent NSURL instance.
    if (selectedURL) {
        NSUInteger indexOfExistingURL = [self.tabURLs indexOfObject:selectedURL];
        selectedURL = self.tabURLs[indexOfExistingURL];
    }
    
    PTDocumentViewController *viewController = selectedURL ? self.urlInfo[selectedURL] : nil;
    
    if (!viewController && selectedURL) {
        [self openDocumentWithURL:selectedURL];
    }
    else
    {
        // NOTE: viewController is nil when last tab is closing.
        viewController.toolbarDelegate = self;
        [viewController.navigationItem setForwardingTargetItem:self.navigationItem animated:NO];
        
        PTDocumentViewController *previousViewController = _selectedURL ? self.urlInfo[_selectedURL] : nil;
        previousViewController.toolbarDelegate = nil;
        [previousViewController.navigationItem setForwardingTargetItem:nil animated:NO];
        
        _selectedURL = selectedURL;
        
        // Transition to the new view controller.
        [self transitionFromViewController:previousViewController toViewController:viewController];
        
        if (self.tabBar.window) {
            self.tabBar.selectedIndex = self.selectedIndex;
        }
        
        // Update status bar appearance.
        [self setNeedsStatusBarAppearanceUpdate];
    }
}

- (NSUInteger)selectedIndex
{
    if (self.selectedURL) {
        return [self.tabURLs indexOfObject:self.selectedURL];
    }
    return NSNotFound;
}

- (void)setSelectedIndex:(NSUInteger)selectedIndex
{
    if (selectedIndex >= self.tabURLs.count) {
        // Invalid index;
        return;
    }
    
    self.selectedURL = self.tabURLs[selectedIndex]; //james
}

// The selectedIndex property is derived from selectedURL.
+ (NSSet<NSString *> *)keyPathsForValuesAffectingSelectedIndex
{
    return [NSSet setWithObject:PT_CLASS_KEY(PTTabbedDocumentViewController, selectedURL)];
}

- (NSUInteger)nextSelectedIndex
{
    return [self nextSelectedIndexForIndex:self.selectedIndex];
}

- (NSUInteger)nextSelectedIndexForIndex:(NSUInteger)index
{
    if (index == NSNotFound) {
        return NSNotFound;
    }
    
    NSUInteger tabsCount = self.tabURLs.count;
    
    if (tabsCount < 2) {
        // No other tabs to select.
        return NSNotFound;
    }
    
    if (index < tabsCount - 1) {
        return index + 1;
    } else {
        return index - 1;
    }
}

- (PTDocumentViewController *)selectedViewController
{
    return (self.selectedURL) ? self.urlInfo[self.selectedURL] : nil;
}

#pragma mark - State persistence

- (void)setRestorationIdentifier:(NSString *)restorationIdentifier
{
    NSString *previousRestorationIdentifier = self.restorationIdentifier;
    
    [super setRestorationIdentifier:restorationIdentifier];
    
    if (![restorationIdentifier isEqualToString:previousRestorationIdentifier]) {
        [self restoreModel];
    }
}

- (NSString *)modelRestorationIdentifier
{
    NSString *restorationIdentifier = self.restorationIdentifier;
    if (!restorationIdentifier) {
        restorationIdentifier = [NSBundle bundleForClass:[PTDocumentViewController class]].bundleIdentifier;
    }
    if (!restorationIdentifier) {
        restorationIdentifier = @"PTCurrentDocumentTabs";
    }
    
    return restorationIdentifier;
}

#pragma mark Preservation

- (void)persistModel
{
    // Save list of document bookmarks.
    NSArray<NSData *> *bookmarks = [self bookmarksForURLs:self.tabURLs];
    if (bookmarks.count == 0) {
        // Clear previously saved values.
        [NSUserDefaults.standardUserDefaults removeObjectForKey:self.modelRestorationIdentifier];
        return;
    }
    
    [[NSUserDefaults standardUserDefaults] setObject:bookmarks forKey:self.modelRestorationIdentifier];
}

- (NSArray<NSData *> *)bookmarksForURLs:(NSArray<NSURL *> *)urls
{
    NSMutableArray<NSData *> *bookmarks = [NSMutableArray array];
    for (NSURL *url in urls) {
        NSURLBookmarkCreationOptions options = NSURLBookmarkCreationSuitableForBookmarkFile;
        NSError *error = nil;
        
        NSData *bookmark = [url bookmarkDataWithOptions:options includingResourceValuesForKeys:nil relativeToURL:nil error:&error];
        //NSAssert(bookmark, @"Bookmark should not be Nil");
        if (!bookmark) {
            PTLog(@"Failed to create bookmark data for URL \"%@\": %@", url, error);
            continue;
        } else {

        }
        
        [bookmarks addObject:bookmark];
    }

    return [bookmarks copy];
}

#pragma mark Restoration

- (void)restoreModel
{
    NSArray<NSData *> *bookmarks = [[NSUserDefaults standardUserDefaults] objectForKey:self.modelRestorationIdentifier];

    NSArray<NSURL *> *urls = [self URLsForBookmarks:bookmarks];

    for (NSURL *url in urls) {
        //optionally print out what URLs were restored

        (void)url;
    }
    
    if (urls.count == 0) {
        
    }

    // short circut setter because no need to persist what we're restoring.
    _tabURLs = [urls copy];
}

- (NSArray<NSURL *> *)URLsForBookmarks:(NSArray<NSData *> *)bookmarks
{
    // Avoid duplicate URLs.
    NSMutableOrderedSet<NSURL *> *urls = [NSMutableOrderedSet orderedSet];

    for (NSData *bookmark in bookmarks) {
        BOOL bookmarkIsStale = NO;
        NSError *error = nil;
        
        // NSURLBookmarkResolutionWithSecurityScope unavailable and not needed on iOS
        NSURL *url = [NSURL URLByResolvingBookmarkData:bookmark options:NSURLBookmarkResolutionWithoutUI relativeToURL:nil bookmarkDataIsStale:&bookmarkIsStale error:&error];

//        NSAssert(bookmarkIsStale == false, @"Need to handle stale bookmark");
        
        if (!url) {
            PTLog(@"Failed to resolve URL bookmark data: %@", error);
            continue;
        }
        
        
        // renew: https://stackoverflow.com/questions/23954662/what-is-the-correct-way-to-handle-stale-nsurl-bookmarks
        
        if ([url isFileURL]) {
            BOOL success = [url startAccessingSecurityScopedResource];
            if (!success) {
                PTLog(@"Failed to access security scoped resource with URL: %@", url);
                continue;
            }
        }
        
        // Standardize and resolve symlinks in URL.
//        url = url.URLByStandardizingPath.URLByResolvingSymlinksInPath;
        
        [urls addObject:url];
    }

    return ((NSOrderedSet<NSURL *> *)[urls copy]).array;
}

#pragma mark - OLD <PTDocumentOpening>

- (void)openDocumentWithURL:(NSURL *)url
{
    [self openDocumentWithURL:url password:nil];
}

- (void)openDocumentWithURL:(NSURL *)url password:(NSString *)password
{
    
    
    [self addTabWithURL:url selected:YES];
}

#pragma mark - <PTDocumentViewControllerToolbarDelegate>

#pragma mark Toolbar

- (void)documentViewController:(PTDocumentViewController *)documentViewController willShowToolbar:(UIToolbar *)toolbar
{
    // Hide document tab bar.
    if (self.tabsEnabled) {
        [self setTabBarHidden:YES animated:YES];
    }
        
    if (@available(iOS 11, *)) {
        [self updateChildViewControllerAdditionalSafeAreaInsets];
    }
}

- (void)documentViewController:(PTDocumentViewController *)documentViewController didShowToolbar:(UIToolbar *)toolbar
{
    
    
    return;
}

- (void)documentViewController:(PTDocumentViewController *)documentViewController willHideToolbar:(UIToolbar *)toolbar
{
    // Show document tab bar.
    if (self.tabsEnabled) {
        [self setTabBarHidden:NO animated:YES];
    }
    
    if (@available(iOS 11, *)) {
        [self updateChildViewControllerAdditionalSafeAreaInsets];
    }
}

- (void)documentViewController:(PTDocumentViewController *)documentViewController didHideToolbar:(UIToolbar *)toolbar
{
    

    return;
}

#pragma mark Navigation bar

- (BOOL)documentViewControllerShouldHideNavigationBar:(PTDocumentViewController *)documentViewController
{
    if ([self.tabBar isInteractivelyMoving]) {
        return NO;
    }
    
    return YES;
}

- (void)documentViewControllerDidHideNavigationBar:(PTDocumentViewController *)documentViewController animated:(BOOL)animated
{
    if (self.tabsEnabled) {
        [self setTabBarHidden:YES animated:animated];
    }
}

- (void)documentViewControllerDidShowNavigationBar:(PTDocumentViewController *)documentViewController animated:(BOOL)animated
{
    if (self.tabsEnabled) {
        [self setTabBarHidden:NO animated:animated];
    }
}

-(void)documentViewControllerDidOpenDocument:(PTDocumentViewController *)documentViewController
{
    NSLog(@"Document successfully opened");
}

- (BOOL)documentViewController:(PTDocumentViewController *)documentViewController shouldOpenExportedFileAttachmentAtURL:(NSURL *)exportedURL
{
    PTLog(@"Tabbed viewer will handle opening exported file attachment at URL %@", exportedURL);
    
    // Open exported file attachment in new tab.
    [self addTabWithURL:exportedURL selected:YES];
    
    // Tabbed viewer handles opening file attachments.
    return NO;
}

- (BOOL)documentViewController:(PTDocumentViewController *)documentViewController shouldOpenFileURL:(NSURL *)fileURL
{
    PTLog(@"Tabbed viewer will handle opening file at URL %@", fileURL);

    // NOTE: the async dispatch is necessary to avoid issues with read-locks in other places.
    dispatch_async(dispatch_get_main_queue(), ^{
        // Open exported file in new tab.
        [self addTabWithURL:fileURL selected:YES];
    });
    
    // Tabbed viewer handles opening files.
    return NO;
}

#pragma mark - Other

-(void)documentViewControllerDidBecomeInvalid:(PTDocumentViewController *)documentViewController
{
    dispatch_async(dispatch_get_main_queue(), ^() {

        [self removeTabWithURL:documentViewController.documentUrl];
    });
}

#pragma mark - <PTDocumentTabBarDataSource>

- (NSUInteger)documentTabBarNumberOfItems:(PTDocumentTabBar *)tabBar
{
    return self.tabURLs.count;
}

- (NSString *)documentTabBar:(PTDocumentTabBar *)tabBar titleForItemAtIndex:(NSUInteger)index
{
    NSURL *url = self.tabURLs[index];
    
    PTDocumentViewController *viewController = self.urlInfo[url];
    if (viewController) {
        NSURL *cachesURL = [NSFileManager.defaultManager URLsForDirectory:NSCachesDirectory
                                                                inDomains:NSUserDomainMask].firstObject;
        
        // Only use the localDocumentURL when *not* inside the Caches folder.
        if (viewController.localDocumentURL
            && ![viewController.localDocumentURL isEqual:viewController.documentUrl]
            && ![viewController.localDocumentURL.absoluteString hasPrefix:cachesURL.absoluteString]) {
            url = viewController.localDocumentURL;
        } else if (viewController.documentUrl) {
            url = viewController.documentUrl;
        }
    }
    
    // Get file name.
    NSString *filename = url.lastPathComponent;
    
    // Strip file extension.
    return filename.stringByDeletingPathExtension;
}

- (void)documentTabBar:(PTDocumentTabBar *)tabBar removeItemAtIndex:(NSUInteger)index
{
    
    @try {
        NSURL *url = self.tabURLs[index];
        
        // Update data source.
        NSMutableArray<NSURL *> *tabURLs = [self.tabURLs mutableCopy];
        [tabURLs removeObjectAtIndex:index];
        self.tabURLs = tabURLs;
        
        PTDocumentViewController *viewController = self.urlInfo[url];
        if (viewController) {
            // Remove child view controller.
            [self PT_removeDocumentViewController:viewController];
        }
    } @catch (NSException *exception) {

    }
}

- (BOOL)documentTabBar:(PTDocumentTabBar *)tabBar canMoveItemAtIndex:(NSUInteger)index
{
    // Multiple items are required for movement.
    return (self.tabURLs.count > 1 && self.selectedIndex == index);
}

- (void)documentTabBar:(PTDocumentTabBar *)tabBar moveItemAtIndex:(NSUInteger)sourceIndex toIndex:(NSUInteger)destinationIndex
{
    // Update data source.
    NSMutableArray<NSURL *> *tabURLs = [self.tabURLs mutableCopy];
    
    NSURL *url = tabURLs[sourceIndex];
    
    [tabURLs removeObjectAtIndex:sourceIndex];
    [tabURLs insertObject:url atIndex:destinationIndex];
    
    self.tabURLs = tabURLs;
}

#pragma mark - <PTDocumentTabBarDelegate>

- (BOOL)documentTabBar:(PTDocumentTabBar *)tabBar shouldSelectItemAtIndex:(NSUInteger)index
{
    // Allow selection of all tabs.
    return YES;
}

- (void)documentTabBar:(PTDocumentTabBar *)tabBar didSelectItemAtIndex:(NSUInteger)index
{
    // Update selection.
    self.selectedIndex = index;
}

- (void)documentTabBar:(PTDocumentTabBar *)tabBar willRemoveItemAtIndex:(NSUInteger)index
{
    // Update selection before the removal occurs.
    NSUInteger selectedIndex = [self nextSelectedIndexForIndex:index];
    if (selectedIndex != NSNotFound) {
        self.selectedIndex = selectedIndex;
    } else {
        self.selectedURL = nil;
    }
    
    // Notify delegate.
    if ([self.delegate respondsToSelector:@selector(tabbedDocumentViewController:willRemoveTabAtIndex:)]) {
        [self.delegate tabbedDocumentViewController:self willRemoveTabAtIndex:index];
    }
}

- (void)documentTabBar:(PTDocumentTabBar *)tabBar didRemoveItemAtIndex:(NSUInteger)index
{

}

- (void)documentTabBar:(PTDocumentTabBar *)tabBar closeButtonTappedForItemWithIndex:(NSUInteger)index
{
    
    
    // Remove tab at specified index.
    [tabBar removeTabAtIndex:index animated:YES];
}

#pragma mark - PTDocumentViewController observation

- (void)documentViewControllerLocalDocumentURLDidChange:(PTKeyValueObservedChange *)change
{
    PTDocumentViewController *documentViewController = (PTDocumentViewController *)change.object;
    NSURL *localDocumentURL = documentViewController.localDocumentURL;
    
    if (localDocumentURL) {
        [self PT_addDocumentViewController:documentViewController forURL:localDocumentURL];
        
        NSURL *oldLocalDocumentURL = change.oldValue;

        if ([oldLocalDocumentURL isEqual:[NSNull null]] &&
            ![documentViewController.documentUrl isFileURL]) {
            oldLocalDocumentURL = documentViewController.documentUrl;
        }

        if (oldLocalDocumentURL) {
            NSMutableArray<NSURL *> *mutableURLs = [self.tabURLs mutableCopy];
            NSUInteger index = [mutableURLs indexOfObject:oldLocalDocumentURL];
            if (index != NSNotFound) {
                [mutableURLs replaceObjectAtIndex:index withObject:localDocumentURL];
            }
            self.tabURLs = [mutableURLs copy];
        }
        
        [self.tabBar reloadItems];
    }
}

@end
